@@ -157,7 +157,7 @@ applier_check_sync(struct applier *applier)
 	if (applier->state == APPLIER_SYNC &&
 	    applier->lag <= replication_sync_lag &&
 	    vclock_compare_ignore0(&applier->remote_vclock_at_subscribe,
-				   &replicaset.vclock) <= 0) {
+				   instance_vclock) <= 0) {
 		/* Applier is synced, switch to "follow". */
 		applier_set_state(applier, APPLIER_FOLLOW);
 	}
@@ -181,6 +181,10 @@ applier_fiber_new(struct applier *applier, const char *name, fiber_func func,
 	return f;
 }
 
+/** Apply all rows in the rows queue as a single transaction. */
+static int
+applier_apply_tx(struct applier *applier, struct stailq *rows);
+
 /*
  * Fiber function to write vclock to replication master.
  * To track connection status, replica answers master
@@ -660,6 +664,7 @@ applier_connect(struct applier *applier)
 				&greeting);
 
 	applier->last_row_time = ev_monotonic_now(loop());
+	applier->txn_last_tm = 0;
 
 	if (applier->version_id != greeting.version_id) {
 		say_info("remote master %s at %s running Tarantool %u.%u.%u",
@@ -783,7 +788,9 @@ applier_wait_snapshot(struct applier *applier)
 		 * Used to initialize the replica's initial
 		 * vclock in bootstrap_from_master()
 		 */
-		xrow_decode_vclock_xc(&row, &replicaset.vclock);
+		struct vclock vclock;
+		xrow_decode_vclock_ignore0_xc(&row, &vclock);
+		box_init_instance_vclock(&vclock);
 	}
 
 	coio_read_xrow(io, ibuf, &row);
@@ -795,8 +802,11 @@ applier_wait_snapshot(struct applier *applier)
 				xrow_decode_error_xc(&row);
 			} else if (iproto_type_is_promote_request(row.type)) {
 				struct synchro_request req;
-				if (xrow_decode_synchro(&row, &req) != 0)
+				struct vclock limbo_vclock;
+				if (xrow_decode_synchro(&row, &req,
+							&limbo_vclock) != 0) {
 					diag_raise();
+				}
 				if (txn_limbo_process(&txn_limbo, &req) != 0)
 					diag_raise();
 			} else if (iproto_type_is_raft_request(row.type)) {
@@ -836,7 +846,9 @@ applier_wait_snapshot(struct applier *applier)
 				 * vclock yet, do it now. In 1.7+
 				 * this vclock is not used.
 				 */
-				xrow_decode_vclock_xc(&row, &replicaset.vclock);
+				struct vclock vclock;
+				xrow_decode_vclock_ignore0_xc(&row, &vclock);
+				box_init_instance_vclock(&vclock);
 			}
 			break; /* end of stream */
 		} else if (iproto_type_is_error(row.type)) {
@@ -858,8 +870,17 @@ applier_fetch_snapshot(struct applier *applier)
 	struct iostream *io = &applier->io;
 	struct xrow_header row;
 
-	memset(&row, 0, sizeof(row));
-	row.type = IPROTO_FETCH_SNAPSHOT;
+	struct vclock vclock;
+	vclock_create(&vclock);
+	struct fetch_snapshot_request req = {
+		.version_id = tarantool_version_id(),
+		/* Applier doesn't support checkpoint join. */
+		.is_checkpoint_join = false,
+		.checkpoint_vclock = vclock,
+		.checkpoint_lsn = 0,
+	};
+	RegionGuard region_guard(&fiber()->gc);
+	xrow_encode_fetch_snapshot(&row, &req);
 	coio_write_xrow(io, &row);
 
 	applier_set_state(applier, APPLIER_WAIT_SNAPSHOT);
@@ -878,9 +899,6 @@ static uint64_t
 applier_read_tx(struct applier *applier, struct stailq *rows,
 		const struct applier_read_ctx *ctx, double timeout);
 
-static int
-apply_final_join_tx(uint32_t replica_id, struct stailq *rows);
-
 /**
  * A helper struct to link xrow objects in a list.
  */
@@ -996,7 +1014,7 @@ applier_wait_register(struct applier *applier, uint64_t row_count)
 						  next)->row);
 			break;
 		}
-		if (apply_final_join_tx(applier->instance_id, &rows) != 0)
+		if (applier_apply_tx(applier, &rows) != 0)
 			diag_raise();
 	}
 
@@ -1102,7 +1120,7 @@ applier_parse_tx_row(struct applier_tx_row *tx_row)
 			diag_raise();
 		}
 	} else if (iproto_type_is_synchro_request(type)) {
-		if (xrow_decode_synchro(row, &tx_row->req.synchro) != 0) {
+		if (xrow_decode_synchro(row, &tx_row->req.synchro, NULL) != 0) {
 			diag_raise();
 		}
 	} else if (iproto_type_is_raft_request(type)) {
@@ -1211,14 +1229,15 @@ applier_rollback_by_wal_io(int64_t signature)
 	trigger_run(&replicaset.applier.on_rollback, NULL);
 
 	/* Rollback applier vclock to the committed one. */
-	vclock_copy(&replicaset.applier.vclock, &replicaset.vclock);
+	vclock_copy(&replicaset.applier.vclock, instance_vclock);
 }
 
 static int
 applier_txn_rollback_cb(struct trigger *trigger, void *event)
 {
-	(void) trigger;
-	struct txn *txn = (struct txn *) event;
+	(void)trigger;
+	(void)event;
+	struct txn *txn = in_txn();
 	/*
 	 * Synchronous transaction rollback due to receiving a
 	 * ROLLBACK entry is a normal event and requires no
@@ -1245,8 +1264,8 @@ static void
 replica_txn_wal_write_cb(struct replica_cb_data *rcb)
 {
 	struct replica *r = replica_by_id(rcb->replica_id);
-	if (likely(r != NULL))
-		r->applier_txn_last_tm = rcb->txn_last_tm;
+	if (likely(r != NULL && r->applier != NULL))
+		r->applier->txn_last_tm = rcb->txn_last_tm;
 }
 
 static int
@@ -1374,8 +1393,7 @@ applier_handle_raft_request(struct applier *applier, struct raft_request *req)
 }
 
 static int
-apply_plain_tx(uint32_t replica_id, struct stailq *rows,
-	       bool skip_conflict, bool use_triggers)
+apply_plain_tx(uint32_t replica_id, struct stailq *rows)
 {
 	/*
 	 * Explicitly begin the transaction so that we can
@@ -1392,7 +1410,7 @@ apply_plain_tx(uint32_t replica_id, struct stailq *rows,
 	stailq_foreach_entry(item, rows, next) {
 		struct xrow_header *row = &item->row;
 		int res = apply_request(&item->req.dml);
-		if (res != 0 && skip_conflict) {
+		if (res != 0 && replication_skip_conflict) {
 			struct error *e = diag_last_error(diag_get());
 			/*
 			 * In case of ER_TUPLE_FOUND error and enabled
@@ -1429,80 +1447,51 @@ apply_plain_tx(uint32_t replica_id, struct stailq *rows,
 			 "distributed transactions");
 		goto fail;
 	}
-
-	if (use_triggers) {
-		/* We are ready to submit txn to wal. */
-		struct trigger *on_rollback, *on_wal_write;
-		size_t size;
-		on_rollback = region_alloc_object(&txn->region, typeof(*on_rollback),
-						  &size);
-		on_wal_write = region_alloc_object(&txn->region, typeof(*on_wal_write),
-						   &size);
-		if (on_rollback == NULL || on_wal_write == NULL) {
-			diag_set(OutOfMemory, size, "region_alloc_object",
-				 "on_rollback/on_wal_write");
-			goto fail;
-		}
-
-		struct replica_cb_data *rcb;
-		rcb = region_alloc_object(&txn->region, typeof(*rcb), &size);
-		if (rcb == NULL) {
-			diag_set(OutOfMemory, size, "region_alloc_object", "rcb");
-			goto fail;
-		}
-
-		trigger_create(on_rollback, applier_txn_rollback_cb, NULL, NULL);
-		txn_on_rollback(txn, on_rollback);
-
-		/*
-		 * We use *last* entry timestamp because ack comes up to
-		 * last entry in transaction. Same time this shows more
-		 * precise result because we're interested in how long
-		 * transaction traversed network + remote WAL bundle before
-		 * ack get received.
-		 */
-		item = stailq_last_entry(rows, struct applier_tx_row, next);
-		rcb->replica_id = replica_id;
-		rcb->txn_last_tm = item->row.tm;
-
-		trigger_create(on_wal_write, applier_txn_wal_write_cb, rcb, NULL);
-		txn_on_wal_write(txn, on_wal_write);
+	/* We are ready to submit txn to wal. */
+	struct trigger *on_rollback, *on_wal_write;
+	size_t size;
+	on_rollback = region_alloc_object(&txn->region, typeof(*on_rollback),
+					  &size);
+	on_wal_write = region_alloc_object(&txn->region, typeof(*on_wal_write),
+					   &size);
+	if (on_rollback == NULL || on_wal_write == NULL) {
+		diag_set(OutOfMemory, size, "region_alloc_object",
+			 "on_rollback/on_wal_write");
+		goto fail;
 	}
+	struct replica_cb_data *rcb;
+	rcb = region_alloc_object(&txn->region, typeof(*rcb), &size);
+	if (rcb == NULL) {
+		diag_set(OutOfMemory, size, "region_alloc_object", "rcb");
+		goto fail;
+	}
+	trigger_create(on_rollback, applier_txn_rollback_cb, NULL, NULL);
+	txn_on_rollback(txn, on_rollback);
+	/*
+	 * We use *last* entry timestamp because ack comes up to last entry in
+	 * transaction. Same time this shows more precise result because we're
+	 * interested in how long transaction traversed network + remote WAL
+	 * bundle before ack get received.
+	 */
+	item = stailq_last_entry(rows, struct applier_tx_row, next);
+	rcb->replica_id = replica_id;
+	rcb->txn_last_tm = item->row.tm;
 
-	return txn_commit_try_async(txn);
+	trigger_create(on_wal_write, applier_txn_wal_write_cb, rcb, NULL);
+	txn_on_wal_write(txn, on_wal_write);
+	return txn_commit_submit(txn);
 fail:
 	txn_abort(txn);
 	return -1;
 }
 
-/** A simpler version of applier_apply_tx() for final join stage. */
-static int
-apply_final_join_tx(uint32_t replica_id, struct stailq *rows)
-{
-	struct applier_tx_row *txr =
-		stailq_first_entry(rows, struct applier_tx_row, next);
-
-	struct xrow_header *last_row =
-		&stailq_last_entry(rows, struct applier_tx_row, next)->row;
-	int rc = 0;
-	/* WAL isn't enabled yet, so follow vclock manually. */
-	vclock_follow_xrow(&replicaset.vclock, last_row);
-	if (unlikely(iproto_type_is_synchro_request(txr->row.type))) {
-		rc = apply_synchro_req(replica_id, &txr->row,
-				       &txr->req.synchro);
-	} else {
-		rc = apply_plain_tx(replica_id, rows, false, false);
-	}
-	return rc;
-}
-
 /**
  * We must filter out synchronous rows coming from an instance that fell behind
  * the current synchro queue owner. This includes both synchronous tx rows and
  * rows for txs following unconfirmed synchronous transactions.
  * The rows are replaced with NOPs to preserve the vclock consistency.
  */
-static void
+static int
 applier_synchro_filter_tx(struct stailq *rows)
 {
 	latch_lock(&txn_limbo.promote_latch);
@@ -1511,38 +1500,102 @@ applier_synchro_filter_tx(struct stailq *rows)
 	});
 	struct xrow_header *row;
 	/*
-	 * It  may happen that we receive the instance's rows via some third
+	 * It may happen that we receive the instance's rows via some third
 	 * node, so cannot check for applier->instance_id here.
 	 */
-	row = &stailq_first_entry(rows, struct applier_tx_row, next)->row;
+	row = &stailq_last_entry(rows, struct applier_tx_row, next)->row;
 	uint64_t term = txn_limbo_replica_term(&txn_limbo, row->replica_id);
 	assert(term <= txn_limbo.promote_greatest_term);
-	if (term == txn_limbo.promote_greatest_term)
-		return;
-
+	bool is_current_term = term == txn_limbo.promote_greatest_term;
+	struct applier_tx_row *item;
+	if (iproto_type_is_dml(row->type)) {
+		if (is_current_term)
+			return 0;
+		/*
+		 * Any asynchronous transaction from an obsolete term when limbo
+		 * is claimed by someone is a marker of split-brain by itself:
+		 * consider it a synchronous transaction, which is committed
+		 * with quorum 1.
+		 */
+		if (row->wait_sync)
+			goto nopify;
+		if (txn_limbo.owner_id == REPLICA_ID_NIL)
+			return 0;
+		stailq_foreach_entry(item, rows, next) {
+			row = &item->row;
+			if (row->type == IPROTO_NOP)
+				continue;
+			diag_set(ClientError, ER_SPLIT_BRAIN,
+				 "got an async transaction from an old term");
+			return -1;
+		}
+		return 0;
+	}
 	/*
-	 * We do not nopify promotion/demotion and confirm/rollback.
+	 * We do not nopify promotion/demotion and most of confirm/rollback.
 	 * Such syncrhonous requests should be filtered by txn_limbo to detect
 	 * possible split brain situations.
 	 *
 	 * This means the only filtered out transactions are synchronous ones or
 	 * the ones depending on them.
-	 *
-	 * Any asynchronous transaction from an obsolete term when limbo is
-	 * claimed by someone is a marker of split-brain by itself: consider it
-	 * a synchronous transaction, which is committed with quorum 1.
 	 */
-	struct xrow_header *last_row =
-		&stailq_last_entry(rows, struct applier_tx_row, next)->row;
-	if (!last_row->wait_sync) {
-		if (iproto_type_is_dml(last_row->type) &&
-		    txn_limbo.owner_id != REPLICA_ID_NIL) {
-			tnt_raise(ClientError, ER_SPLIT_BRAIN,
-				  "got an async transaction from an old term");
-		}
-		return;
+	assert(iproto_type_is_synchro_request(row->type));
+	const struct synchro_request *req;
+	int64_t confirmed_lsn;
+	item = stailq_last_entry(rows, typeof(*item), next);
+	req = &item->req.synchro;
+	/* Note! Might be different from row->replica_id. */
+	confirmed_lsn = txn_limbo_replica_confirmed_lsn(&txn_limbo,
+							req->replica_id);
+	switch (row->type) {
+	case IPROTO_RAFT_PROMOTE:
+	case IPROTO_RAFT_DEMOTE:
+		/*
+		 * Never need to be nopified. Every PROMOTE / DEMOTE is either a
+		 * valid one or is a split brain. There is no such thing as an
+		 * "old already applied promotion".
+		 */
+		return 0;
+	case IPROTO_RAFT_CONFIRM:
+		if (req->lsn > confirmed_lsn)
+			return 0;
+		/*
+		 * A CONFIRM with lsn <= known confirm lsn for this replica may
+		 * be nopified without a second thought. The transactions it's
+		 * going to confirm were already confirmed by one of the
+		 * PROMOTE/DEMOTE requests in a new term.
+		 *
+		 * See that the CONFIRM can be nopified even in the current term
+		 * if it wants to commit already committed txns. This is a niche
+		 * case which might happen when a replica joins a master and
+		 * receives a valid fully confirmed read-view from it, but some
+		 * CONFIRM WAL entries might have been written by the master
+		 * after the read-view is sent. Then the replica would receive
+		 * those "already known" CONFIRMs during xlogs catch up.
+		 *
+		 * Besides, logically a confirmation of already confirmed txns
+		 * doesn't contradict anything.
+		 */
+		break;
+	case IPROTO_RAFT_ROLLBACK:
+		if (req->lsn <= confirmed_lsn)
+			return 0;
+		/*
+		 * Rollback in the current term wants to roll some currently
+		 * waiting transactions back. No case when it can be considered
+		 * outdated.
+		 */
+		if (is_current_term)
+			return 0;
+		/*
+		 * In older terms though this is fine to nopify it. Those txns
+		 * must have already been cancelled by the new leader anyway.
+		 */
+		break;
+	default:
+		unreachable();
 	}
-	struct applier_tx_row *item;
+nopify:
 	stailq_foreach_entry(item, rows, next) {
 		row = &item->row;
 		row->type = IPROTO_NOP;
@@ -1551,6 +1604,7 @@ applier_synchro_filter_tx(struct stailq *rows)
 		item->req.dml.header = row;
 		item->req.dml.type = IPROTO_NOP;
 	}
+	return 0;
 }
 
 /**
@@ -1579,11 +1633,6 @@ applier_process_heartbeat(struct applier *applier, struct applier_tx_row *txr)
 	return 0;
 }
 
-/**
- * Apply all rows in the rows queue as a single transaction.
- *
- * Return 0 for success or -1 in case of an error.
- */
 static int
 applier_apply_tx(struct applier *applier, struct stailq *rows)
 {
@@ -1643,7 +1692,10 @@ applier_apply_tx(struct applier *applier, struct stailq *rows)
 			}
 		}
 	}
-	applier_synchro_filter_tx(rows);
+	rc = applier_synchro_filter_tx(rows);
+	if (rc != 0)
+		goto finish;
+
 	if (unlikely(iproto_type_is_synchro_request(first_row->type))) {
 		/*
 		 * Synchro messages are not transactions, in terms
@@ -1654,8 +1706,7 @@ applier_apply_tx(struct applier *applier, struct stailq *rows)
 		rc = apply_synchro_req(applier->instance_id, &txr->row,
 				       &txr->req.synchro);
 	} else {
-		rc = apply_plain_tx(applier->instance_id, rows,
-				    replication_skip_conflict, true);
+		rc = apply_plain_tx(applier->instance_id, rows);
 	}
 	if (rc != 0)
 		goto finish;
@@ -1680,12 +1731,15 @@ applier_signal_ack(struct applier *applier)
 		 * timestamp in tm field. If user delete the node from _cluster
 		 * space, we obtain a nil pointer here.
 		 */
-		struct replica *r = replica_by_id(applier->instance_id);
-		applier->ack_msg.txn_last_tm = (r == NULL ? 0 :
-						r->applier_txn_last_tm);
+		applier->ack_msg.txn_last_tm = applier->txn_last_tm;
+		/*
+		 * Send each timestamp only once. New timestamp is treated by
+		 * relay like something new was acked from that specific relay.
+		 */
+		applier->txn_last_tm = 0;
 		applier->ack_msg.vclock_sync = applier->last_vclock_sync;
 		applier->ack_msg.term = box_raft()->term;
-		vclock_copy(&applier->ack_msg.vclock, &replicaset.vclock);
+		vclock_copy(&applier->ack_msg.vclock, instance_vclock);
 		cmsg_init(&applier->ack_msg.base, applier->ack_route);
 		cpipe_push(&applier->applier_thread->thread_pipe,
 			   &applier->ack_msg.base);
@@ -2105,7 +2159,7 @@ applier_init(void)
 						  sizeof(struct applier_thread *));
 	for (int i = 0; i < replication_threads; i++) {
 		struct applier_thread *thread =
-			(struct applier_thread *)xmalloc(sizeof(*thread));
+			xalloc_object(struct applier_thread);
 		applier_thread_create(thread);
 		applier_threads[i] = thread;
 	}
@@ -2225,6 +2279,8 @@ applier_thread_detach_applier(struct cbus_call_msg *base)
 	applier->thread.reader = NULL;
 	lsregion_destroy(&applier->thread.lsr);
 	fiber_cond_destroy(&applier->thread.writer_cond);
+	ibuf_destroy(&applier->thread.ibuf);
+
 	return 0;
 }
 
@@ -2312,7 +2368,7 @@ applier_subscribe(struct applier *applier)
 
 	struct subscribe_request req;
 	memset(&req, 0, sizeof(req));
-	vclock_copy(&req.vclock, &replicaset.vclock);
+	vclock_copy(&req.vclock, instance_vclock);
 	ERROR_INJECT(ERRINJ_REPLICASET_VCLOCK, {
 		vclock_create(&req.vclock);
 	});
@@ -2437,6 +2493,7 @@ applier_subscribe(struct applier *applier)
 	 * Process a stream of rows from the binary log.
 	 */
 	while (true) {
+		ERROR_INJECT_YIELD(ERRINJ_APPLIER_SUBSCRIBE_DELAY);
 		if (applier->pending_msg_cnt == 0) {
 			fiber_cond_wait(&applier->msg_cond);
 		}
@@ -2504,6 +2561,19 @@ applier_f(va_list ap)
 				 * is retrying final join.
 				 */
 				applier_register(applier, was_anon);
+				/*
+				 * It's hard to catch the exact moment when
+				 * instance id settles, for example, it could
+				 * change multiple times during register, but we
+				 * have to configure raft with it before
+				 * `applier_subscribe` starts, because we'll
+				 * start receiving raft messages there.
+				 */
+				if (was_anon) {
+					assert(instance_id != 0);
+					raft_cfg_instance_id(box_raft(),
+							     instance_id);
+				}
 			}
 			applier_subscribe(applier);
 			/*
@@ -2669,8 +2739,8 @@ applier_stop(struct applier *applier)
 struct applier *
 applier_new(struct uri *uri)
 {
-	struct applier *applier = (struct applier *)
-		xcalloc(1, sizeof(struct applier));
+	struct applier *applier = xalloc_object(struct applier);
+	memset(applier, 0, sizeof(*applier));
 	if (iostream_ctx_create(&applier->io_ctx, IOSTREAM_CLIENT, uri) != 0) {
 		free(applier);
 		diag_raise();
