@@ -57,12 +57,18 @@ struct vclock;
  * Pointer to TX thread local vclock.
  *
  * During recovery it points to the current recovery position.
- * Once recovery is complete, it is set to &replicaset.vclock.
+ * Once recovery is complete, it is set to instance_vclock.
  *
  * We need it for reporting the actual vclock in box.info while
  * the instance is in hot standby mode.
  */
 extern const struct vclock *box_vclock;
+/**
+ * Vclock of the instance as it is stored on disk. It means that during
+ * recovery this vclock points to the end of WAL, not to the current recovery
+ * position.
+ */
+extern const struct vclock *instance_vclock;
 
 /**
  * Name of the authentication method that is currently used on
@@ -100,6 +106,18 @@ extern bool box_is_force_recovery;
 void
 box_init(void);
 
+/**
+ * Set instance's vclock to the given value. Works only if the instance vclock
+ * isn't already set. It must be initialized manually and explicitly this way
+ * when the instance boots from a snapshot - snapshot rows have no LSNs, so
+ * setting the instance vclock manually is the only way.
+ *
+ * It has to be a "public" API of box, because the snapshot can be remote, i.e.
+ * coming from the applier.
+ */
+void
+box_init_instance_vclock(const struct vclock *vclock);
+
 /**
  * Cleanup box library
  */
@@ -154,12 +172,18 @@ box_check_writable(void);
 void
 box_set_ro(void);
 
-bool
+/** \cond public */
+
+API_EXPORT bool
 box_is_ro(void);
 
+/** \endcond public */
+
 bool
 box_is_orphan(void);
 
+/** \cond public */
+
 /**
  * Wait until the instance switches to a desired mode.
  * \param ro wait read-only if set or read-write if unset
@@ -167,9 +191,11 @@ box_is_orphan(void);
  * \retval -1 timeout or fiber is cancelled
  * \retval 0 success
  */
-int
+API_EXPORT int
 box_wait_ro(bool ro, double timeout);
 
+/** \endcond public */
+
 /**
  * Switch this instance from 'orphan' to 'running' state or
  * vice versa depending on the value of the function argument.
@@ -195,13 +221,17 @@ box_do_set_orphan(bool orphan);
 void
 box_update_ro_summary(void);
 
+/** \cond public */
+
 /**
  * Get the reason why the instance is read only if it is. Can't be called on a
  * writable instance.
  */
-const char *
+API_EXPORT const char *
 box_ro_reason(void);
 
+/** \endcond public */
+
 /**
  * Iterate over all spaces and save them to the
  * snapshot file.
@@ -338,6 +368,8 @@ box_init_say();
 extern "C" {
 #endif /* defined(__cplusplus) */
 
+int box_set_wal_retention_period(void);
+
 typedef struct tuple box_tuple_t;
 
 bool
@@ -688,6 +720,13 @@ API_EXPORT int
 box_iproto_override(uint32_t req_type, iproto_handler_t handler,
 		    iproto_handler_destroy_t destroy, void *ctx);
 
+/**
+ * Get self LSN component of box vclock, -1 if no one or bootstrap haven't
+ * succeeded.
+ */
+API_EXPORT int64_t
+box_info_lsn(void);
+
 /** \endcond public */
 
 /**
