@@ -194,6 +194,21 @@ extern const char *iproto_flag_bit_strs[];
 	 * authentication method.
 	 */								\
 	_(AUTH_TYPE, 0x5b, MP_STR)					\
+	 /**
+	  * Flag indicating whether checkpoint join should be done.
+	  */								\
+	 _(IS_CHECKPOINT_JOIN, 0x62, MP_BOOL)				\
+	 /**
+	  * Shows the signature of the checkpoint to read from.
+	  * Requires CHECKPOINT_JOIN to be true.
+	  */								\
+	 _(CHECKPOINT_VCLOCK, 0x63, MP_MAP)				\
+	 /**
+	  * Shows the lsn to start sending from. Server sends all rows
+	  * >= IPROTO_CHECKPOINT_LSN. Requires CHECKPOINT_JOIN to be
+	  * true and CHECKPOINT_VCLOCK to be set.
+	  */								\
+	 _(CHECKPOINT_LSN, 0x64, MP_UINT)				\
 
 #define IPROTO_KEY_MEMBER(s, v, ...) IPROTO_ ## s = v,
 
@@ -437,7 +452,7 @@ iproto_type_name(uint16_t type)
 }
 
 /** Predefined replication group identifiers. */
-enum {
+enum group_id {
 	/**
 	 * Default replication group: changes made to the space
 	 * are replicated throughout the entire cluster.
@@ -547,11 +562,13 @@ request_replace_body_create(struct request_replace_body *body,
 	/** Number of pages in the run. */				\
 	_(PAGE_COUNT, 5)						\
 	/** Legacy bloom filter implementation. */			\
-	_(BLOOM_FILTER_LEGACY, 6)					\
-	/** Bloom filter for keys. */					\
-	_(BLOOM_FILTER, 7)						\
+	_(BLOOM_FILTER_LEGACY_V1, 6)					\
+	/** Legacy bloom filter implementation. */			\
+	_(BLOOM_FILTER_LEGACY_V2, 7)					\
 	/** Number of statements of each type (map). */			\
 	_(STMT_STAT, 8)							\
+	/** Bloom filter for keys. */					\
+	_(BLOOM_FILTER, 9)						\
 
 #define VY_RUN_INFO_KEY_MEMBER(s, v) VY_RUN_INFO_ ## s = v,
 
