@@ -268,6 +268,7 @@ struct key_def *
 key_def_new(const struct key_part_def *parts, uint32_t part_count,
 	    bool for_func_index)
 {
+	assert(part_count > 0);
 	size_t sz = 0;
 	for (uint32_t i = 0; i < part_count; i++)
 		sz += parts[i].path != NULL ? strlen(parts[i].path) : 0;
@@ -324,7 +325,7 @@ error:
 	return NULL;
 }
 
-int
+void
 key_def_dump_parts(const struct key_def *def, struct key_part_def *parts,
 		   struct region *region)
 {
@@ -338,12 +339,7 @@ key_def_dump_parts(const struct key_def *def, struct key_part_def *parts,
 		part_def->nullable_action = part->nullable_action;
 		part_def->coll_id = part->coll_id;
 		if (part->path != NULL) {
-			char *path = region_alloc(region, part->path_len + 1);
-			if (path == NULL) {
-				diag_set(OutOfMemory, part->path_len + 1,
-					 "region", "part_def->path");
-				return -1;
-			}
+			char *path = xregion_alloc(region, part->path_len + 1);
 			memcpy(path, part->path, part->path_len);
 			path[part->path_len] = '\0';
 			part_def->path = path;
@@ -351,7 +347,6 @@ key_def_dump_parts(const struct key_def *def, struct key_part_def *parts,
 			part_def->path = NULL;
 		}
 	}
-	return 0;
 }
 
  /* {{{ Module API helpers */
@@ -988,10 +983,12 @@ key_def_decode_parts(struct key_part_def *parts, uint32_t part_count,
 const struct key_part *
 key_def_find_by_fieldno(const struct key_def *key_def, uint32_t fieldno)
 {
-	struct key_part part;
-	memset(&part, 0, sizeof(struct key_part));
-	part.fieldno = fieldno;
-	return key_def_find(key_def, &part);
+	const struct key_part *part = key_def->parts;
+	const struct key_part *end = part + key_def->part_count;
+	for (; part != end; part++)
+		if (part->fieldno == fieldno)
+			return part;
+	return NULL;
 }
 
 const struct key_part *
@@ -1143,16 +1140,10 @@ key_def_find_pk_in_cmp_def(const struct key_def *cmp_def,
 	size_t region_svp = region_used(region);
 
 	/* First, dump primary key parts as is. */
-	size_t size;
-	struct key_part_def *parts =
-		region_alloc_array(region, typeof(parts[0]), pk_def->part_count,
-				   &size);
-	if (parts == NULL) {
-		diag_set(OutOfMemory, size, "region_alloc_array", "parts");
-		goto out;
-	}
-	if (key_def_dump_parts(pk_def, parts, region) != 0)
-		goto out;
+	struct key_part_def *parts = xregion_alloc_array(
+			region, typeof(parts[0]), pk_def->part_count);
+	key_def_dump_parts(pk_def, parts, region);
+
 	/*
 	 * Second, update field numbers to match the primary key
 	 * parts in a secondary key.
@@ -1167,7 +1158,6 @@ key_def_find_pk_in_cmp_def(const struct key_def *cmp_def,
 
 	/* Finally, allocate the new key definition. */
 	extracted_def = key_def_new(parts, pk_def->part_count, false);
-out:
 	region_truncate(region, region_svp);
 	return extracted_def;
 }
