@@ -163,9 +163,6 @@ typedef char *(*tuple_extract_key_raw_t)(const char *data,
 /** @copydoc tuple_hash() */
 typedef uint32_t (*tuple_hash_t)(struct tuple *tuple,
 				 struct key_def *key_def);
-/** @copydoc key_hash() */
-typedef uint32_t (*key_hash_t)(const char *key,
-				struct key_def *key_def);
 /** @copydoc tuple_hint() */
 typedef hint_t (*tuple_hint_t)(struct tuple *tuple,
 			       struct key_def *key_def);
@@ -185,8 +182,6 @@ struct key_def {
 	tuple_extract_key_raw_t tuple_extract_key_raw;
 	/** @see tuple_hash() */
 	tuple_hash_t tuple_hash;
-	/** @see key_hash() */
-	key_hash_t key_hash;
 	/** @see tuple_hint() */
 	tuple_hint_t tuple_hint;
 	/** @see key_hint() */
@@ -660,9 +655,8 @@ key_def_new(const struct key_part_def *parts, uint32_t part_count,
 /**
  * Dump part definitions of the given key def.
  * The region is used for allocating JSON paths, if any.
- * Return -1 on memory allocation error, 0 on success.
  */
-int
+void
 key_def_dump_parts(const struct key_def *def, struct key_part_def *parts,
 		   struct region *region);
 
@@ -714,7 +708,9 @@ key_def_decode_parts(struct key_part_def *parts, uint32_t part_count,
 		     uint32_t field_count, struct region *region);
 
 /**
- * Returns the part in index_def->parts for the specified fieldno.
+ * Returns the first part in index_def->parts for the specified fieldno.
+ * The part is returned regardless of whether it is indexed by JSON path or not.
+ *
  * If fieldno is not in index_def->parts returns NULL.
  */
 const struct key_part *
@@ -1067,9 +1063,11 @@ tuple_compare_with_key(struct tuple *tuple, hint_t tuple_hint,
 					       part_count, key_hint, key_def);
 }
 
+uint32_t
+tuple_hash_null(uint32_t *ph1, uint32_t *pcarry);
+
 /**
  * Compute hash of a tuple field.
- * @param type - type of the field key part
  * @param ph1 - pointer to running hash
  * @param pcarry - pointer to carry
  * @param field - pointer to field data
@@ -1081,7 +1079,7 @@ tuple_compare_with_key(struct tuple *tuple, hint_t tuple_hint,
  */
 uint32_t
 tuple_hash_field(uint32_t *ph1, uint32_t *pcarry, const char **field,
-		 enum field_type type, struct coll *coll);
+		 struct coll *coll);
 
 /**
  * Compute hash of a key part.
@@ -1116,11 +1114,8 @@ tuple_hash(struct tuple *tuple, struct key_def *key_def)
  * @param key_def - key_def for field description
  * @return - hash value
  */
-static inline uint32_t
-key_hash(const char *key, struct key_def *key_def)
-{
-	return key_def->key_hash(key, key_def);
-}
+uint32_t
+key_hash(const char *key, struct key_def *key_def);
 
  /*
  * Get comparison hint for a tuple.
