@@ -854,11 +854,7 @@ const char console_lua[] =
 "    -- We don't know if the remote end is binary or Lua console so we first try\n"
 "    -- to connect to it as binary using net.box and fall back on Lua console if\n"
 "    -- it fails.\n"
-"    local remote = net_box.connect(u.host, u.service, {\n"
-"            connect_timeout = opts.timeout,\n"
-"            user = u.login,\n"
-"            password = u.password,\n"
-"    })\n"
+"    local remote = net_box.connect(uri, {connect_timeout = opts.timeout})\n"
 "    if remote.state == 'error' then\n"
 "        local err = remote.error\n"
 "        remote = nil\n"
