@@ -168,7 +168,7 @@ lbox_pushreplica(lua_State *L, struct replica *replica)
 	lua_settable(L, -3);
 
 	lua_pushstring(L, "lsn");
-	luaL_pushuint64(L, vclock_get(&replicaset.vclock, replica->id));
+	luaL_pushuint64(L, vclock_get(instance_vclock, replica->id));
 	lua_settable(L, -3);
 
 	if (applier != NULL && applier->state != APPLIER_OFF) {
@@ -285,13 +285,7 @@ lbox_info_uuid(struct lua_State *L)
 static int
 lbox_info_lsn(struct lua_State *L)
 {
-	/* See comments in lbox_info_id */
-	struct replica *self = replica_by_uuid(&INSTANCE_UUID);
-	if (self != NULL && (self->id != REPLICA_ID_NIL || replication_anon)) {
-		luaL_pushint64(L, vclock_get(box_vclock, self->id));
-	} else {
-		luaL_pushint64(L, -1);
-	}
+	luaL_pushint64(L, box_info_lsn());
 	return 1;
 }
 
@@ -380,6 +374,9 @@ lbox_info_cluster(struct lua_State *L)
 static int
 lbox_info_memory_call(struct lua_State *L)
 {
+	if (box_check_configured() != 0)
+		return luaT_error(L);
+
 	struct engine_memory_stat stat;
 	engine_memory_stat(&stat);
 
@@ -408,7 +405,7 @@ lbox_info_memory_call(struct lua_State *L)
 	lua_settable(L, -3);
 
 	lua_pushstring(L, "lua");
-	lua_pushinteger(L, G(L)->gc.total);
+	lua_pushinteger(L, luaL_getgctotal(L));
 	lua_settable(L, -3);
 
 	return 1;
@@ -432,6 +429,9 @@ lbox_info_memory(struct lua_State *L)
 static int
 lbox_info_gc_call(struct lua_State *L)
 {
+	if (box_check_configured() != 0)
+		return luaT_error(L);
+
 	int count;
 
 	lua_newtable(L);
@@ -452,6 +452,15 @@ lbox_info_gc_call(struct lua_State *L)
 	lua_pushboolean(L, gc.is_paused);
 	lua_settable(L, -3);
 
+	lua_pushstring(L, "wal_retention_vclock");
+	struct vclock retention_vclock;
+	wal_get_retention_vclock(&retention_vclock);
+	if (vclock_is_set(&retention_vclock))
+		luaT_pushvclock(L, &retention_vclock);
+	else
+		luaL_pushnull(L);
+	lua_settable(L, -3);
+
 	lua_pushstring(L, "checkpoints");
 	lua_newtable(L);
 
@@ -530,6 +539,9 @@ lbox_info_gc(struct lua_State *L)
 static int
 lbox_info_vinyl_call(struct lua_State *L)
 {
+	if (box_check_configured() != 0)
+		return luaT_error(L);
+
 	struct info_handler h;
 	luaT_info_handler_create(&h, L);
 	struct engine *vinyl = engine_by_name("vinyl");
@@ -557,6 +569,9 @@ lbox_info_vinyl(struct lua_State *L)
 static int
 lbox_info_sql_call(struct lua_State *L)
 {
+	if (box_check_configured() != 0)
+		return luaT_error(L);
+
 	struct info_handler h;
 	luaT_info_handler_create(&h, L);
 	sql_stmt_cache_stat(&h);
