@@ -82,6 +82,12 @@ const char load_cfg_lua[] =
 "    end\n"
 "end\n"
 "\n"
+"local function ifdef_wal_retention_period(value)\n"
+"    if private.cfg_set_wal_retention_period ~= nil then\n"
+"        return value\n"
+"    end\n"
+"end\n"
+"\n"
 "-- all available options\n"
 "local default_cfg = {\n"
 "    listen              = nil,\n"
@@ -152,6 +158,7 @@ const char load_cfg_lua[] =
 "    wal_dir_rescan_delay= 2,\n"
 "    wal_queue_max_size  = 16 * 1024 * 1024,\n"
 "    wal_cleanup_delay   = 4 * 3600,\n"
+"    experimental_wal_retention_period = ifdef_wal_retention_period(0),\n"
 "    wal_ext             = ifdef_wal_ext(nil),\n"
 "    force_recovery      = false,\n"
 "    replication         = nil,\n"
@@ -341,6 +348,7 @@ const char load_cfg_lua[] =
 "    wal_max_size        = 'number',\n"
 "    wal_dir_rescan_delay= 'number',\n"
 "    wal_cleanup_delay   = 'number',\n"
+"    experimental_wal_retention_period = ifdef_wal_retention_period('number'),\n"
 "    wal_ext             = ifdef_wal_ext('table'),\n"
 "    force_recovery      = 'boolean',\n"
 "    replication         = 'string, number, table',\n"
@@ -400,31 +408,49 @@ const char load_cfg_lua[] =
 "    replication        = normalize_uri_list_for_replication,\n"
 "}\n"
 "\n"
-"local function purge_password_from_uri(uri)\n"
+"local function purge_sensitive_data_from_uri(uri)\n"
 "    local parsed = urilib.parse(uri)\n"
-"    if parsed ~= nil and parsed.password ~= nil then\n"
+"    if parsed ~= nil then\n"
 "        return urilib.format(parsed, false)\n"
 "    end\n"
 "    return uri\n"
 "end\n"
 "\n"
-"local function purge_password_from_uris(uri)\n"
+"local function purge_sensitive_data_from_uris(uri)\n"
 "    if uri == nil then\n"
 "        return nil\n"
 "    end\n"
 "    if type(uri) == 'table' then\n"
+"        uri = purge_sensitive_data_from_uri(uri)\n"
+"        if type(uri) == 'string' then\n"
+"            return uri\n"
+"        end\n"
 "        local new_table = {}\n"
 "        for k, v in pairs(uri) do\n"
-"            new_table[k] = purge_password_from_uri(v)\n"
+"            new_table[k] = purge_sensitive_data_from_uris(v)\n"
 "        end\n"
 "        return new_table\n"
 "    end\n"
-"    return purge_password_from_uri(uri)\n"
+"    return purge_sensitive_data_from_uri(uri)\n"
+"end\n"
+"\n"
+"local function purge_sensitive_data_from_listen_options_uri(uri)\n"
+"    if uri == nil or type(uri) == \"table\" then\n"
+"        return purge_sensitive_data_from_uris(uri)\n"
+"    end\n"
+"\n"
+"    local parts = {}\n"
+"    for sub_uri in string.gmatch(uri, \"[^, ]+\") do\n"
+"        table.insert(parts, purge_sensitive_data_from_uri(sub_uri))\n"
+"    end\n"
+"    return table.concat(parts, \", \")\n"
 "end\n"
 "\n"
 "-- options that require modification for logging\n"
 "local log_cfg_option = {\n"
-"    replication = purge_password_from_uris,\n"
+"    listen = purge_sensitive_data_from_listen_options_uri,\n"
+"    replication = purge_sensitive_data_from_uris,\n"
+"    bootstrap_leader = purge_sensitive_data_from_uris,\n"
 "}\n"
 "\n"
 "\n"
@@ -468,6 +494,7 @@ const char load_cfg_lua[] =
 "    -- do nothing, affects new replicas, which query this value on start\n"
 "    wal_dir_rescan_delay    = nop,\n"
 "    wal_cleanup_delay       = private.cfg_set_wal_cleanup_delay,\n"
+"    experimental_wal_retention_period = private.cfg_set_wal_retention_period,\n"
 "    custom_proc_title       = function()\n"
 "        require('title').update(box.cfg.custom_proc_title)\n"
 "    end,\n"
@@ -643,6 +670,7 @@ const char load_cfg_lua[] =
 "    replication_anon        = true,\n"
 "    bootstrap_strategy      = true,\n"
 "    wal_dir_rescan_delay    = true,\n"
+"    wal_queue_max_size      = true,\n"
 "    custom_proc_title       = true,\n"
 "    force_recovery          = true,\n"
 "    instance_uuid           = true,\n"
@@ -653,6 +681,7 @@ const char load_cfg_lua[] =
 "    auth_delay              = ifdef_security(true),\n"
 "    disable_guest           = ifdef_security(true),\n"
 "    password_lifetime_days  = ifdef_security(true),\n"
+"    experimental_wal_retention_period = ifdef_wal_retention_period(true),\n"
 "}\n"
 "\n"
 "-- Options that are not part of dynamic_cfg_modules and applied individually\n"
@@ -772,11 +801,10 @@ const char load_cfg_lua[] =
 "    end\n"
 "end\n"
 "\n"
-"local function prepare_cfg(cfg, default_cfg, template_cfg, modify_cfg)\n"
+"local function prepare_cfg(cfg, old_cfg, default_cfg, template_cfg, modify_cfg)\n"
 "    if cfg == nil then\n"
-"        return {}\n"
-"    end\n"
-"    if type(cfg) ~= 'table' then\n"
+"        cfg = {}\n"
+"    elseif type(cfg) ~= 'table' then\n"
 "        error(\"Error: cfg should be a table\")\n"
 "    end\n"
 "    local new_cfg = {}\n"
@@ -794,6 +822,15 @@ const char load_cfg_lua[] =
 "        end\n"
 "        new_cfg[k] = v\n"
 "    end\n"
+"    -- Use the old config for omitted options.\n"
+"    for k, v in pairs(old_cfg) do\n"
+"        -- Don't override options set to box.NULL (which equals nil but\n"
+"        -- evaluates to true) because setting an option to box.NULL must\n"
+"        -- be equivalent to resetting it to the default value.\n"
+"        if cfg[k] == nil and not cfg[k] then\n"
+"            new_cfg[k] = v\n"
+"        end\n"
+"    end\n"
 "    return new_cfg\n"
 "end\n"
 "\n"
@@ -810,16 +847,6 @@ const char load_cfg_lua[] =
 "    end\n"
 "end\n"
 "\n"
-"local function merge_cfg(cfg, cur_cfg)\n"
-"    for k,v in pairs(cur_cfg) do\n"
-"        if cfg[k] == nil then\n"
-"            cfg[k] = v\n"
-"        elseif type(v) == 'table' then\n"
-"            merge_cfg(cfg[k], v)\n"
-"        end\n"
-"    end\n"
-"end\n"
-"\n"
 "-- Return true if two configurations are equivalent.\n"
 "local function compare_cfg(cfg1, cfg2)\n"
 "    if type(cfg1) ~= type(cfg2) then\n"
@@ -908,8 +935,8 @@ const char load_cfg_lua[] =
 "\n"
 "local function reload_cfg(oldcfg, cfg)\n"
 "    cfg = upgrade_cfg(cfg, translate_cfg)\n"
-"    local newcfg = prepare_cfg(cfg, default_cfg, template_cfg, modify_cfg)\n"
-"\n"
+"    local newcfg = prepare_cfg(cfg, {}, default_cfg, template_cfg,\n"
+"                               modify_cfg)\n"
 "    local module_keys = {}\n"
 "    -- iterate over original table because prepare_cfg() may store NILs\n"
 "    for key in pairs(cfg) do\n"
@@ -1005,8 +1032,7 @@ const char load_cfg_lua[] =
 "    -- Set options passed through environment variables.\n"
 "    apply_env_cfg(cfg, box.internal.cfg.env, pre_load_cfg_is_set)\n"
 "\n"
-"    cfg = prepare_cfg(cfg, default_cfg, template_cfg, modify_cfg)\n"
-"    merge_cfg(cfg, pre_load_cfg);\n"
+"    cfg = prepare_cfg(cfg, pre_load_cfg, default_cfg, template_cfg, modify_cfg)\n"
 "\n"
 "    -- Save new box.cfg\n"
 "    box.cfg = cfg\n"
@@ -1155,9 +1181,12 @@ const char load_cfg_lua[] =
 "    if param_type:find('table') and (raw_value:startswith('{') or\n"
 "                                     raw_value:startswith('[')) then\n"
 "        return json.decode(raw_value)\n"
-"    elseif param_type:find('table') and raw_value:find('=') then\n"
+"    end\n"
+"\n"
+"    if param_type:find('table') and raw_value:find('=') then\n"
 "        assert(not param_type:find('boolean'))\n"
 "        local res = {}\n"
+"        local contains_uri = false\n"
 "        for _, v in ipairs(raw_value:split(',')) do\n"
 "            local eq = v:find('=')\n"
 "            if eq == nil then\n"
@@ -1172,17 +1201,37 @@ const char load_cfg_lua[] =
 "                                         'in `key=value` or `value` format, ' ..\n"
 "                                         '`key` must not be empty'))\n"
 "            end\n"
+"            -- Don't interpret `=` as a key-value separator if\n"
+"            -- there is `\?` in a key.\n"
+"            --\n"
+"            -- Otherwise, for example,\n"
+"            -- `localhost:3301\?transport=plain` would be\n"
+"            -- interpreted as the following map.\n"
+"            --\n"
+"            -- {\n"
+"            --     ['localhost:3301\?transport'] = 'plain',\n"
+"            -- }\n"
+"            if lhs:find('\?') then\n"
+"                contains_uri = true\n"
+"            end\n"
 "            res[lhs] = tonumber(rhs) or rhs\n"
 "        end\n"
-"        return res\n"
-"    elseif param_type:find('table') and raw_value:find(',') then\n"
+"        if not contains_uri then\n"
+"            return res\n"
+"        end\n"
+"        -- Fall through otherwise.\n"
+"    end\n"
+"\n"
+"    if param_type:find('table') and raw_value:find(',') then\n"
 "        assert(not param_type:find('boolean'))\n"
 "        local res = {}\n"
 "        for i, v in ipairs(raw_value:split(',')) do\n"
 "            res[i] = tonumber(v) or v\n"
 "        end\n"
 "        return res\n"
-"    elseif param_type:find('boolean') then\n"
+"    end\n"
+"\n"
+"    if param_type:find('boolean') then\n"
 "        assert(param_type == 'boolean')\n"
 "        if raw_value:lower() == 'false' then\n"
 "            return false\n"
@@ -1190,20 +1239,24 @@ const char load_cfg_lua[] =
 "            return true\n"
 "        end\n"
 "        error(err_msg_fmt:format(env_var_name, option, '\"true\" or \"false\"'))\n"
-"    elseif param_type == 'number' then\n"
+"    end\n"
+"\n"
+"    if param_type == 'number' then\n"
 "        local res = tonumber(raw_value)\n"
 "        if res == nil then\n"
 "            error(err_msg_fmt:format(env_var_name, option,\n"
 "                'convertible to a number'))\n"
 "        end\n"
 "        return res\n"
-"    elseif param_type:find('number') then\n"
+"    end\n"
+"\n"
+"    if param_type:find('number') then\n"
 "        assert(not param_type:find('boolean'))\n"
 "        return tonumber(raw_value) or raw_value\n"
-"    else\n"
-"        assert(param_type == 'string')\n"
-"        return raw_value\n"
 "    end\n"
+"\n"
+"    assert(param_type == 'string')\n"
+"    return raw_value\n"
 "end\n"
 "\n"
 "-- Get options from env vars for given set.\n"
@@ -1228,7 +1281,6 @@ const char load_cfg_lua[] =
 "\n"
 "box.internal.prepare_cfg = prepare_cfg\n"
 "box.internal.apply_env_cfg = apply_env_cfg\n"
-"box.internal.merge_cfg = merge_cfg\n"
 "box.internal.check_cfg_option_type = check_cfg_option_type\n"
 "box.internal.update_cfg = update_cfg\n"
 "box.internal.env_cfg = env_cfg\n"
