@@ -235,6 +235,8 @@ struct netbox_transport {
 	struct error *last_error;
 	/** Fiber doing I/O and dispatching responses. */
 	struct fiber *worker;
+	/** Signalled when the worker is stopped. */
+	struct fiber_cond on_worker_stop;
 	/**
 	 * Lua reference to the Lua state used by the worker fiber or
 	 * LUA_NOREF if the worker fiber isn't running.
@@ -254,6 +256,8 @@ struct netbox_transport {
 	struct ibuf send_buf;
 	/** Connection receive buffer. */
 	struct ibuf recv_buf;
+	/** Size of the last received message. */
+	size_t last_msg_size;
 	/** Signalled when send_buf becomes empty. */
 	struct fiber_cond on_send_buf_empty;
 	/** Next request id. */
@@ -331,6 +335,12 @@ struct netbox_request {
 	 * the response hasn't been received yet.
 	 */
 	struct error *error;
+	/**
+	 * Lua reference to the connection object if the request is asynchronous
+	 * or LUA_NOREF if the request is synchronous. Used to prevent garbage
+	 * collection in case the user discards the connection.
+	 */
+	int remote_ref;
 };
 
 static const char netbox_transport_typename[] = "net.box.transport";
@@ -356,6 +366,7 @@ netbox_request_destroy(struct netbox_request *request)
 	luaL_unref(tarantool_L, LUA_REGISTRYINDEX, request->index_ref);
 	if (request->error != NULL)
 		error_unref(request->error);
+	luaL_unref(tarantool_L, LUA_REGISTRYINDEX, request->remote_ref);
 }
 
 /**
@@ -500,12 +511,14 @@ netbox_transport_create(struct netbox_transport *transport)
 	transport->is_closing = false;
 	transport->last_error = NULL;
 	transport->worker = NULL;
+	fiber_cond_create(&transport->on_worker_stop);
 	transport->coro_ref = LUA_NOREF;
 	transport->self_ref = LUA_NOREF;
 	iostream_ctx_clear(&transport->io_ctx);
 	iostream_clear(&transport->io);
 	ibuf_create(&transport->send_buf, &cord()->slabc, NETBOX_READAHEAD);
 	ibuf_create(&transport->recv_buf, &cord()->slabc, NETBOX_READAHEAD);
+	transport->last_msg_size = 0;
 	fiber_cond_create(&transport->on_send_buf_empty);
 	transport->next_sync = 1;
 	transport->requests = mh_i64ptr_new();
@@ -519,6 +532,7 @@ netbox_transport_destroy(struct netbox_transport *transport)
 	if (transport->last_error != NULL)
 		error_unref(transport->last_error);
 	assert(transport->worker == NULL);
+	fiber_cond_destroy(&transport->on_worker_stop);
 	assert(transport->coro_ref == LUA_NOREF);
 	assert(transport->self_ref == LUA_NOREF);
 	iostream_ctx_destroy(&transport->io_ctx);
@@ -563,6 +577,7 @@ netbox_transport_set_error(struct netbox_transport *transport)
 	/* Reset buffers. */
 	ibuf_reinit(&transport->send_buf);
 	ibuf_reinit(&transport->recv_buf);
+	transport->last_msg_size = 0;
 	fiber_cond_broadcast(&transport->on_send_buf_empty);
 	/* Complete requests and clean up the hash. */
 	struct mh_i64ptr_t *h = transport->requests;
@@ -1120,7 +1135,7 @@ netbox_transport_communicate(struct netbox_transport *transport, size_t limit)
 						"Peer closed");
 				return -1;
 			} if (rc > 0) {
-				recv_buf->wpos += rc;
+				VERIFY(ibuf_alloc(recv_buf, rc) != NULL);
 			} else if (rc == IOSTREAM_ERROR) {
 				goto io_error;
 			} else {
@@ -1134,7 +1149,7 @@ netbox_transport_communicate(struct netbox_transport *transport, size_t limit)
 			ssize_t rc = iostream_write(io, send_buf->rpos,
 						    ibuf_used(send_buf));
 			if (rc >= 0) {
-				send_buf->rpos += rc;
+				ibuf_consume(send_buf, rc);
 				if (ibuf_used(send_buf) == 0)
 					fiber_cond_broadcast(on_send_buf_empty);
 			} else if (rc == IOSTREAM_ERROR) {
@@ -1171,6 +1186,7 @@ static int
 netbox_transport_send_and_recv(struct netbox_transport *transport,
 			       struct xrow_header *hdr)
 {
+	ibuf_consume(&transport->recv_buf, transport->last_msg_size);
 	while (true) {
 		size_t required;
 		size_t data_len = ibuf_used(&transport->recv_buf);
@@ -1190,10 +1206,11 @@ netbox_transport_send_and_recv(struct netbox_transport *transport,
 			required = size + len;
 			if (data_len >= required) {
 				const char *body_end = rpos + len;
-				transport->recv_buf.rpos = (char *)body_end;
-				return xrow_header_decode(
-					hdr, &rpos, body_end,
-					/*end_is_exact=*/true);
+				int rc = xrow_header_decode(
+						hdr, &rpos, body_end,
+						/*end_is_exact=*/true);
+				transport->last_msg_size = body_end - bufpos;
+				return rc;
 			}
 		}
 		if (netbox_transport_communicate(transport, required) != 0)
@@ -2309,6 +2326,7 @@ luaT_netbox_transport_make_request(struct lua_State *L, int idx,
 	request->index_ref = LUA_NOREF;
 	request->result_ref = LUA_NOREF;
 	request->error = NULL;
+	request->remote_ref = LUA_NOREF;
 	netbox_request_register(request, transport);
 	return 0;
 }
@@ -2317,9 +2335,13 @@ static int
 luaT_netbox_transport_perform_async_request(struct lua_State *L)
 {
 	struct netbox_transport *transport = luaT_check_netbox_transport(L, 1);
+	/* The connection object. */
+	assert(lua_istable(L, 2));
 	struct netbox_request *request = lua_newuserdata(L, sizeof(*request));
-	if (luaT_netbox_transport_make_request(L, 2, transport, request) != 0)
+	if (luaT_netbox_transport_make_request(L, 3, transport, request) != 0)
 		return luaT_push_nil_and_error(L);
+	lua_pushvalue(L, 2);
+	request->remote_ref = luaL_ref(L, LUA_REGISTRYINDEX);
 	luaL_getmetatable(L, netbox_request_typename);
 	lua_setmetatable(L, -2);
 	return 1;
@@ -2738,21 +2760,28 @@ netbox_transport_fetch_schema(struct netbox_transport *transport,
 restart:
 	lua_newtable(L);
 	int schema_table_idx = lua_gettop(L);
+	int inprogress_request_count = 0;
 	uint64_t vspace_sync = transport->next_sync++;
 	netbox_encode_select_all(L, &transport->send_buf, vspace_sync,
 				 BOX_VSPACE_ID);
+	inprogress_request_count++;
 	uint64_t vindex_sync = transport->next_sync++;
 	netbox_encode_select_all(L, &transport->send_buf, vindex_sync,
 				 BOX_VINDEX_ID);
+	inprogress_request_count++;
 	uint64_t vcollation_sync = transport->next_sync++;
-	if (peer_has_vcollation)
+	if (peer_has_vcollation) {
 		netbox_encode_select_all(L, &transport->send_buf,
 					 vcollation_sync, BOX_VCOLLATION_ID);
+		inprogress_request_count++;
+	}
 	uint64_t vspace_sequence_sync = transport->next_sync++;
-	if (peer_has_vspace_sequence)
+	if (peer_has_vspace_sequence) {
 		netbox_encode_select_all(L, &transport->send_buf,
 					 vspace_sequence_sync,
 					 BOX_VSPACE_SEQUENCE_ID);
+		inprogress_request_count++;
+	}
 	bool got_vspace = false;
 	bool got_vindex = false;
 	bool got_vcollation = false;
@@ -2769,6 +2798,8 @@ restart:
 			netbox_transport_dispatch_response(transport, L, &hdr);
 			continue;
 		}
+		assert(inprogress_request_count > 0);
+		inprogress_request_count--;
 		if (iproto_type_is_error(hdr.type)) {
 			uint32_t errcode = hdr.type & (IPROTO_TYPE_ERROR - 1);
 			if (errcode == ER_NO_SUCH_SPACE) {
@@ -2792,6 +2823,13 @@ restart:
 			 * Restart loader.
 			 */
 			lua_pop(L, 1);
+			/*
+			 * Account the pending requests so that the in-progress
+			 * request counter doesn't go negative when we finally
+			 * get the replies.
+			 */
+			transport->inprogress_request_count +=
+					inprogress_request_count;
 			goto restart;
 		}
 		const char *data = hdr.body[0].iov_base;
@@ -2818,6 +2856,7 @@ restart:
 	} while (!(got_vspace && got_vindex &&
 		   (got_vcollation || !peer_has_vcollation) &&
 		   (got_vspace_sequence || !peer_has_vspace_sequence)));
+	assert(inprogress_request_count == 0);
 	/* Invoke the 'did_fetch_schema' callback. */
 	lua_rawgeti(L, LUA_REGISTRYINDEX, transport->opts.callback_ref);
 	lua_pushliteral(L, "did_fetch_schema");
@@ -2885,18 +2924,12 @@ netbox_connection_handler_f(struct lua_State *L)
 static int
 netbox_worker_f(va_list ap)
 {
-	struct netbox_transport *transport;
-	transport = va_arg(ap, struct netbox_transport *);
-	struct lua_State *L = va_arg(ap, struct lua_State *);
+	(void)ap;
+	struct netbox_transport *transport = fiber()->f_arg;
+	struct lua_State *L = fiber()->storage.lua.stack;
 	assert(transport->worker == fiber());
 	assert(transport->coro_ref != LUA_NOREF);
 	assert(transport->self_ref != LUA_NOREF);
-	/*
-	 * Code that needs a temporary fiber-local Lua state may save some time
-	 * and resources for creating a new state and use this one.
-	 */
-	assert(fiber()->storage.lua.stack == NULL);
-	fiber()->storage.lua.stack = L;
 	const double reconnect_after = transport->opts.reconnect_after;
 	while (!fiber_is_cancelled()) {
 		if (netbox_transport_connect(transport) == 0) {
@@ -2907,7 +2940,7 @@ netbox_worker_f(va_list ap)
 			(void)rc;
 			iostream_close(&transport->io);
 		}
-		if (transport->state == NETBOX_CLOSED)
+		if (fiber_is_cancelled())
 			break;
 		netbox_transport_set_error(transport);
 		transport->state = (reconnect_after > 0 ?
@@ -2919,7 +2952,14 @@ netbox_worker_f(va_list ap)
 			break;
 		}
 	}
+	if (transport->state != NETBOX_ERROR) {
+		box_error_raise(ER_NO_CONNECTION, "Connection closed");
+		netbox_transport_set_error(transport);
+		transport->state = NETBOX_CLOSED;
+		netbox_transport_on_state_change_pcall(transport, L);
+	}
 	transport->worker = NULL;
+	fiber_cond_broadcast(&transport->on_worker_stop);
 	luaL_unref(tarantool_L, LUA_REGISTRYINDEX, transport->coro_ref);
 	transport->coro_ref = LUA_NOREF;
 	/* Careful: luaL_unref may delete this transport object. */
@@ -2954,7 +2994,14 @@ luaT_netbox_transport_start(struct lua_State *L)
 		transport->self_ref = LUA_NOREF;
 		return luaT_error(L);
 	}
-	fiber_start(transport->worker, transport, fiber_L);
+	transport->worker->f_arg = transport;
+	/*
+	 * Code that needs a temporary fiber-local Lua state may save some time
+	 * and resources for creating a new state and use this one.
+	 */
+	assert(transport->worker->storage.lua.stack == NULL);
+	transport->worker->storage.lua.stack = fiber_L;
+	fiber_wakeup(transport->worker);
 	return 0;
 }
 
@@ -2962,17 +3009,16 @@ luaT_netbox_transport_start(struct lua_State *L)
  * Stops a worker fiber.
  *
  * Takes an optional boolean argument 'wait': if set the function will wait
- * for all pending requests to be sent.
+ * for all pending requests to be sent and for the connection to be closed.
+ * It is ignored if `stop` is being called from the worker fiber.
  */
 static int
 luaT_netbox_transport_stop(struct lua_State *L)
 {
 	struct netbox_transport *transport = luaT_check_netbox_transport(L, 1);
-	bool wait = lua_toboolean(L, 2);
-	if (wait && fiber() != transport->worker &&
-	    transport->state != NETBOX_CLOSED &&
-	    transport->state != NETBOX_ERROR) {
-		transport->is_closing = true;
+	transport->is_closing = true;
+	bool wait = fiber() == transport->worker ? false : lua_toboolean(L, 2);
+	if (wait) {
 		/*
 		 * Here we are waiting until send buf became empty:
 		 * it is necessary to ensure that all requests are
@@ -2980,18 +3026,6 @@ luaT_netbox_transport_stop(struct lua_State *L)
 		 */
 		while (ibuf_used(&transport->send_buf) > 0)
 			fiber_cond_wait(&transport->on_send_buf_empty);
-		transport->is_closing = false;
-	}
-	/*
-	 * While we were waiting for the send buffer to be empty,
-	 * the state could change.
-	 */
-	if (transport->state != NETBOX_CLOSED &&
-	    transport->state != NETBOX_ERROR) {
-		box_error_raise(ER_NO_CONNECTION, "Connection closed");
-		netbox_transport_set_error(transport);
-		transport->state = NETBOX_CLOSED;
-		netbox_transport_on_state_change(transport, L);
 	}
 	/* Cancel the worker fiber. */
 	if (transport->worker != NULL) {
@@ -2999,6 +3033,10 @@ luaT_netbox_transport_stop(struct lua_State *L)
 		/* Check if we cancelled ourselves. */
 		luaL_testcancel(L);
 	}
+	if (wait) {
+		while (transport->worker != NULL)
+			fiber_cond_wait(&transport->on_worker_stop);
+	}
 	return 0;
 }
 
