@@ -7,6 +7,7 @@ const char net_box_lua[] =
 "local urilib   = require('uri')\n"
 "local internal = require('net.box.lib')\n"
 "local trigger  = require('internal.trigger')\n"
+"local utils    = require('internal.utils')\n"
 "\n"
 "local this_module\n"
 "\n"
@@ -17,7 +18,7 @@ const char net_box_lua[] =
 "local check_index_arg     = box.internal.check_index_arg\n"
 "local check_space_arg     = box.internal.check_space_arg\n"
 "local check_primary_index = box.internal.check_primary_index\n"
-"local check_param_table   = box.internal.check_param_table\n"
+"local check_param_table   = utils.check_param_table\n"
 "\n"
 "local ibuf_t = ffi.typeof('struct ibuf')\n"
 "local is_tuple = box.tuple.is\n"
@@ -324,6 +325,7 @@ const char net_box_lua[] =
 "        end\n"
 "        if what == 'state_changed' then\n"
 "            local state, err = ...\n"
+"            remote.state, remote.error = state, err\n"
 "            local was_connected = remote._is_connected\n"
 "            if state == 'active' then\n"
 "                if not was_connected then\n"
@@ -334,7 +336,11 @@ const char net_box_lua[] =
 "                   state == 'closed' then\n"
 "                if was_connected then\n"
 "                    remote._is_connected = false\n"
-"                    remote._on_disconnect:run(remote)\n"
+"                    local ok, trigger_err = pcall(remote._on_disconnect.run,\n"
+"                                                  remote._on_disconnect, remote)\n"
+"                    if not ok then\n"
+"                        log.error(trigger_err)\n"
+"                    end\n"
 "                end\n"
 "                -- A server may exit after initiating a graceful shutdown but\n"
 "                -- before all clients close their connections (for example, on\n"
@@ -344,7 +350,6 @@ const char net_box_lua[] =
 "                -- time we finish running on_shutdown triggers.\n"
 "                remote._shutdown_pending = nil\n"
 "            end\n"
-"            remote.state, remote.error = state, err\n"
 "            if state == 'error_reconnect' then\n"
 "                -- Repeat the same error in verbose log only.\n"
 "                -- Else the error clogs the log. See gh-3175.\n"
@@ -775,7 +780,7 @@ const char net_box_lua[] =
 "            if opts.on_push or opts.on_push_ctx then\n"
 "                error('To handle pushes in an async request use future:pairs()')\n"
 "            end\n"
-"            return transport:perform_async_request(buffer, skip_header,\n"
+"            return transport:perform_async_request(self, buffer, skip_header,\n"
 "                                                   return_raw, table.insert,\n"
 "                                                   {}, format, stream_id,\n"
 "                                                   method, ...)\n"
@@ -1055,7 +1060,11 @@ const char net_box_lua[] =
 "\n"
 "    self.schema_version = schema_version\n"
 "    self.space = sl\n"
-"    self._on_schema_reload:run(self)\n"
+"    local ok, err = pcall(self._on_schema_reload.run, self._on_schema_reload,\n"
+"                          self)\n"
+"    if not ok then\n"
+"        log.error(err)\n"
+"    end\n"
 "end\n"
 "\n"
 "local function nothing_or_data(value)\n"
@@ -1296,6 +1305,12 @@ const char net_box_lua[] =
 "        check_call_args(args)\n"
 "        args = args or {}\n"
 "        proc_name = tostring(proc_name)\n"
+"        if box.ctl.is_recovery_finished() then\n"
+"            local f = box.func[proc_name]\n"
+"            if f ~= nil then\n"
+"                return handle_eval_result(pcall(f.call, f, args))\n"
+"            end\n"
+"        end\n"
 "        local status, proc, obj = pcall(box.internal.call_loadproc, proc_name)\n"
 "        if not status then\n"
 "            rollback()\n"
