@@ -6,8 +6,15 @@ const char upgrade_lua[] =
 "local xlog = require('xlog')\n"
 "local ffi = require('ffi')\n"
 "local fun = require('fun')\n"
+"local tarantool = require('tarantool')\n"
+"local utils = require('internal.utils')\n"
 "\n"
-"ffi.cdef('uint32_t box_dd_version_id(void);')\n"
+"ffi.cdef([[\n"
+"    uint32_t box_dd_version_id(void);\n"
+"    int box_schema_upgrade_begin(void);\n"
+"    void box_schema_upgrade_end(void);\n"
+"]])\n"
+"local builtin = ffi.C\n"
 "\n"
 "-- Guest user id - the default user\n"
 "local GUEST = 0\n"
@@ -289,7 +296,7 @@ const char upgrade_lua[] =
 "end\n"
 "\n"
 "local function user_trig_1_7_5(_, tuple)\n"
-"    if tuple and not tuple[5] then\n"
+"    if tuple and (type(tuple[5]) ~= 'table' or next(tuple[5]) == nil) then\n"
 "        tuple = tuple:update{{'=', 5, setmap({})}}\n"
 "        log.info(\"Set empty password to %s '%s'\", tuple[4], tuple[3])\n"
 "    end\n"
@@ -576,6 +583,58 @@ const char upgrade_lua[] =
 "-- Tarantool 1.7.6\n"
 "--------------------------------------------------------------------------------\n"
 "\n"
+"--\n"
+"-- Update format of user spaces. Format of system spaces have been updated\n"
+"-- during upgrade to 1.7.5. However commit 519bc82e (\"Parse and validate\n"
+"-- space formats\") introduced strict checking of format field. Now tarantool\n"
+"-- requires, that format is written as map with explicit 'name' and 'type'\n"
+"-- keys: {name = 'a', type = 'number'}. Previously, it wasn't a case,\n"
+"-- anything could be written in the space format.\n"
+"--\n"
+"local function space_trig_1_7_6(_, tuple)\n"
+"    -- Do nothing for system spaces.\n"
+"    if tuple == nil or tuple[1] <= box.schema.SYSTEM_ID_MAX then\n"
+"        return tuple\n"
+"    end\n"
+"\n"
+"    -- If it's a complete garbage or tuple[7] is missing, use empty format.\n"
+"    -- Also, overwrite the table, if it's empty to be sure, that it's\n"
+"    -- encoded as array and not as map.\n"
+"    local new_format = {}\n"
+"    if type(tuple[7]) == 'table' and next(tuple[7]) ~= nil then\n"
+"        -- Non empty format. Check format and change if it's possible.\n"
+"        for i, f in ipairs(tuple[7]) do\n"
+"            -- Create the fields from ground up. Better not to use existing\n"
+"            -- ones, anything can be there, we must be sure, that every field\n"
+"            -- is encoded as map and has 'name' and 'type' keys.\n"
+"            local field = setmap({\n"
+"                name = f['name'],\n"
+"                type = f['type'],\n"
+"            })\n"
+"\n"
+"            if field['name'] == nil then\n"
+"                if type(f[1]) ~= 'string' then\n"
+"                    local err = \"Cannot find name for field %d in space %s\"\n"
+"                    error(err:format(i, tuple[3]))\n"
+"                end\n"
+"                field['name'] = f[1]\n"
+"            end\n"
+"            if field['type'] == nil then\n"
+"                field['type'] = type(f[2]) == 'string' and f[2] or 'any'\n"
+"            end\n"
+"            table.insert(new_format, field)\n"
+"        end\n"
+"    end\n"
+"    if not table.equals(tuple[7], new_format) then\n"
+"        -- Better to be as verbose as it's possible. So that user can\n"
+"        -- restore the previous format if smth goes wrong.\n"
+"        log.info(\"Update space '%s' format: old format %s, new format %s\",\n"
+"                 tuple[3], json.encode(tuple[7]), json.encode(new_format))\n"
+"        tuple = tuple:update{{'=', 7, new_format}}\n"
+"    end\n"
+"    return tuple\n"
+"end\n"
+"\n"
 "local function create_sequence_space()\n"
 "    local _space = box.space[box.schema.SPACE_ID]\n"
 "    local _index = box.space[box.schema.INDEX_ID]\n"
@@ -1400,7 +1459,15 @@ const char upgrade_lua[] =
 "local function upgrade_to_2_11_1()\n"
 "    drop_schema_max_id()\n"
 "end\n"
+"\n"
 "--------------------------------------------------------------------------------\n"
+"-- Tarantool 2.11.5\n"
+"--------------------------------------------------------------------------------\n"
+"local function upgrade_to_2_11_5()\n"
+"    -- NoOp. We need to bump schema version to distinguish, where persistent\n"
+"    -- names properly work before box.schema.upgrade is called.\n"
+"    return\n"
+"end\n"
 "\n"
 "local handlers = {\n"
 "    {version = mkversion(1, 7, 5), func = upgrade_to_1_7_5},\n"
@@ -1422,11 +1489,12 @@ const char upgrade_lua[] =
 "    {version = mkversion(2, 10, 5), func = upgrade_to_2_10_5},\n"
 "    {version = mkversion(2, 11, 0), func = upgrade_to_2_11_0},\n"
 "    {version = mkversion(2, 11, 1), func = upgrade_to_2_11_1},\n"
+"    {version = mkversion(2, 11, 5), func = upgrade_to_2_11_5},\n"
 "}\n"
 "\n"
 "-- Schema version of the snapshot.\n"
 "local function get_version()\n"
-"    local version = ffi.C.box_dd_version_id()\n"
+"    local version = builtin.box_dd_version_id()\n"
 "    local major = bit.band(bit.rshift(version, 16), 0xff)\n"
 "    local minor = bit.band(bit.rshift(version, 8), 0xff)\n"
 "    local patch = bit.band(version, 0xff)\n"
@@ -1476,6 +1544,9 @@ const char upgrade_lua[] =
 "        _space = space_trig_1_7_5,\n"
 "        _user  = user_trig_1_7_5,\n"
 "    }},\n"
+"    {version = mkversion(1, 7, 6), tbl = {\n"
+"        _space = space_trig_1_7_6,\n"
+"    }},\n"
 "    {version = mkversion(1, 7, 7), tbl = {\n"
 "        _priv   = priv_trig_1_7_7,\n"
 "    }},\n"
@@ -1549,8 +1620,22 @@ const char upgrade_lua[] =
 "    end\n"
 "end\n"
 "\n"
+"-- Runs the given function with the permission to execute DDL operations\n"
+"-- with an old schema. Used by upgrade/downgrade scripts.\n"
+"local function run_upgrade(func, ...)\n"
+"    if builtin.box_schema_upgrade_begin() ~= 0 then\n"
+"        box.error()\n"
+"    end\n"
+"    local ok, err = pcall(func, ...)\n"
+"    builtin.box_schema_upgrade_end()\n"
+"    if not ok then\n"
+"        error(err)\n"
+"    end\n"
+"end\n"
+"\n"
 "local function upgrade()\n"
-"    upgrade_from(get_version())\n"
+"    local version = get_version()\n"
+"    run_upgrade(upgrade_from, version)\n"
 "end\n"
 "\n"
 "--------------------------------------------------------------------------------\n"
@@ -1582,20 +1667,15 @@ const char upgrade_lua[] =
 "    }\n"
 "    local datetime = os.date(\"%Y-%m-%d %H:%M:%S\")\n"
 "    local _func = box.space._func\n"
-"    -- Otherwise we can't insert SQL_BUILTIN function. It is prohibited\n"
-"    -- to add since 2.9.0.\n"
-"    with_disabled_system_triggers(function()\n"
-"        for _, func in ipairs(sql_builtin_list) do\n"
-"            if _func.index.name:get(func) == nil then\n"
-"                local t = _func:auto_increment{\n"
-"                    ADMIN, func, 1, 'SQL_BUILTIN', '', 'function', {}, 'any',\n"
-"                    'none', 'none', false, false, true, {}, setmap({}), '',\n"
-"                    datetime, datetime}\n"
-"                box.space._priv:replace{ADMIN, PUBLIC, 'function', t.id,\n"
-"                                        box.priv.X}\n"
-"            end\n"
+"    for _, func in ipairs(sql_builtin_list) do\n"
+"        if _func.index.name:get(func) == nil then\n"
+"            local t = _func:auto_increment{\n"
+"                ADMIN, func, 1, 'SQL_BUILTIN', '', 'function', {}, 'any',\n"
+"                'none', 'none', false, false, true, {}, setmap({}), '',\n"
+"                datetime, datetime}\n"
+"            box.space._priv:replace{ADMIN, PUBLIC, 'function', t.id, box.priv.X}\n"
 "        end\n"
-"    end)\n"
+"    end\n"
 "end\n"
 "\n"
 "local function downgrade_from_2_9_1(issue_handler)\n"
@@ -1765,16 +1845,13 @@ const char upgrade_lua[] =
 "    end\n"
 "    log.info(\"revoke grants for 'public' role for _vspace_sequence\")\n"
 "    box.space._priv:delete{PUBLIC, 'space', box.schema.VSPACE_SEQUENCE_ID}\n"
-"    local indexes = box.space._index:select(box.schema.VSPACE_SEQUENCE_ID)\n"
-"    -- Otherwise we can't drop neither the primary index nor the space.\n"
-"    with_disabled_system_triggers(function()\n"
-"        for _, index in pairs(indexes) do\n"
-"            log.info(\"drop index %s on _vspace_sequence\", index[3])\n"
-"            box.space._index:delete{index[1], index[2]}\n"
-"        end\n"
-"        log.info(\"drop view _vspace_sequence\")\n"
-"        box.space._space:delete{box.schema.VSPACE_SEQUENCE_ID}\n"
-"    end)\n"
+"    for _, index in box.space._index:pairs(box.schema.VSPACE_SEQUENCE_ID,\n"
+"                                           {iterator = 'REQ'}) do\n"
+"        log.info(\"drop index %s on _vspace_sequence\", index[3])\n"
+"        box.space._index:delete{index[1], index[2]}\n"
+"    end\n"
+"    log.info(\"drop view _vspace_sequence\")\n"
+"    box.space._space:delete{box.schema.VSPACE_SEQUENCE_ID}\n"
 "end\n"
 "\n"
 "local function downgrade_from_2_10_5(issue_handler)\n"
@@ -2001,6 +2078,7 @@ const char upgrade_lua[] =
 "\n"
 "-- List of all Tarantool releases we can downgrade to.\n"
 "local downgrade_versions = {\n"
+"    -- DOWNGRADE VERSIONS BEGIN\n"
 "    \"2.8.2\",\n"
 "    \"2.8.3\",\n"
 "    \"2.8.4\",\n"
@@ -2010,8 +2088,19 @@ const char upgrade_lua[] =
 "    \"2.10.3\",\n"
 "    \"2.10.4\",\n"
 "    \"2.10.5\",\n"
+"    \"2.10.6\",\n"
+"    \"2.10.7\",\n"
+"    \"2.10.8\",\n"
 "    \"2.11.0\",\n"
 "    \"2.11.1\",\n"
+"    \"2.11.2\",\n"
+"    \"2.11.3\",\n"
+"    \"2.11.4\",\n"
+"    \"2.11.5\",\n"
+"    \"2.11.6\",\n"
+"    \"2.11.7\",\n"
+"    \"2.11.8\",\n"
+"    -- DOWNGRADE VERSIONS END\n"
 "}\n"
 "\n"
 "-- Downgrade or list downgrade issues depending of dry_run argument value.\n"
@@ -2022,7 +2111,7 @@ const char upgrade_lua[] =
 "-- In case of downgrade check for issues is done before making any changes.\n"
 "-- If any issue is found then downgrade is failed and no any changes are done.\n"
 "local function downgrade_impl(version_str, dry_run)\n"
-"    box.internal.check_param(version_str, 'version_str', 'string')\n"
+"    utils.check_param(version_str, 'version_str', 'string')\n"
 "    local version = mkversion.parse(version_str)\n"
 "    if fun.index(version_str, downgrade_versions) == nil then\n"
 "        error(\"Downgrade is only possible to version listed in\" ..\n"
@@ -2030,6 +2119,12 @@ const char upgrade_lua[] =
 "    end\n"
 "\n"
 "    local schema_version_cur = get_version()\n"
+"    local app_version = tarantool.version:match('^%d+%.%d+%.%d+')\n"
+"    if schema_version_cur > mkversion.parse(app_version) then\n"
+"        local err = \"Cannot downgrade as current schema version %s is newer\" ..\n"
+"                    \" than Tarantool version %s\"\n"
+"        error(err:format(schema_version_cur, app_version))\n"
+"    end\n"
 "    local schema_version_dst = app2schema_version(version)\n"
 "    if schema_version_cur < schema_version_dst then\n"
 "        local err = \"Cannot downgrade as current schema version %s is older\" ..\n"
@@ -2088,7 +2183,9 @@ const char upgrade_lua[] =
 "box.schema.downgrade_versions = function()\n"
 "    return table.copy(downgrade_versions)\n"
 "end\n"
-"box.schema.downgrade = function(version) downgrade_impl(version, false) end\n"
+"box.schema.downgrade = function(version)\n"
+"    run_upgrade(downgrade_impl, version, false)\n"
+"end\n"
 "box.schema.downgrade_issues = function(version)\n"
 "    return downgrade_impl(version, true)\n"
 "end\n"
@@ -2097,5 +2194,9 @@ const char upgrade_lua[] =
 "box.internal.get_snapshot_version = get_snapshot_version;\n"
 "box.internal.set_recovery_triggers = set_recovery_triggers;\n"
 "box.internal.clear_recovery_triggers = clear_recovery_triggers;\n"
+"\n"
+"-- Export the run_upgrade() helper to let users perform schema upgrade\n"
+"-- manually in case box.schema.upgrade() failed.\n"
+"box.internal.run_schema_upgrade = run_upgrade\n"
 ""
 ;
