@@ -371,15 +371,13 @@ memtx_hash_index_replace(struct index *base, struct tuple *old_tuple,
 		if (pos == light_index_end)
 			pos = light_index_insert(hash_table, h, new_tuple);
 
-		ERROR_INJECT(ERRINJ_INDEX_ALLOC,
-		{
+		ERROR_INJECT(ERRINJ_HASH_INDEX_REPLACE, {
 			light_index_delete(hash_table, pos);
 			pos = light_index_end;
 		});
 
 		if (pos == light_index_end) {
-			diag_set(OutOfMemory,
-				 (ssize_t)light_index_count(hash_table),
+			diag_set(OutOfMemory, MEMTX_EXTENT_SIZE,
 				 "hash_table", "key");
 			return -1;
 		}
@@ -644,7 +642,7 @@ memtx_hash_index_create_read_view(struct index *base)
 	index_read_view_create(&rv->base, &vtab, base->def);
 	struct space *space = space_by_id(base->def->space_id);
 	assert(space != NULL);
-	memtx_tx_snapshot_cleaner_create(&rv->cleaner, space);
+	memtx_tx_snapshot_cleaner_create(&rv->cleaner, space, base);
 	rv->index = index;
 	index_ref(base);
 	light_index_view_create(&rv->view, &index->hash_table);
@@ -690,9 +688,8 @@ memtx_hash_index_new(struct memtx_engine *memtx, struct index_def *def)
 	index_create(&index->base, (struct engine *)memtx,
 		     &memtx_hash_index_vtab, def);
 
-	light_index_create(&index->hash_table, MEMTX_EXTENT_SIZE,
-			   memtx_index_extent_alloc, memtx_index_extent_free,
-			   memtx, index->base.def->key_def);
+	light_index_create(&index->hash_table, index->base.def->key_def,
+			   &memtx->index_extent_allocator);
 	return &index->base;
 }
 
