@@ -844,6 +844,21 @@ memtx_space_check_index_def(struct space *space, struct index_def *index_def)
 			return -1;
 		}
 	}
+
+	/* Checks for memtx MVCC unsupported features. */
+	if (memtx_tx_manager_use_mvcc_engine) {
+		if (key_def->is_multikey) {
+			diag_set(ClientError, ER_UNSUPPORTED,
+				 "Memtx MVCC engine", "multikey indexes");
+			return -1;
+		}
+		if (key_def->for_func_index) {
+			diag_set(ClientError, ER_UNSUPPORTED,
+				 "Memtx MVCC engine", "functional indexes");
+			return -1;
+		}
+	}
+
 	return 0;
 }
 
@@ -942,6 +957,18 @@ struct memtx_ddl_state {
 	struct tuple *cursor;
 	/* Primary key key_def to compare new tuples with cursor. */
 	struct key_def *cmp_def;
+	/*
+	 * List of all transactional triggers created by the DDL for concurrent
+	 * statements. They must be destroyed when the DDL is over since the
+	 * object is going to be invalidated. It is OK because all owners of
+	 * the triggers are already prepared and on their rollback the DDL
+	 * will be rolled back as well and on their commit we won't need the
+	 * triggers anymore.
+	 *
+	 * NB: must be used only without MVCC, otherwise something must have
+	 * gone wrong.
+	 */
+	struct rlist stmt_triggers;
 	struct diag diag;
 	int rc;
 };
@@ -1000,6 +1027,7 @@ memtx_space_check_format(struct space *space, struct tuple_format *format)
 	state.cmp_def = pk->def->key_def;
 	state.rc = 0;
 	diag_create(&state.diag);
+	rlist_create(&state.stmt_triggers);
 
 	struct trigger on_replace;
 	trigger_create(&on_replace, memtx_check_on_replace, &state, NULL);
@@ -1037,6 +1065,7 @@ memtx_space_check_format(struct space *space, struct tuple_format *format)
 	iterator_delete(it);
 	diag_destroy(&state.diag);
 	trigger_clear(&on_replace);
+	assert(rlist_empty(&state.stmt_triggers));
 	return rc;
 }
 
@@ -1052,7 +1081,6 @@ memtx_space_drop_primary_key(struct space *space)
 	 *   can be put back online properly.
 	 */
 	memtx_space->replace = memtx_space_replace_no_keys;
-	memtx_space->bsize = 0;
 }
 
 static void
@@ -1069,21 +1097,35 @@ memtx_init_ephemeral_space(struct space *space)
 }
 
 /*
- * Ongoing index build state with statement used by
- * corresponding on_rollback triggers to prevent rollbacked changes appearance.
+ * Struct to allocate by memtx_build_on_replace trigger, contains
+ * transactional statement triggers with their data.
  */
-struct index_build_on_rollback_data {
+struct memtx_build_stmt_trigger {
+	/** Link in memtx_ddl_state::stmt_triggers. */
+	struct rlist in_state;
+	/** See memtx_build_on_replace_rollback. */
+	struct trigger on_rollback;
+	/** See memtx_build_on_replace_commit. */
+	struct trigger on_commit;
+	/** State of the DDL. */
 	struct memtx_ddl_state *state;
-	struct txn_stmt *stmt;
 };
 
+/**
+ * Rolls back a replace in the index that was done during its
+ * background build.
+ */
 static int
-memtx_build_on_replace_rollback(struct trigger *trigger, void *event)
+memtx_build_on_replace_rollback(struct trigger *base, void *event)
 {
-	(void)event;
-	struct index_build_on_rollback_data *data = trigger->data;
-	struct txn_stmt *stmt = data->stmt;
-	struct memtx_ddl_state *state = data->state;
+	struct txn_stmt *stmt = (struct txn_stmt *)event;
+	struct memtx_build_stmt_trigger *trigger =
+		container_of(base, struct memtx_build_stmt_trigger,
+			     on_rollback);
+	struct memtx_ddl_state *state = trigger->state;
+
+	/* The trigger is fired and will be deleted - remove from the state. */
+	rlist_del_entry(trigger, in_state);
 	/*
 	 * Old tuple's format is valid if it exists.
 	 */
@@ -1119,14 +1161,19 @@ memtx_build_on_replace_rollback(struct trigger *trigger, void *event)
 	return 0;
 }
 
-/*
- * Struct to allocate by memtx_build_on_replace trigger
- * (on_rollback trigger with its data).
+/**
+ * Deletes the transactional statement triggers from
+ * memtx_ddl_state::stmt_triggers since they are going to be deleted.
  */
-struct on_rollback_trigger_with_data {
-	struct trigger on_rollback;
-	struct index_build_on_rollback_data data;
-};
+static int
+memtx_build_on_replace_commit(struct trigger *base, void *event)
+{
+	(void)event;
+	struct memtx_build_stmt_trigger *trigger =
+		container_of(base, struct memtx_build_stmt_trigger, on_commit);
+	rlist_del_entry(trigger, in_state);
+	return 0;
+}
 
 static int
 memtx_build_on_replace(struct trigger *trigger, void *event)
@@ -1179,30 +1226,18 @@ memtx_build_on_replace(struct trigger *trigger, void *event)
 	 * problem when rollbacked changes appears in
 	 * built-in-background index.
 	 */
-	struct on_rollback_trigger_with_data *on_rollback_associates = NULL;
-	struct errinj *inj = errinj(ERRINJ_BUILD_INDEX_ON_ROLLBACK_ALLOC,
-				    ERRINJ_BOOL);
-	if (inj == NULL || inj->bparam == false) {
-		on_rollback_associates = region_aligned_alloc(
-			&in_txn()->region,
-			sizeof(struct on_rollback_trigger_with_data),
-			alignof(struct on_rollback_trigger_with_data));
-	}
-	if (on_rollback_associates == NULL) {
-		diag_set(OutOfMemory,
-			 sizeof(struct on_rollback_trigger_with_data),
-			 "region_aligned_alloc",
-			 "struct on_rollback_trigger_with_data");
-		diag_move(diag_get(), &state->diag);
-		state->rc = -1;
-		return 0;
-	}
-	on_rollback_associates->data.stmt = stmt;
-	on_rollback_associates->data.state = state;
-	trigger_create(&on_rollback_associates->on_rollback,
-		       memtx_build_on_replace_rollback,
-		       &on_rollback_associates->data, NULL);
-	txn_stmt_on_rollback(stmt, &on_rollback_associates->on_rollback);
+	struct memtx_build_stmt_trigger *stmt_trigger =
+		xregion_alloc_object(&in_txn()->region,
+				     struct memtx_build_stmt_trigger);
+	rlist_add_entry(&state->stmt_triggers, stmt_trigger, in_state);
+	stmt_trigger->state = state;
+
+	trigger_create(&stmt_trigger->on_rollback,
+		       memtx_build_on_replace_rollback, NULL, NULL);
+	trigger_create(&stmt_trigger->on_commit,
+		       memtx_build_on_replace_commit, NULL, NULL);
+	txn_stmt_on_rollback(stmt, &stmt_trigger->on_rollback);
+	txn_stmt_on_commit(stmt, &stmt_trigger->on_commit);
 	return 0;
 }
 
@@ -1253,6 +1288,10 @@ memtx_space_build_index(struct space *src_space, struct index *new_index,
 	 */
 	bool can_yield = pk->def->type != HASH;
 
+	inj = errinj(ERRINJ_BUILD_INDEX_DISABLE_YIELD, ERRINJ_BOOL);
+	if (inj != NULL && inj->bparam == true)
+		can_yield = false;
+
 	struct memtx_engine *memtx = (struct memtx_engine *)src_space->engine;
 	struct memtx_ddl_state state;
 	struct trigger on_replace;
@@ -1266,10 +1305,20 @@ memtx_space_build_index(struct space *src_space, struct index *new_index,
 		state.cmp_def = pk->def->key_def;
 		state.rc = 0;
 		diag_create(&state.diag);
+		rlist_create(&state.stmt_triggers);
 
 		trigger_create(&on_replace, memtx_build_on_replace, &state,
 			       NULL);
-		trigger_add(&src_space->on_replace, &on_replace);
+		/*
+		 * If MVCC is enabled, all concurrent writes that we haven't
+		 * read will be conflicted, so we don't need to set the trigger.
+		 * Moreover, concurrent transaction can be rolled back much
+		 * later than the index will be built so the memtx_ddl_state
+		 * will be invalid in this case since it's allocated on stack
+		 * of this function.
+		 */
+		if (!memtx_tx_manager_use_mvcc_engine)
+			trigger_add(&src_space->on_replace, &on_replace);
 	}
 
 	/*
@@ -1349,6 +1398,11 @@ memtx_space_build_index(struct space *src_space, struct index *new_index,
 	}
 	iterator_delete(it);
 	if (can_yield) {
+		struct memtx_build_stmt_trigger *trg;
+		rlist_foreach_entry(trg, &state.stmt_triggers, in_state) {
+			trigger_clear(&trg->on_rollback);
+			trigger_clear(&trg->on_commit);
+		}
 		diag_destroy(&state.diag);
 		trigger_clear(&on_replace);
 	}
@@ -1369,10 +1423,43 @@ memtx_space_prepare_alter(struct space *old_space, struct space *new_space)
 	}
 
 	new_memtx_space->replace = old_memtx_space->replace;
-	new_memtx_space->bsize = old_memtx_space->bsize;
 	return 0;
 }
 
+/**
+ * Copy bsize to the newly altered space from the old space.
+ * In case of DropIndex or TruncateIndex alter operations, the new space will be
+ * empty, and bsize must not be copied.
+ */
+static void
+memtx_space_finish_alter(struct space *old_space, struct space *new_space)
+{
+	struct memtx_space *old_memtx_space = (struct memtx_space *)old_space;
+	struct memtx_space *new_memtx_space = (struct memtx_space *)new_space;
+
+	bool is_empty = new_space->index_count == 0 ||
+			index_size(new_space->index[0]) == 0;
+	if (!is_empty)
+		new_memtx_space->bsize = old_memtx_space->bsize;
+}
+
+/**
+ * Invalidate the space in transaction manager.
+ */
+static void
+memtx_space_invalidate(struct space *space)
+{
+	/*
+	 * Abort all concurrent transactions and invalidate space in MVCC
+	 * only if it is enabled. Otherwise, all concurrent transactions
+	 * will be aborted when DDL will be prepared.
+	 */
+	if (memtx_tx_manager_use_mvcc_engine) {
+		memtx_tx_abort_all_for_ddl(in_txn());
+		memtx_tx_invalidate_space(space, in_txn());
+	}
+}
+
 /* }}} DDL */
 
 static const struct space_vtab memtx_space_vtab = {
@@ -1395,8 +1482,9 @@ static const struct space_vtab memtx_space_vtab = {
 	/* .build_index = */ memtx_space_build_index,
 	/* .swap_index = */ generic_space_swap_index,
 	/* .prepare_alter = */ memtx_space_prepare_alter,
+	/* .finish_alter = */ memtx_space_finish_alter,
 	/* .prepare_upgrade = */ memtx_space_prepare_upgrade,
-	/* .invalidate = */ generic_space_invalidate,
+	/* .invalidate = */ memtx_space_invalidate,
 };
 
 struct space *
@@ -1414,10 +1502,6 @@ memtx_space_new(struct memtx_engine *memtx,
 	int key_count = 0;
 	size_t region_svp = region_used(&fiber()->gc);
 	struct key_def **keys = index_def_to_key_def(key_list, &key_count);
-	if (keys == NULL) {
-		free(memtx_space);
-		return NULL;
-	}
 	struct tuple_format *format =
 		space_tuple_format_new(&memtx_tuple_format_vtab,
 				       memtx, keys, key_count, def);
