@@ -89,7 +89,8 @@ static void
 box_raft_election_fencing_resume(void);
 
 static void
-box_raft_msg_to_request(const struct raft_msg *msg, struct raft_request *req)
+box_raft_msg_to_request(const struct raft_msg *msg, struct raft_request *req,
+			enum group_id group_id)
 {
 	*req = (struct raft_request) {
 		.term = msg->term,
@@ -98,6 +99,7 @@ box_raft_msg_to_request(const struct raft_msg *msg, struct raft_request *req)
 		.is_leader_seen = msg->is_leader_seen,
 		.state = msg->state,
 		.vclock = msg->vclock,
+		.group_id = group_id,
 	};
 }
 
@@ -122,14 +124,22 @@ box_raft_update_synchro_queue(struct raft *raft)
 		return;
 	int rc = 0;
 	uint32_t errcode = 0;
+	bool try_again;
 	do {
+		try_again = false;
 		rc = box_promote_qsync();
 		if (rc != 0) {
 			struct error *err = diag_last_error(diag_get());
 			errcode = box_error_code(err);
 			diag_log();
+			if (!fiber_is_cancelled() &&
+			    (errcode == ER_QUORUM_WAIT ||
+			    errcode == ER_IN_ANOTHER_PROMOTE)) {
+				try_again = true;
+				fiber_sleep(0);
+			}
 		}
-	} while (rc != 0 && errcode == ER_QUORUM_WAIT && !fiber_is_cancelled());
+	} while (try_again);
 }
 
 static int
@@ -292,6 +302,19 @@ box_raft_fence(void)
 	raft_resign(raft);
 }
 
+void
+box_raft_leader_step_off(void)
+{
+	struct raft *raft = box_raft();
+	if (!raft->is_enabled || raft->state != RAFT_STATE_LEADER)
+		return;
+
+	/* It will be unfenced the next time new term is written. */
+	txn_limbo_fence(&txn_limbo);
+	raft_resign(raft);
+	raft_restore(raft);
+}
+
 /**
  * Configure the raft node according to whether it has a quorum of connected
  * peers or not. It can't start elections, when it doesn't.
@@ -343,7 +366,7 @@ box_raft_checkpoint_local(struct raft_request *req)
 {
 	struct raft_msg msg;
 	raft_checkpoint_local(box_raft(), &msg);
-	box_raft_msg_to_request(&msg, req);
+	box_raft_msg_to_request(&msg, req, GROUP_LOCAL);
 }
 
 void
@@ -351,7 +374,7 @@ box_raft_checkpoint_remote(struct raft_request *req)
 {
 	struct raft_msg msg;
 	raft_checkpoint_remote(box_raft(), &msg);
-	box_raft_msg_to_request(&msg, req);
+	box_raft_msg_to_request(&msg, req, GROUP_DEFAULT);
 }
 
 int
@@ -368,7 +391,7 @@ box_raft_broadcast(struct raft *raft, const struct raft_msg *msg)
 	(void)raft;
 	assert(raft == box_raft());
 	struct raft_request req;
-	box_raft_msg_to_request(msg, &req);
+	box_raft_msg_to_request(msg, &req, GROUP_DEFAULT);
 	replicaset_foreach(replica)
 		relay_push_raft(replica->relay, &req);
 	trigger_run(&box_raft_on_broadcast, NULL);
@@ -384,7 +407,7 @@ box_raft_write(struct raft *raft, const struct raft_msg *msg)
 	assert(msg->state == 0);
 
 	struct raft_request req;
-	box_raft_msg_to_request(msg, &req);
+	box_raft_msg_to_request(msg, &req, GROUP_LOCAL);
 	struct region *region = &fiber()->gc;
 	uint32_t svp = region_used(region);
 	struct xrow_header row;
