@@ -442,7 +442,7 @@ void
 space_delete(struct space *space)
 {
 	assert(rlist_empty(&space->alter_stmts));
-	memtx_tx_on_space_delete(space);
+	assert(rlist_empty(&space->memtx_stories));
 	for (uint32_t j = 0; j <= space->index_id_max; j++) {
 		struct index *index = space->index_map[j];
 		if (index != NULL)
@@ -516,15 +516,6 @@ space_index_def(struct space *space, int n)
 	return space->index[n]->def;
 }
 
-const char *
-index_name_by_id(struct space *space, uint32_t id)
-{
-	struct index *index = space_index(space, id);
-	if (index != NULL)
-		return index->def->name;
-	return NULL;
-}
-
 /**
  * Run BEFORE triggers and foreign key constraint checks registered for a space.
  * If a trigger changes the current statement, this function updates the
@@ -680,10 +671,12 @@ after_old_tuple_lookup:;
 				container_of(h, struct tuple_constraint,
 					     space_cache_holder);
 			assert(constr->def.type == CONSTR_FKEY);
-			if (tuple_constraint_fkey_check_delete(constr,
-							       old_tuple,
-							       new_tuple) != 0)
+			if (tuple_constraint_fkey_check_delete(
+					constr, old_tuple, new_tuple) != 0) {
+				if (new_tuple != NULL)
+					tuple_unref(new_tuple);
 				return -1;
+			}
 		}
 	}
 
@@ -971,6 +964,13 @@ generic_space_prepare_alter(struct space *old_space, struct space *new_space)
 	return 0;
 }
 
+void
+generic_space_finish_alter(struct space *old_space, struct space *new_space)
+{
+	(void)old_space;
+	(void)new_space;
+}
+
 int
 generic_space_prepare_upgrade(struct space *old_space, struct space *new_space)
 {
