@@ -304,7 +304,7 @@ mem_set_null(struct Mem *mem)
 }
 
 void
-mem_set_int(struct Mem *mem, int64_t value, bool is_neg)
+mem_set_int_with_sign(struct Mem *mem, int64_t value, bool is_neg)
 {
 	mem_clear(mem);
 	mem->u.i = value;
@@ -817,7 +817,7 @@ str_to_int(struct Mem *mem)
 	int64_t i;
 	if (sql_atoi64(mem->z, &i, &is_neg, mem->n) != 0)
 		return -1;
-	mem_set_int(mem, i, is_neg);
+	mem_set_int_with_sign(mem, i, is_neg);
 	return 0;
 }
 
@@ -865,7 +865,7 @@ str_to_datetime(struct Mem *mem)
 {
 	assert(mem->type == MEM_TYPE_STR);
 	struct datetime dt;
-	if (datetime_parse_full(&dt, mem->z, mem->n, NULL, 0) <= 0)
+	if (datetime_parse_full(&dt, mem->z, mem->n) <= 0)
 		return -1;
 	mem_set_datetime(mem, &dt);
 	return 0;
@@ -1739,88 +1739,38 @@ mem_cast_implicit_number(struct Mem *mem, enum field_type type)
 	return 0;
 }
 
-int
-mem_get_int(const struct Mem *mem, int64_t *i, bool *is_neg)
+int64_t
+mem_get_int_unsafe(const struct Mem *mem)
 {
-	if (mem->type == MEM_TYPE_INT) {
-		*i = mem->u.i;
-		*is_neg = true;
-		return 0;
-	}
-	if (mem->type == MEM_TYPE_UINT) {
-		*i = mem->u.i;
-		*is_neg = false;
-		return 0;
-	}
-	if ((mem->type & (MEM_TYPE_STR | MEM_TYPE_BIN)) != 0)
-		return sql_atoi64(mem->z, i, is_neg, mem->n);
-	if (mem->type == MEM_TYPE_DOUBLE) {
-		double d = mem->u.r;
-		if (d <= -1.0 && d >= (double)INT64_MIN) {
-			*i = (int64_t)d;
-			*is_neg = true;
-			return 0;
-		}
-		if (d > -1.0 && d < (double)UINT64_MAX) {
-			*i = (int64_t)(uint64_t)d;
-			*is_neg = false;
-			return 0;
-		}
-		return -1;
-	}
-	if (mem->type == MEM_TYPE_DEC) {
-		if (decimal_is_neg(&mem->u.d)) {
-			if (decimal_to_int64(&mem->u.d, i) == NULL)
-				return -1;
-			*is_neg = *i < 0;
-			return 0;
-		}
-		if (decimal_to_uint64(&mem->u.d, (uint64_t *)i) == NULL)
-			return -1;
-		*is_neg = false;
-		return 0;
-	}
-	return -1;
-}
-
-int
-mem_get_uint(const struct Mem *mem, uint64_t *u)
-{
-	if (mem->type == MEM_TYPE_INT)
-		return -1;
-	if (mem->type == MEM_TYPE_UINT) {
-		*u = mem->u.u;
-		return 0;
-	}
+	/* Yes, several places here can return values > INT64_MAX. */
+	if ((mem->type & (MEM_TYPE_INT | MEM_TYPE_UINT)) != 0)
+		return mem->u.i;
 	if ((mem->type & (MEM_TYPE_STR | MEM_TYPE_BIN)) != 0) {
-		bool is_neg;
-		if (sql_atoi64(mem->z, (int64_t *)u, &is_neg, mem->n) != 0 ||
-		    is_neg)
-			return -1;
-		return 0;
+		bool is_neg = false;
+		int64_t val = 0;
+		return sql_atoi64(mem->z, &val, &is_neg, mem->n) != 0 ? 0 : val;
 	}
 	if (mem->type == MEM_TYPE_DOUBLE) {
 		double d = mem->u.r;
-		if (d > -1.0 && d < (double)UINT64_MAX) {
-			*u = (uint64_t)d;
-			return 0;
-		}
-		return -1;
+		if (d <= -1.0 && d >= (double)INT64_MIN)
+			return (int64_t)d;
+		if (d > -1.0 && d < (double)UINT64_MAX)
+			return (int64_t)(uint64_t)d;
+		return 0;
 	}
 	if (mem->type == MEM_TYPE_DEC) {
 		if (decimal_is_neg(&mem->u.d)) {
-			int64_t i;
-			if (decimal_to_int64(&mem->u.d, &i) == NULL || i < 0)
-				return -1;
-			assert(i == 0);
-			*u = 0;
-			return 0;
+			int64_t val = 0;
+			if (decimal_to_int64(&mem->u.d, &val) == NULL)
+				return 0;
+			return val;
 		}
-		if (decimal_to_uint64(&mem->u.d, u) == NULL)
-			return -1;
-		return 0;
+		uint64_t val = 0;
+		if (decimal_to_uint64(&mem->u.d, &val) == NULL)
+			return 0;
+		return val;
 	}
-	return -1;
+	return 0;
 }
 
 int
@@ -1878,38 +1828,6 @@ mem_get_dec(const struct Mem *mem, decimal_t *d)
 	return -1;
 }
 
-int
-mem_get_bool(const struct Mem *mem, bool *b)
-{
-	if (mem->type == MEM_TYPE_BOOL) {
-		*b = mem->u.b;
-		return 0;
-	}
-	return -1;
-}
-
-int
-mem_get_bin(const struct Mem *mem, const char **s)
-{
-	if (mem->type == MEM_TYPE_STR) {
-		*s = mem->n > 0 ? mem->z : NULL;
-		return 0;
-	}
-	if (mem->type != MEM_TYPE_BIN)
-		return -1;
-	*s = mem->z;
-	return 0;
-}
-
-int
-mem_len(const struct Mem *mem, size_t *len)
-{
-	if (!mem_is_bytes(mem))
-		return -1;
-	*len = mem->n;
-	return 0;
-}
-
 int
 mem_copy(struct Mem *to, const struct Mem *from)
 {
@@ -2086,7 +2004,7 @@ mem_add_num(const struct Mem *left, const struct Mem *right, struct Mem *result)
 		diag_set(ClientError, ER_SQL_EXECUTE, "integer is overflowed");
 		return -1;
 	}
-	mem_set_int(result, res, is_neg);
+	mem_set_int_with_sign(result, res, is_neg);
 	return 0;
 }
 
@@ -2192,7 +2110,7 @@ mem_sub_num(const struct Mem *left, const struct Mem *right, struct Mem *result)
 		diag_set(ClientError, ER_SQL_EXECUTE, "integer is overflowed");
 		return -1;
 	}
-	mem_set_int(result, res, is_neg);
+	mem_set_int_with_sign(result, res, is_neg);
 	return 0;
 }
 
@@ -2300,7 +2218,7 @@ mem_mul(const struct Mem *left, const struct Mem *right, struct Mem *result)
 		diag_set(ClientError, ER_SQL_EXECUTE, "integer is overflowed");
 		return -1;
 	}
-	mem_set_int(result, res, is_neg);
+	mem_set_int_with_sign(result, res, is_neg);
 	return 0;
 }
 
@@ -2358,7 +2276,7 @@ mem_div(const struct Mem *left, const struct Mem *right, struct Mem *result)
 		diag_set(ClientError, ER_SQL_EXECUTE, "integer is overflowed");
 		return -1;
 	}
-	mem_set_int(result, res, is_neg);
+	mem_set_int_with_sign(result, res, is_neg);
 	return 0;
 }
 
@@ -2390,7 +2308,7 @@ mem_rem(const struct Mem *left, const struct Mem *right, struct Mem *result)
 		diag_set(ClientError, ER_SQL_EXECUTE, "integer is overflowed");
 		return -1;
 	}
-	mem_set_int(result, res, is_neg);
+	mem_set_int_with_sign(result, res, is_neg);
 	return 0;
 }
 
@@ -2682,8 +2600,8 @@ mem_cmp_msgpack(const struct Mem *a, const char **b, int *result,
 		mem.u.u = mp_decode_uint(b);
 		break;
 	case MP_INT:
-		mem.type = MEM_TYPE_INT;
 		mem.u.i = mp_decode_int(b);
+		mem.type = mem.u.i < 0 ? MEM_TYPE_INT : MEM_TYPE_UINT;
 		break;
 	case MP_FLOAT:
 		mem.type = MEM_TYPE_DOUBLE;
@@ -3109,7 +3027,7 @@ mem_from_mp_ephemeral(struct Mem *mem, const char *buf, uint32_t *len)
 	}
 	case MP_INT: {
 		mem->u.i = mp_decode_int(&buf);
-		mem->type = MEM_TYPE_INT;
+		mem->type = mem->u.i < 0 ? MEM_TYPE_INT : MEM_TYPE_UINT;
 		mem->flags = 0;
 		break;
 	}
@@ -3237,13 +3155,7 @@ mem_to_mp(const struct Mem *mem, uint32_t *size, struct region *region)
 		return NULL;
 	}
 	*size = region_used(region) - used;
-	char *data = region_join(region, *size);
-	if (data == NULL) {
-		region_truncate(region, used);
-		diag_set(OutOfMemory, *size, "region_join", "data");
-		return NULL;
-	}
-	return data;
+	return xregion_join(region, *size);
 }
 
 char *
@@ -3266,12 +3178,7 @@ mem_encode_array(const struct Mem *mems, uint32_t count, uint32_t *size,
 		return NULL;
 	}
 	*size = region_used(region) - used;
-	char *array = region_join(region, *size);
-	if (array == NULL) {
-		region_truncate(region, used);
-		diag_set(OutOfMemory, *size, "region_join", "array");
-		return NULL;
-	}
+	char *array = xregion_join(region, *size);
 	mp_tuple_assert(array, array + *size);
 	return array;
 }
@@ -3306,10 +3213,7 @@ mem_encode_map(const struct Mem *mems, uint32_t count, uint32_t *size,
 		goto error;
 	}
 	*size = region_used(region) - used;
-	char *map = region_join(region, *size);
-	if (map != NULL)
-		return map;
-	diag_set(OutOfMemory, *size, "region_join", "map");
+	return xregion_join(region, *size);
 error:
 	region_truncate(region, used);
 	return NULL;
@@ -3382,13 +3286,7 @@ static struct Mem *
 vdbemem_alloc_on_region(uint32_t count)
 {
 	struct region *region = &fiber()->gc;
-	size_t size;
-	struct Mem *ret = region_alloc_array(region, typeof(*ret), count,
-					     &size);
-	if (ret == NULL) {
-		diag_set(OutOfMemory, size, "region_alloc_array", "ret");
-		return NULL;
-	}
+	struct Mem *ret = xregion_alloc_array(region, struct Mem, count);
 	memset(ret, 0, count * sizeof(*ret));
 	for (uint32_t i = 0; i < count; i++) {
 		mem_create(&ret[i]);
@@ -3463,15 +3361,11 @@ port_vdbemem_get_msgpack(struct port *base, uint32_t *size)
 		mem_to_mpstream((struct Mem *)port->mem + i, &stream);
 	mpstream_flush(&stream);
 	*size = region_used(region) - region_svp;
-	if (is_error)
-		goto error;
-	const char *ret = (char *)region_join(region, *size);
-	if (ret == NULL)
-		goto error;
-	return ret;
-error:
-	diag_set(OutOfMemory, *size, "region", "ret");
-	return NULL;
+	if (is_error) {
+		diag_set(OutOfMemory, *size, "region", "ret");
+		return NULL;
+	}
+	return xregion_join(region, *size);
 }
 
 static const struct port_vtab port_vdbemem_vtab;
@@ -3521,8 +3415,6 @@ port_lua_get_vdbemem(struct port *base, uint32_t *size)
 	struct region *region = &fiber()->gc;
 	size_t region_svp = region_used(region);
 	struct Mem *val = vdbemem_alloc_on_region(argc);
-	if (val == NULL)
-		return NULL;
 	for (int i = 0; i < argc; i++) {
 		struct luaL_field field;
 		int index = -1 - i;
@@ -3590,12 +3482,7 @@ port_lua_get_vdbemem(struct port *base, uint32_t *size)
 				goto error;
 			}
 			uint32_t size = region_used(region) - used;
-			char *raw = region_join(region, size);
-			if (raw == NULL) {
-				diag_set(OutOfMemory, size, "region_join",
-					 "raw");
-				goto error;
-			}
+			char *raw = xregion_join(region, size);
 			rc = is_map ? mem_copy_map(&val[i], raw, size) :
 			     mem_copy_array(&val[i], raw, size);
 			if (rc != 0)
@@ -3649,8 +3536,6 @@ port_c_get_vdbemem(struct port *base, uint32_t *size)
 	struct region *region = &fiber()->gc;
 	size_t region_svp = region_used(region);
 	struct Mem *val = vdbemem_alloc_on_region(port->size);
-	if (val == NULL)
-		return NULL;
 	int i = 0;
 	const char *data;
 	struct port_c_entry *pe;
@@ -3696,9 +3581,10 @@ port_c_get_vdbemem(struct port *base, uint32_t *size)
 			val[i].u.r = d;
 			break;
 		case MP_INT:
-			val[i].type = MEM_TYPE_INT;
 			assert(val[i].flags == 0);
 			val[i].u.i = mp_decode_int(&data);
+			val[i].type = val[i].u.i < 0 ?
+				MEM_TYPE_INT : MEM_TYPE_UINT;
 			break;
 		case MP_UINT:
 			val[i].type = MEM_TYPE_UINT;
