@@ -156,30 +156,18 @@ char *
 sql_normalized_name_region_new(struct region *r, const char *name, int len)
 {
 	int size = len + 1;
-	ERROR_INJECT(ERRINJ_SQL_NAME_NORMALIZATION, {
-		diag_set(OutOfMemory, size, "region_alloc", "res");
-		return NULL;
-	});
 	size_t region_svp = region_used(r);
-	char *res = region_alloc(r, size);
-	if (res == NULL)
-		goto oom_error;
+	char *res = xregion_alloc(r, size);
 	int rc = sql_normalize_name(res, size, name, len);
 	if (rc <= size)
 		return res;
 
 	size = rc;
 	region_truncate(r, region_svp);
-	res = region_alloc(r, size);
-	if (res == NULL)
-		goto oom_error;
+	res = xregion_alloc(r, size);
 	if (sql_normalize_name(res, size, name, len) > size)
 		unreachable();
 	return res;
-
-oom_error:
-	diag_set(OutOfMemory, size, "region_alloc", "res");
-	return NULL;
 }
 
 /* Convenient short-hand */
@@ -790,86 +778,6 @@ sqlGetVarint(const unsigned char *p, u64 * v)
 	return 9;
 }
 
-/*
- * Read a 32-bit variable-length integer from memory starting at p[0].
- * Return the number of bytes read.  The value is stored in *v.
- *
- * If the varint stored in p[0] is larger than can fit in a 32-bit unsigned
- * integer, then set *v to 0xffffffff.
- *
- * A MACRO version, getVarint32, is provided which inlines the
- * single-byte case.  All code should use the MACRO version as
- * this function assumes the single-byte case has already been handled.
- */
-u8
-sqlGetVarint32(const unsigned char *p, u32 * v)
-{
-	u32 a, b;
-
-	/* The 1-byte case.  Overwhelmingly the most common.  Handled inline
-	 * by the getVarin32() macro
-	 */
-	a = *p;
-	/* a: p0 (unmasked) */
-#ifndef getVarint32
-	if (!(a & 0x80)) {
-		/* Values between 0 and 127 */
-		*v = a;
-		return 1;
-	}
-#endif
-
-	/* The 2-byte case */
-	p++;
-	b = *p;
-	/* b: p1 (unmasked) */
-	if (!(b & 0x80)) {
-		/* Values between 128 and 16383 */
-		a &= 0x7f;
-		a = a << 7;
-		*v = a | b;
-		return 2;
-	}
-
-	/* The 3-byte case */
-	p++;
-	a = a << 14;
-	a |= *p;
-	/* a: p0<<14 | p2 (unmasked) */
-	if (!(a & 0x80)) {
-		/* Values between 16384 and 2097151 */
-		a &= (0x7f << 14) | (0x7f);
-		b &= 0x7f;
-		b = b << 7;
-		*v = a | b;
-		return 3;
-	}
-
-	/* A 32-bit varint is used to store size information in btrees.
-	 * Objects are rarely larger than 2MiB limit of a 3-byte varint.
-	 * A 3-byte varint is sufficient, for example, to record the size
-	 * of a 1048569-byte BLOB or string.
-	 *
-	 * We only unroll the first 1-, 2-, and 3- byte cases.  The very
-	 * rare larger cases can be handled by the slower 64-bit varint
-	 * routine.
-	 */
-	{
-		u64 v64;
-		u8 n;
-
-		p -= 2;
-		n = sqlGetVarint(p, &v64);
-		assert(n > 3 && n <= 9);
-		if ((v64 & SQL_MAX_U32) != v64) {
-			*v = 0xffffffff;
-		} else {
-			*v = (u32) v64;
-		}
-		return n;
-	}
-}
-
 /*
  * Return the number of bytes that will be needed to store the given
  * 64-bit integer.
@@ -912,6 +820,13 @@ sqlHexToBlob(const char *z, int n)
 	return zBlob;
 }
 
+#if ENABLE_UB_SANITIZER
+/* See https://github.com/tarantool/tarantool/issues/10703. */
+int
+sql_add_int(int64_t lhs, bool is_lhs_neg, int64_t rhs, bool is_rhs_neg,
+	    int64_t *res, bool *is_res_neg)
+  __attribute__((no_sanitize("signed-integer-overflow")));
+#endif
 int
 sql_add_int(int64_t lhs, bool is_lhs_neg, int64_t rhs, bool is_rhs_neg,
 	    int64_t *res, bool *is_res_neg)
@@ -941,6 +856,13 @@ sql_add_int(int64_t lhs, bool is_lhs_neg, int64_t rhs, bool is_rhs_neg,
 	return 0;
 }
 
+#if ENABLE_UB_SANITIZER
+/* See https://github.com/tarantool/tarantool/issues/10703. */
+int
+sql_sub_int(int64_t lhs, bool is_lhs_neg, int64_t rhs, bool is_rhs_neg,
+	    int64_t *res, bool *is_res_neg)
+  __attribute__((no_sanitize("signed-integer-overflow")));
+#endif
 int
 sql_sub_int(int64_t lhs, bool is_lhs_neg, int64_t rhs, bool is_rhs_neg,
 	    int64_t *res, bool *is_res_neg)
@@ -1026,6 +948,13 @@ sql_mul_int(int64_t lhs, bool is_lhs_neg, int64_t rhs, bool is_rhs_neg,
 	return 0;
 }
 
+#if ENABLE_UB_SANITIZER
+/* See https://github.com/tarantool/tarantool/issues/10703. */
+int
+sql_div_int(int64_t lhs, bool is_lhs_neg, int64_t rhs, bool is_rhs_neg,
+	    int64_t *res, bool *is_res_neg)
+  __attribute__((no_sanitize("signed-integer-overflow")));
+#endif
 int
 sql_div_int(int64_t lhs, bool is_lhs_neg, int64_t rhs, bool is_rhs_neg,
 	    int64_t *res, bool *is_res_neg)
