@@ -1232,13 +1232,13 @@ sqlVdbeList(Vdbe * p)
 			}
 		}
 
-		mem_set_int(pMem, pOp->p1, pOp->p1 < 0);
+		mem_set_int(pMem, pOp->p1);
 		pMem++;
 
-		mem_set_int(pMem, pOp->p2, pOp->p2 < 0);
+		mem_set_int(pMem, pOp->p2);
 		pMem++;
 
-		mem_set_int(pMem, pOp->p3, pOp->p3 < 0);
+		mem_set_int(pMem, pOp->p3);
 		pMem++;
 
 		char *buf = sql_xmalloc(256);
