@@ -105,23 +105,6 @@ mp_extension_class(const char *data)
 	return mp_ext_classes[type];
 }
 
-/*
- * @brief Read the msgpack value at \p field (whatever it is: int, uint, float
- *        or double) as double.
- * @param field the field to read the value from
- * @retval the read-and-cast result
- */
-static double
-mp_read_as_double(const char *field)
-{
-	double result = NAN;
-	assert(mp_classof(mp_typeof(*field)) == MP_CLASS_NUMBER);
-	/* This can only fail on non-numeric msgpack field, so it shouldn't. */
-	if (mp_read_double_lossy(&field, &result) == -1)
-		unreachable();
-	return result;
-}
-
 static int
 mp_compare_bool(const char *field_a, const char *field_b)
 {
@@ -130,14 +113,6 @@ mp_compare_bool(const char *field_a, const char *field_b)
 	return COMPARE_RESULT(a_val, b_val);
 }
 
-static int
-mp_compare_as_double(const char *field_a, const char *field_b)
-{
-	double a_val = mp_read_as_double(field_a);
-	double b_val = mp_read_as_double(field_b);
-	return COMPARE_RESULT(a_val, b_val);
-}
-
 static int
 mp_compare_integer_with_type(const char *field_a, enum mp_type a_type,
 			     const char *field_b, enum mp_type b_type)
@@ -238,16 +213,17 @@ mp_compare_decimal(const char *lhs, const char *rhs)
 }
 
 /**
- * Compare decimal to 'not a number' which is either NaN or inf. Decimal value
- * does not matter then. It is always a valid number.
+ * Compare a decimal to something not representable as decimal. Like NaN, Inf or
+ * just a value outside the (-1e38, 1e38) range. In all these cases the decimal
+ * value doesn't matter.
  */
 static inline int
-decimal_compare_nan(double rhs)
+decimal_compare_nan_or_huge(double rhs)
 {
 	/* We assume NaN is less than everything else. */
 	if (isnan(rhs))
 		return 1;
-	assert(isinf(rhs));
+	assert(fabs(rhs) >= 1e38);
 	return (rhs < 0) - (rhs > 0);
 }
 
@@ -263,7 +239,7 @@ mp_compare_decimal_any_number(decimal_t *lhs, const char *rhs,
 		double d = mp_decode_float(&rhs);
 		rc = decimal_from_double(&rhs_dec, d);
 		if (rc == NULL)
-			return decimal_compare_nan(d) * k;
+			return decimal_compare_nan_or_huge(d) * k;
 		break;
 	}
 	case MP_DOUBLE:
@@ -271,7 +247,7 @@ mp_compare_decimal_any_number(decimal_t *lhs, const char *rhs,
 		double d = mp_decode_double(&rhs);
 		rc = decimal_from_double(&rhs_dec, d);
 		if (rc == NULL)
-			return decimal_compare_nan(d) * k;
+			return decimal_compare_nan_or_huge(d) * k;
 		break;
 	}
 	case MP_INT:
@@ -507,9 +483,8 @@ tuple_compare_field(const char *field_a, const char *field_b,
 						    field_b,
 						    mp_typeof(*field_b));
 	case FIELD_TYPE_NUMBER:
-		return mp_compare_number(field_a, field_b);
 	case FIELD_TYPE_DOUBLE:
-		return mp_compare_as_double(field_a, field_b);
+		return mp_compare_number(field_a, field_b);
 	case FIELD_TYPE_BOOLEAN:
 		return mp_compare_bool(field_a, field_b);
 	case FIELD_TYPE_VARBINARY:
@@ -546,10 +521,9 @@ tuple_compare_field_with_type(const char *field_a, enum mp_type a_type,
 		return mp_compare_integer_with_type(field_a, a_type,
 						    field_b, b_type);
 	case FIELD_TYPE_NUMBER:
+	case FIELD_TYPE_DOUBLE:
 		return mp_compare_number_with_type(field_a, a_type,
 						   field_b, b_type);
-	case FIELD_TYPE_DOUBLE:
-		return mp_compare_as_double(field_a, field_b);
 	case FIELD_TYPE_BOOLEAN:
 		return mp_compare_bool(field_a, field_b);
 	case FIELD_TYPE_VARBINARY:
@@ -1696,7 +1670,8 @@ hint_uint(uint64_t u)
 static inline hint_t
 hint_int(int64_t i)
 {
-	assert(i < 0);
+	if (i >= 0)
+		return hint_uint((uint64_t)i);
 	uint64_t val = (i <= HINT_VALUE_INT_MIN ? 0 : i - HINT_VALUE_INT_MIN);
 	return hint_create(MP_CLASS_NUMBER, val);
 }
@@ -1847,12 +1822,6 @@ field_hint_integer(const char *field)
 	return HINT_NONE;
 }
 
-static inline hint_t
-field_hint_double(const char *field)
-{
-	return hint_double(mp_read_as_double(field));
-}
-
 static inline hint_t
 field_hint_number(const char *field)
 {
@@ -2004,9 +1973,8 @@ field_hint(const char *field, struct coll *coll)
 	case FIELD_TYPE_INTEGER:
 		return field_hint_integer(field);
 	case FIELD_TYPE_NUMBER:
-		return field_hint_number(field);
 	case FIELD_TYPE_DOUBLE:
-		return field_hint_double(field);
+		return field_hint_number(field);
 	case FIELD_TYPE_STRING:
 		return field_hint_string(field, coll);
 	case FIELD_TYPE_VARBINARY:
