@@ -339,11 +339,15 @@ txn_stmt_prepare_rollback_info(struct txn_stmt *stmt, struct tuple *old_tuple,
 }
 
 /*
- * Undo changes done by a statement and run the corresponding
- * rollback triggers.
+ * Run the statement's rollback triggers and undo changes done by the statement.
+ *
+ * Logically, we call triggers after running statements. These triggers can
+ * make significant changes (for instance, DDL triggers), so, for consistency,
+ * we should call the statement's `on_rollback` triggers before rolling back
+ * the corresponding statement.
  *
  * Note, a trigger set by a particular statement must be run right
- * after the statement is rolled back, because rollback triggers
+ * before the statement is rolled back, because rollback triggers
  * installed by DDL statements restore the schema cache, which is
  * necessary to roll back previous statements. For example, to roll
  * back a DML statement applied to a space whose index is dropped
@@ -353,12 +357,12 @@ txn_stmt_prepare_rollback_info(struct txn_stmt *stmt, struct tuple *old_tuple,
 static void
 txn_rollback_one_stmt(struct txn *txn, struct txn_stmt *stmt)
 {
-	if (txn->engine != NULL && stmt->space != NULL)
-		engine_rollback_statement(txn->engine, txn, stmt);
-	if (stmt->has_triggers && trigger_run(&stmt->on_rollback, txn) != 0) {
+	if (stmt->has_triggers && trigger_run(&stmt->on_rollback, stmt) != 0) {
 		diag_log();
 		panic("statement rollback trigger failed");
 	}
+	if (txn->engine != NULL && stmt->space != NULL)
+		engine_rollback_statement(txn->engine, txn, stmt);
 }
 
 static void
@@ -420,10 +424,7 @@ txn_new(void)
 	return txn;
 }
 
-/*
- * Free txn memory and return it to a cache.
- */
-inline static void
+void
 txn_free(struct txn *txn)
 {
 	assert(txn->limbo_entry == NULL);
@@ -505,6 +506,11 @@ txn_begin(void)
 	 * if they are not supported.
 	 */
 	txn_set_flags(txn, TXN_CAN_YIELD);
+	/*
+	 * A transaction is unaffected by concurrent DDL as long as it has
+	 * no statements.
+	 */
+	txn_set_flags(txn, TXN_HANDLES_DDL);
 	memtx_tx_register_txn(txn);
 	rmean_collect(rmean_box, IPROTO_BEGIN, 1);
 	return txn;
@@ -517,7 +523,10 @@ txn_begin_in_engine(struct engine *engine, struct txn *txn)
 		return 0;
 	if (txn->engine == NULL) {
 		txn->engine = engine;
-		return engine_begin(engine, txn);
+		if (engine_begin(engine, txn) != 0)
+			return -1;
+		if ((engine->flags & ENGINE_TXM_HANDLES_DDL) == 0)
+			txn_clear_flags(txn, TXN_HANDLES_DDL);
 	} else if (txn->engine != engine) {
 		/**
 		 * Only one engine can be used in
@@ -704,7 +713,7 @@ txn_complete_fail(struct txn *txn)
 	if (txn->engine != NULL)
 		engine_rollback(txn->engine, txn);
 	if (txn_has_flag(txn, TXN_HAS_TRIGGERS)) {
-		if (trigger_run(&txn->on_rollback, txn) != 0) {
+		if (trigger_run(&txn->on_rollback, txn_first_stmt(txn)) != 0) {
 			diag_log();
 			panic("transaction rollback trigger failed");
 		}
@@ -733,7 +742,8 @@ txn_complete_success(struct txn *txn)
 		 * so that a trigger sees the changes done by previous triggers
 		 * (this is vital for DDL).
 		 */
-		if (trigger_run_reverse(&txn->on_commit, txn) != 0) {
+		if (trigger_run_reverse(&txn->on_commit,
+					txn_first_stmt(txn)) != 0) {
 			diag_log();
 			panic("transaction commit trigger failed");
 		}
@@ -794,16 +804,35 @@ txn_on_journal_write(struct journal_entry *entry)
 	} else {
 		int64_t lsn;
 		/*
-		 * XXX: that is quite ugly. Need a more reliable way to get the
-		 * synchro LSN.
+		 * Synchro lsn is taken from the last global row of a
+		 * transaction. When this is a remote transaction
+		 * (n_applier_rows > 0) we know for sure the last applier row is
+		 * the last global row. Otherwise we need to find the last
+		 * global row.
 		 */
-		if (txn->n_applier_rows > 0)
+		if (txn->n_applier_rows > 0) {
 			lsn = entry->rows[txn->n_applier_rows - 1]->lsn;
-		else
-			lsn = entry->rows[entry->n_rows - 1]->lsn;
+		} else {
+			int i = entry->n_rows - 1;
+			struct xrow_header **rows = entry->rows;
+			/*
+			 * We don't care which lsn to choose for a fully local
+			 * transaction.
+			 */
+			if (!txn_is_fully_local(txn)) {
+				while (rows[i]->group_id == GROUP_LOCAL)
+					i--;
+				assert(i >= 0);
+			}
+			lsn = rows[i]->lsn;
+		}
 		txn_limbo_assign_lsn(&txn_limbo, txn->limbo_entry, lsn);
 		if (txn->fiber != NULL)
 			fiber_wakeup(txn->fiber);
+		if (txn_limbo_is_owned_by_current_instance(&txn_limbo) &&
+		    txn_has_flag(txn, TXN_WAIT_ACK))
+			txn_limbo_ack(&txn_limbo, txn_limbo.owner_id,
+				      txn->limbo_entry->lsn);
 	}
 finish:
 	fiber_set_txn(fiber(), NULL);
@@ -819,9 +848,8 @@ txn_journal_entry_new(struct txn *txn)
 
 	assert(txn->n_new_rows + txn->n_applier_rows > 0);
 
-	/* Save space for an additional NOP row just in case. */
 	struct region *txn_region = tx_region_acquire(txn);
-	req = journal_entry_new(txn->n_new_rows + txn->n_applier_rows + 1,
+	req = journal_entry_new(txn->n_new_rows + txn->n_applier_rows,
 				txn_region, txn_on_journal_write, txn);
 	tx_region_release(txn, TX_ALLOC_SYSTEM);
 	txn_region = NULL;
@@ -886,30 +914,6 @@ txn_journal_entry_new(struct txn *txn)
 	assert(remote_row == req->rows + txn->n_applier_rows);
 	assert(local_row == remote_row + txn->n_new_rows);
 
-	/*
-	 * Append a dummy NOP statement to preserve replication tx
-	 * boundaries when the last tx row is a local one, and the
-	 * transaction has at least one global row.
-	 */
-	if (txn->n_local_rows > 0 &&
-	    (txn->n_local_rows != txn->n_new_rows || txn->n_applier_rows > 0) &&
-	    (*(local_row - 1))->group_id == GROUP_LOCAL) {
-		size_t size;
-		*local_row =
-			tx_region_alloc_object(txn, TX_OBJECT_XROW_HEADER,
-					       &size);
-		if (*local_row == NULL) {
-			diag_set(OutOfMemory, size, "tx_region_alloc_object",
-				 "row");
-			return NULL;
-		}
-		memset(*local_row, 0, sizeof(**local_row));
-		(*local_row)->type = IPROTO_NOP;
-		(*local_row)->group_id = GROUP_DEFAULT;
-	} else {
-		--req->n_rows;
-	}
-
 	static const uint8_t flags_map[] = {
 		[TXN_WAIT_SYNC] = IPROTO_FLAG_WAIT_SYNC,
 		[TXN_WAIT_ACK] = IPROTO_FLAG_WAIT_ACK,
@@ -949,6 +953,12 @@ txn_prepare(struct txn *txn)
 			return -1;
 		}
 	}
+	/*
+	 * Non-memtx transactions can use memtx MVCC (for example, transactions
+	 * reading from sysview), so we need to finalize prepare of all
+	 * transactions, not only memtx ones.
+	 */
+	memtx_tx_prepare_finalize(txn);
 
 	trigger_clear(&txn->fiber_on_stop);
 	trigger_clear(&txn->fiber_on_yield);
@@ -987,71 +997,44 @@ txn_add_limbo_entry(struct txn *txn, const struct journal_entry *req)
 	 * instance id in the first row.
 	 */
 	uint32_t origin_id = req->rows[0]->replica_id;
-	txn->limbo_entry = txn_limbo_append(&txn_limbo, origin_id, txn);
-	if (txn->limbo_entry == NULL)
-		return -1;
-	return 0;
+	return txn_limbo_submit(&txn_limbo, origin_id, txn);
 }
 
-int
-txn_commit_try_async(struct txn *txn)
+static int
+txn_commit_impl(struct txn *txn, enum txn_commit_wait_mode wait_mode)
 {
 	struct journal_entry *req;
-
-	ERROR_INJECT(ERRINJ_TXN_COMMIT_ASYNC, {
-		diag_set(ClientError, ER_INJECTION,
-			 "txn commit async injection");
-		goto rollback;
-	});
-
-	if (txn_prepare(txn) != 0)
-		goto rollback;
-
-	if (txn_commit_nop(txn))
-		return 0;
-
-	req = txn_journal_entry_new(txn);
-	if (req == NULL)
-		goto rollback;
-
-	if (txn_has_flag(txn, TXN_WAIT_SYNC) &&
-	    txn_add_limbo_entry(txn, req) != 0) {
-		goto rollback;
-	}
-
-	fiber_set_txn(fiber(), NULL);
-	if (journal_write_try_async(req) != 0) {
-		fiber_set_txn(fiber(), txn);
-		diag_log();
-		goto rollback;
-	}
-
-	return 0;
-
-rollback:
-	assert(txn->fiber == NULL);
-	txn_abort(txn);
-	return -1;
-}
-
-int
-txn_commit(struct txn *txn)
-{
-	struct journal_entry *req;
-
 	txn->fiber = fiber();
-
 	if (txn_prepare(txn) != 0)
 		goto rollback_abort;
-
 	if (txn_commit_nop(txn)) {
 		txn_free(txn);
 		return 0;
 	}
-
 	req = txn_journal_entry_new(txn);
 	if (req == NULL)
 		goto rollback_abort;
+	if (wait_mode != TXN_COMMIT_WAIT_MODE_COMPLETE) {
+		assert(wait_mode == TXN_COMMIT_WAIT_MODE_SUBMIT);
+		ERROR_INJECT(ERRINJ_TXN_COMMIT_ASYNC, {
+			diag_set(ClientError, ER_INJECTION,
+				 "txn commit async injection");
+			goto rollback_abort;
+		});
+		/*
+		 * The main reason why it is disabled is that the sync txn's
+		 * original instance wants to write CONFIRM. But it can't be
+		 * done asynchronously in the current code design. Also it is
+		 * logically strange to commit a synchronous transaction in an
+		 * async way, but the CONFIRM problem is the technical one.
+		 */
+		bool is_original = req->rows[0]->replica_id == 0;
+		if (txn_has_flag(txn, TXN_WAIT_ACK) && is_original) {
+			diag_set(ClientError, ER_UNSUPPORTED, "Tarantool",
+				 "Non-blocking commit of a synchronous txn");
+			goto rollback_abort;
+		}
+	}
 	/*
 	 * Do not cache the flag value in a variable. The flag might be deleted
 	 * during WAL write. This can happen for async transactions created
@@ -1063,10 +1046,17 @@ txn_commit(struct txn *txn)
 	    txn_add_limbo_entry(txn, req) != 0) {
 		goto rollback_abort;
 	}
-
 	fiber_set_txn(fiber(), NULL);
-	if (journal_write(req) != 0)
+	if (journal_write_submit(req) != 0)
 		goto rollback_io;
+	if (wait_mode != TXN_COMMIT_WAIT_MODE_COMPLETE) {
+		if (txn_has_flag(txn, TXN_IS_DONE))
+			goto finish_done;
+		txn->fiber = NULL;
+		return 0;
+	}
+	while (!req->is_complete)
+		fiber_yield();
 	if (req->res < 0) {
 		diag_set_journal_res(req->res);
 		goto rollback_io;
@@ -1074,22 +1064,19 @@ txn_commit(struct txn *txn)
 	if (txn_has_flag(txn, TXN_WAIT_SYNC)) {
 		struct txn_limbo_entry *limbo_entry = txn->limbo_entry;
 		assert(limbo_entry->lsn > 0);
-		/*
-		 * XXX: ACK should be done on WAL write too. But it can make
-		 * another WAL write. Can't be done until it works
-		 * asynchronously.
-		 */
-		if (txn_has_flag(txn, TXN_WAIT_ACK)) {
-			txn_limbo_ack(&txn_limbo, txn_limbo.owner_id,
-				      limbo_entry->lsn);
+		int rc = txn_limbo_wait_complete(&txn_limbo, limbo_entry);
+		if (rc < 0) {
+			if (fiber_is_cancelled()) {
+				txn->fiber = NULL;
+				return -1;
+			} else {
+				goto rollback;
+			}
 		}
-		if (txn_limbo_wait_complete(&txn_limbo, limbo_entry) < 0)
-			goto rollback;
 	}
+finish_done:
 	assert(txn_has_flag(txn, TXN_IS_DONE));
 	assert(txn->signature >= 0);
-
-	/* Synchronous transactions are freed by the calling fiber. */
 	txn_free(txn);
 	return 0;
 
@@ -1110,6 +1097,18 @@ rollback:
 	return -1;
 }
 
+int
+txn_commit_submit(struct txn *txn)
+{
+	return txn_commit_impl(txn, TXN_COMMIT_WAIT_MODE_SUBMIT);
+}
+
+int
+txn_commit(struct txn *txn)
+{
+	return txn_commit_impl(txn, TXN_COMMIT_WAIT_MODE_COMPLETE);
+}
+
 void
 txn_rollback_stmt(struct txn *txn)
 {
@@ -1277,6 +1276,26 @@ box_txn_set_timeout(double timeout)
 	return 0;
 }
 
+void
+txn_set_xrow_flags(uint8_t xrow_flags)
+{
+	struct txn *txn = in_txn();
+	/*
+	 * Do nothing if transaction is not started,
+	 * it's the same as BEGIN + COMMIT.
+	 */
+	if (!txn)
+		return;
+
+	const uint8_t flags_map[] = {
+		[IPROTO_FLAG_WAIT_SYNC] = TXN_WAIT_SYNC,
+		[IPROTO_FLAG_WAIT_ACK] = TXN_WAIT_ACK,
+	};
+
+	txn->flags |= flags_map[xrow_flags & IPROTO_FLAG_WAIT_SYNC];
+	txn->flags |= flags_map[xrow_flags & IPROTO_FLAG_WAIT_ACK];
+}
+
 /** Wait for a linearization point for a transaction. */
 static int
 txn_make_linearizable(struct txn *txn)
