@@ -30,16 +30,99 @@
  */
 #include "txn.h"
 #include "txn_limbo.h"
-#include "replication.h"
 #include "iproto_constants.h"
 #include "journal.h"
 #include "box.h"
 #include "raft.h"
 #include "tt_static.h"
 #include "session.h"
+#include "trivia/config.h"
 
 struct txn_limbo txn_limbo;
 
+static struct txn_limbo_entry *
+txn_limbo_first_entry(struct txn_limbo *limbo)
+{
+	return rlist_first_entry(&limbo->queue, struct txn_limbo_entry,
+				 in_queue);
+}
+
+static struct txn_limbo_entry *
+txn_limbo_last_entry(struct txn_limbo *limbo)
+{
+	return rlist_last_entry(&limbo->queue, struct txn_limbo_entry,
+				in_queue);
+}
+
+/** Create a request for a specific limbo and write it to WAL. */
+static int
+txn_limbo_write_synchro(struct txn_limbo *limbo, uint16_t type, int64_t lsn,
+			uint64_t term, struct vclock *vclock);
+
+/** Confirm all the entries <= @a lsn. */
+static void
+txn_limbo_read_confirm(struct txn_limbo *limbo, int64_t lsn);
+
+/**
+ * Write a confirmation entry to the WAL. After it's written all the
+ * transactions waiting for confirmation may be finished.
+ */
+static int
+txn_limbo_write_confirm(struct txn_limbo *limbo, int64_t lsn)
+{
+	assert(lsn > limbo->confirmed_lsn);
+	assert(!limbo->is_in_rollback);
+	return txn_limbo_write_synchro(limbo, IPROTO_RAFT_CONFIRM, lsn, 0,
+				       NULL);
+}
+
+/**
+ * Bring the persisted confirmed LSN closer to the currently known volatile
+ * confirmed LSN.
+ */
+static int
+txn_limbo_worker_bump_confirmed_lsn(struct txn_limbo *limbo)
+{
+	assert(limbo->volatile_confirmed_lsn >= limbo->confirmed_lsn);
+	while (txn_limbo_is_owned_by_current_instance(limbo) &&
+	       limbo->volatile_confirmed_lsn > limbo->confirmed_lsn) {
+		if (limbo->is_in_rollback)
+			return -1;
+		/* It can get bumped again while we are writing. */
+		int64_t volatile_confirmed_lsn = limbo->volatile_confirmed_lsn;
+		if (txn_limbo_write_confirm(limbo,
+					    volatile_confirmed_lsn) != 0) {
+			diag_log();
+			return -1;
+		}
+		ERROR_INJECT_YIELD(ERRINJ_TXN_LIMBO_WORKER_DELAY);
+		txn_limbo_read_confirm(limbo, volatile_confirmed_lsn);
+	}
+	assert(limbo->volatile_confirmed_lsn >= limbo->confirmed_lsn);
+	return 0;
+}
+
+static int
+txn_limbo_worker_f(va_list args)
+{
+	(void)args;
+	struct txn_limbo *limbo = fiber()->f_arg;
+	assert(limbo == &txn_limbo);
+	while (!fiber_is_cancelled()) {
+		fiber_check_gc();
+		ERROR_INJECT_YIELD(ERRINJ_TXN_LIMBO_WORKER_DELAY);
+		if (txn_limbo_worker_bump_confirmed_lsn(limbo) != 0)
+#ifdef TEST_BUILD
+			fiber_sleep(0.01);
+#else
+			fiber_sleep(1);
+#endif
+		else
+			fiber_yield();
+	}
+	return 0;
+}
+
 static inline void
 txn_limbo_create(struct txn_limbo *limbo)
 {
@@ -49,15 +132,38 @@ txn_limbo_create(struct txn_limbo *limbo)
 	fiber_cond_create(&limbo->wait_cond);
 	vclock_create(&limbo->vclock);
 	vclock_create(&limbo->promote_term_map);
+	vclock_create(&limbo->confirmed_vclock);
 	limbo->promote_greatest_term = 0;
 	latch_create(&limbo->promote_latch);
 	limbo->confirmed_lsn = 0;
+	limbo->volatile_confirmed_lsn = 0;
+	limbo->entry_to_confirm = NULL;
 	limbo->rollback_count = 0;
 	limbo->is_in_rollback = false;
 	limbo->svp_confirmed_lsn = -1;
 	limbo->frozen_reasons = 0;
 	limbo->is_frozen_until_promotion = true;
 	limbo->do_validate = false;
+	limbo->worker = fiber_new_system("txn_limbo_worker",
+					 txn_limbo_worker_f);
+	if (limbo->worker == NULL)
+		panic("failed to allocate synchronous queue worker fiber");
+	limbo->worker->f_arg = limbo;
+	fiber_set_joinable(limbo->worker, true);
+}
+
+static inline void
+txn_limbo_destroy(struct txn_limbo *limbo)
+{
+	struct txn_limbo_entry *entry;
+	while (!rlist_empty(&limbo->queue)) {
+		entry = rlist_shift_entry(&limbo->queue,
+					  typeof(*entry), in_queue);
+		entry->txn->limbo_entry = NULL;
+		txn_free(entry->txn);
+	}
+	fiber_cond_destroy(&limbo->wait_cond);
+	TRASH(limbo);
 }
 
 static inline bool
@@ -70,10 +176,21 @@ bool
 txn_limbo_is_ro(struct txn_limbo *limbo)
 {
 	return limbo->owner_id != REPLICA_ID_NIL &&
-	       (limbo->owner_id != instance_id || txn_limbo_is_frozen(limbo));
+		(!txn_limbo_is_owned_by_current_instance(limbo) ||
+		txn_limbo_is_frozen(limbo));
 }
 
-void
+/** Pop the first entry. */
+static inline void
+txn_limbo_pop_first(struct txn_limbo *limbo, struct txn_limbo_entry *entry)
+{
+	assert(!rlist_empty(&entry->in_queue));
+	assert(txn_limbo_first_entry(limbo) == entry);
+	rlist_del_entry(entry, in_queue);
+	limbo->len--;
+}
+
+static void
 txn_limbo_complete(struct txn *txn, bool is_success)
 {
 	/*
@@ -107,6 +224,38 @@ txn_limbo_complete(struct txn *txn, bool is_success)
 	fiber_set_session(fiber(), orig_session);
 }
 
+/** Complete the given limbo entry with a failure and the given reason. */
+static void
+txn_limbo_complete_fail(struct txn_limbo *limbo, struct txn_limbo_entry *entry,
+			int64_t signature)
+{
+	assert(entry->state == TXN_LIMBO_ENTRY_SUBMITTED);
+	struct txn *txn = entry->txn;
+	txn->signature = signature;
+	txn->limbo_entry = NULL;
+	txn_limbo_abort(limbo, entry);
+	txn_clear_flags(txn, TXN_WAIT_SYNC | TXN_WAIT_ACK);
+	txn_limbo_complete(txn, false);
+}
+
+/** Complete the given limbo entry with a success. */
+static void
+txn_limbo_complete_success(struct txn_limbo *limbo, struct txn_limbo_entry *entry)
+{
+	assert(entry->state == TXN_LIMBO_ENTRY_SUBMITTED);
+	struct txn *txn = entry->txn;
+	entry->state = TXN_LIMBO_ENTRY_COMMIT;
+	txn->limbo_entry = NULL;
+	txn_limbo_pop_first(limbo, entry);
+	txn_clear_flags(txn, TXN_WAIT_SYNC | TXN_WAIT_ACK);
+	/*
+	 * Should be written to WAL by now. Confirm is always written after the
+	 * affected transactions.
+	 */
+	assert(txn->signature >= 0);
+	txn_limbo_complete(txn, true);
+}
+
 struct txn_limbo_entry *
 txn_limbo_last_synchro_entry(struct txn_limbo *limbo)
 {
@@ -118,8 +267,8 @@ txn_limbo_last_synchro_entry(struct txn_limbo *limbo)
 	return NULL;
 }
 
-struct txn_limbo_entry *
-txn_limbo_append(struct txn_limbo *limbo, uint32_t id, struct txn *txn)
+int
+txn_limbo_submit(struct txn_limbo *limbo, uint32_t id, struct txn *txn)
 {
 	assert(txn_has_flag(txn, TXN_WAIT_SYNC));
 	assert(limbo == &txn_limbo);
@@ -140,13 +289,13 @@ txn_limbo_append(struct txn_limbo *limbo, uint32_t id, struct txn *txn)
 		 * it should be done right now. See in the limbo comments why.
 		 */
 		diag_set(ClientError, ER_SYNC_ROLLBACK);
-		return NULL;
+		return -1;
 	}
 	if (id == 0)
 		id = instance_id;
 	if  (limbo->owner_id == REPLICA_ID_NIL) {
 		diag_set(ClientError, ER_SYNC_QUEUE_UNCLAIMED);
-		return NULL;
+		return -1;
 	} else if (limbo->owner_id != id && !txn_is_fully_local(txn)) {
 		if (txn_limbo_is_empty(limbo)) {
 			diag_set(ClientError, ER_SYNC_QUEUE_FOREIGN,
@@ -155,57 +304,47 @@ txn_limbo_append(struct txn_limbo *limbo, uint32_t id, struct txn *txn)
 			diag_set(ClientError, ER_UNCOMMITTED_FOREIGN_SYNC_TXNS,
 				 limbo->owner_id);
 		}
-		return NULL;
+		return -1;
 	}
 	size_t size;
 	struct txn_limbo_entry *e = region_alloc_object(&txn->region,
 							typeof(*e), &size);
+	if (limbo->entry_to_confirm == NULL &&
+	    txn_has_flag(txn, TXN_WAIT_ACK)) {
+		limbo->entry_to_confirm = e;
+		limbo->ack_count = 0;
+	}
 	if (e == NULL) {
 		diag_set(OutOfMemory, size, "region_alloc_object", "e");
-		return NULL;
+		return -1;
 	}
 	e->txn = txn;
 	e->lsn = -1;
-	e->ack_count = 0;
-	e->is_commit = false;
-	e->is_rollback = false;
+	e->state = TXN_LIMBO_ENTRY_SUBMITTED;
+	txn->limbo_entry = e;
 	rlist_add_tail_entry(&limbo->queue, e, in_queue);
 	limbo->len++;
-	return e;
-}
-
-static inline void
-txn_limbo_remove(struct txn_limbo *limbo, struct txn_limbo_entry *entry)
-{
-	assert(!rlist_empty(&entry->in_queue));
-	assert(txn_limbo_first_entry(limbo) == entry);
-	rlist_del_entry(entry, in_queue);
-	limbo->len--;
-}
-
-static inline void
-txn_limbo_pop(struct txn_limbo *limbo, struct txn_limbo_entry *entry)
-{
-	assert(!rlist_empty(&entry->in_queue));
-	assert(txn_limbo_last_entry(limbo) == entry);
-	assert(entry->is_rollback);
-
-	rlist_del_entry(entry, in_queue);
-	limbo->len--;
-	++limbo->rollback_count;
+	return 0;
 }
 
 void
 txn_limbo_abort(struct txn_limbo *limbo, struct txn_limbo_entry *entry)
 {
-	entry->is_rollback = true;
+	assert(entry->state == TXN_LIMBO_ENTRY_SUBMITTED);
+	assert(!rlist_empty(&entry->in_queue));
 	/*
 	 * The simple rule about rollback/commit order applies
 	 * here as well: commit always in the order of WAL write,
 	 * rollback in the reversed order. Rolled back transaction
 	 * is always the last.
 	 */
-	txn_limbo_pop(limbo, entry);
+	assert(txn_limbo_last_entry(limbo) == entry);
+	entry->state = TXN_LIMBO_ENTRY_ROLLBACK;
+	if (entry == limbo->entry_to_confirm)
+		limbo->entry_to_confirm = NULL;
+	rlist_del_entry(entry, in_queue);
+	limbo->len--;
+	++limbo->rollback_count;
 }
 
 void
@@ -213,7 +352,7 @@ txn_limbo_assign_remote_lsn(struct txn_limbo *limbo,
 			    struct txn_limbo_entry *entry, int64_t lsn)
 {
 	assert(limbo->owner_id != REPLICA_ID_NIL);
-	assert(limbo->owner_id != instance_id);
+	assert(!txn_limbo_is_owned_by_current_instance(limbo));
 	assert(entry->lsn == -1);
 	assert(lsn > 0);
 	(void) limbo;
@@ -227,36 +366,28 @@ txn_limbo_assign_remote_lsn(struct txn_limbo *limbo,
 	entry->lsn = lsn;
 }
 
+static void
+txn_limbo_confirm(struct txn_limbo *limbo);
+
 void
 txn_limbo_assign_local_lsn(struct txn_limbo *limbo,
 			   struct txn_limbo_entry *entry, int64_t lsn)
 {
 	assert(limbo->owner_id != REPLICA_ID_NIL);
-	assert(limbo->owner_id == instance_id);
+	assert(txn_limbo_is_owned_by_current_instance(limbo));
 	assert(entry->lsn == -1);
 	assert(lsn > 0);
 
 	entry->lsn = lsn;
-	/*
-	 * The entry just got its LSN after a WAL write. It could
-	 * happen that this LSN was already ACKed by some
-	 * replicas. Update the ACK counter to take them into
-	 * account.
-	 */
-	struct vclock_iterator iter;
-	vclock_iterator_init(&iter, &limbo->vclock);
-	int ack_count = 0;
-	vclock_foreach(&iter, vc)
-		ack_count += vc.lsn >= lsn;
-	assert(ack_count >= entry->ack_count);
-	entry->ack_count = ack_count;
+	if (entry == limbo->entry_to_confirm)
+		limbo->ack_count = vclock_count_ge(&limbo->vclock, entry->lsn);
 }
 
 void
 txn_limbo_assign_lsn(struct txn_limbo *limbo, struct txn_limbo_entry *entry,
 		     int64_t lsn)
 {
-	if (limbo->owner_id == instance_id)
+	if (txn_limbo_is_owned_by_current_instance(limbo))
 		txn_limbo_assign_local_lsn(limbo, entry, lsn);
 	else
 		txn_limbo_assign_remote_lsn(limbo, entry, lsn);
@@ -282,6 +413,8 @@ txn_limbo_wait_complete(struct txn_limbo *limbo, struct txn_limbo_entry *entry)
 		int rc = fiber_cond_wait_timeout(&limbo->wait_cond, timeout);
 		if (txn_limbo_entry_is_complete(entry))
 			goto complete;
+		if (rc != 0 && fiber_is_cancelled())
+			return -1;
 		if (txn_limbo_is_frozen(limbo))
 			goto wait;
 		if (rc != 0)
@@ -303,7 +436,7 @@ txn_limbo_wait_complete(struct txn_limbo *limbo, struct txn_limbo_entry *entry)
 
 	/* First in the queue is always a synchronous transaction. */
 	assert(entry->lsn > 0);
-	if (entry->lsn <= limbo->confirmed_lsn) {
+	if (entry->lsn <= limbo->volatile_confirmed_lsn) {
 		/*
 		 * Yes, the wait timed out, but there is an on-going CONFIRM WAL
 		 * write in another fiber covering this LSN. Can't rollback it
@@ -318,14 +451,9 @@ txn_limbo_wait_complete(struct txn_limbo *limbo, struct txn_limbo_entry *entry)
 	struct txn_limbo_entry *e, *tmp;
 	rlist_foreach_entry_safe_reverse(e, &limbo->queue,
 					 in_queue, tmp) {
-		e->txn->signature = TXN_SIGNATURE_QUORUM_TIMEOUT;
-		e->txn->limbo_entry = NULL;
-		txn_limbo_abort(limbo, e);
-		txn_clear_flags(e->txn, TXN_WAIT_SYNC | TXN_WAIT_ACK);
-		txn_limbo_complete(e->txn, false);
+		txn_limbo_complete_fail(limbo, e, TXN_SIGNATURE_QUORUM_TIMEOUT);
 		if (e == entry)
 			break;
-		fiber_wakeup(e->txn->fiber);
 	}
 	diag_set(ClientError, ER_SYNC_QUORUM_TIMEOUT);
 	return -1;
@@ -347,7 +475,7 @@ complete:
 	 * The first tx to be rolled back already performed all
 	 * the necessary cleanups for us.
 	 */
-	if (entry->is_rollback) {
+	if (entry->state == TXN_LIMBO_ENTRY_ROLLBACK) {
 		diag_set(ClientError, ER_SYNC_ROLLBACK);
 		return -1;
 	}
@@ -355,44 +483,53 @@ complete:
 }
 
 void
-txn_limbo_checkpoint(const struct txn_limbo *limbo,
-		     struct synchro_request *req)
+txn_limbo_checkpoint(const struct txn_limbo *limbo, struct synchro_request *req,
+		     struct vclock *vclock)
 {
 	req->type = IPROTO_RAFT_PROMOTE;
 	req->replica_id = limbo->owner_id;
 	req->lsn = limbo->confirmed_lsn;
 	req->term = limbo->promote_greatest_term;
+	if (vclock != NULL)
+		vclock_copy(vclock, &limbo->confirmed_vclock);
+	req->confirmed_vclock = vclock;
 }
 
-/** Write a request to WAL. */
-static void
+static int
 synchro_request_write(const struct synchro_request *req)
 {
 	/*
 	 * This is a synchronous commit so we can
 	 * allocate everything on a stack.
 	 */
-	char body[XROW_SYNCHRO_BODY_LEN_MAX];
+	char body[XROW_BODY_LEN_MAX];
 	struct xrow_header row;
-	char buf[sizeof(struct journal_entry) +
-		 sizeof(struct xrow_header *)];
-
-	struct journal_entry *entry = (struct journal_entry *)buf;
-	entry->rows[0] = &row;
-
 	xrow_encode_synchro(&row, body, req);
+	return journal_write_row(&row);
+}
 
-	journal_entry_create(entry, 1, xrow_approx_len(&row),
-			     journal_entry_fiber_wakeup_cb, fiber());
+static int
+txn_limbo_write_synchro(struct txn_limbo *limbo, uint16_t type, int64_t lsn,
+			uint64_t term, struct vclock *vclock)
+{
+	assert(lsn >= 0);
 
-	if (journal_write(entry) != 0)
-		goto fail;
-	if (entry->res < 0) {
-		diag_set_journal_res(entry->res);
-		goto fail;
-	}
-	return;
-fail:
+	struct synchro_request req = {
+		.type = type,
+		.replica_id = limbo->owner_id,
+		.lsn = lsn,
+		.term = term,
+		.confirmed_vclock = vclock,
+	};
+	return synchro_request_write(&req);
+}
+
+/** Write a request to WAL or panic. */
+static void
+synchro_request_write_or_panic(const struct synchro_request *req)
+{
+	if (synchro_request_write(req) == 0)
+		return;
 	diag_log();
 	/*
 	 * XXX: the stub is supposed to be removed once it is defined what to do
@@ -405,10 +542,11 @@ fail:
 	      "type = %s\n", (long long)req->lsn, iproto_type_name(req->type));
 }
 
-/** Create a request for a specific limbo and write it to WAL. */
+/** Create a request for a specific limbo and write it to WAL or panic. */
 static void
-txn_limbo_write_synchro(struct txn_limbo *limbo, uint16_t type, int64_t lsn,
-			uint64_t term)
+txn_limbo_write_synchro_or_panic(struct txn_limbo *limbo, uint16_t type,
+				 int64_t lsn, uint64_t term,
+				 struct vclock *vclock)
 {
 	assert(lsn >= 0);
 
@@ -417,29 +555,17 @@ txn_limbo_write_synchro(struct txn_limbo *limbo, uint16_t type, int64_t lsn,
 		.replica_id = limbo->owner_id,
 		.lsn = lsn,
 		.term = term,
+		.confirmed_vclock = vclock,
 	};
-	synchro_request_write(&req);
+	synchro_request_write_or_panic(&req);
 }
 
-/**
- * Write a confirmation entry to WAL. After it's written all the
- * transactions waiting for confirmation may be finished.
- */
-static void
-txn_limbo_write_confirm(struct txn_limbo *limbo, int64_t lsn)
-{
-	assert(lsn > limbo->confirmed_lsn);
-	assert(!limbo->is_in_rollback);
-	limbo->confirmed_lsn = lsn;
-	txn_limbo_write_synchro(limbo, IPROTO_RAFT_CONFIRM, lsn, 0);
-}
-
-/** Confirm all the entries <= @a lsn. */
 static void
 txn_limbo_read_confirm(struct txn_limbo *limbo, int64_t lsn)
 {
 	assert(limbo->owner_id != REPLICA_ID_NIL || txn_limbo_is_empty(limbo));
 	assert(limbo == &txn_limbo);
+	assert(limbo->confirmed_lsn <= lsn);
 	struct txn_limbo_entry *e, *tmp;
 	rlist_foreach_entry_safe(e, &limbo->queue, in_queue, tmp) {
 		/*
@@ -458,6 +584,22 @@ txn_limbo_read_confirm(struct txn_limbo *limbo, int64_t lsn)
 			 */
 			if (e->lsn == -1)
 				break;
+			if (!rlist_empty(&e->txn->on_commit)) {
+				/*
+				 * Bump the confirmed LSN right now, do not
+				 * batch with any newer txns. So on-commit
+				 * triggers would see the confirmation LSN
+				 * matching this txn exactly. Making an illusion
+				 * like each txn has its own confirmation.
+				 */
+				if (limbo->confirmed_lsn < e->lsn) {
+					limbo->confirmed_lsn = e->lsn;
+					vclock_follow(&limbo->confirmed_vclock,
+						      limbo->owner_id, e->lsn);
+				} else {
+					assert(limbo->confirmed_lsn == lsn);
+				}
+			}
 		} else if (e->txn->signature == TXN_SIGNATURE_UNKNOWN) {
 			/*
 			 * A transaction might be covered by the CONFIRM even if
@@ -466,6 +608,7 @@ txn_limbo_read_confirm(struct txn_limbo *limbo, int64_t lsn)
 			 * CONFIRM was being written to WAL.
 			 */
 			assert(e->txn->status == TXN_PREPARED);
+			assert(e->state == TXN_LIMBO_ENTRY_SUBMITTED);
 			/*
 			 * Let it complete normally as a plain transaction. It
 			 * is important to remove the limbo entry, because the
@@ -476,7 +619,7 @@ txn_limbo_read_confirm(struct txn_limbo *limbo, int64_t lsn)
 			 * created in the applier.
 			 */
 			txn_clear_flags(e->txn, TXN_WAIT_SYNC);
-			txn_limbo_remove(limbo, e);
+			txn_limbo_pop_first(limbo, e);
 			/*
 			 * The limbo entry now should not be used by the owner
 			 * transaction since it just became a plain one. Nullify
@@ -487,24 +630,21 @@ txn_limbo_read_confirm(struct txn_limbo *limbo, int64_t lsn)
 			e->txn = NULL;
 			continue;
 		}
-		e->is_commit = true;
-		e->txn->limbo_entry = NULL;
-		txn_limbo_remove(limbo, e);
-		txn_clear_flags(e->txn, TXN_WAIT_SYNC | TXN_WAIT_ACK);
-		/*
-		 * Should be written to WAL by now. Confirm is always written
-		 * after the affected transactions.
-		 */
-		assert(e->txn->signature >= 0);
-		txn_limbo_complete(e->txn, true);
+		txn_limbo_complete_success(limbo, e);
 	}
-	/*
-	 * Track CONFIRM lsn on replica in order to detect split-brain by
-	 * comparing existing confirm_lsn with the one arriving from a remote
-	 * instance.
-	 */
-	if (limbo->confirmed_lsn < lsn)
+	if (limbo->confirmed_lsn < lsn) {
 		limbo->confirmed_lsn = lsn;
+		vclock_follow(&limbo->confirmed_vclock, limbo->owner_id, lsn);
+	}
+}
+
+/** Confirm an LSN in the limbo. */
+static void
+txn_limbo_confirm_lsn(struct txn_limbo *limbo, int64_t confirm_lsn)
+{
+	assert(confirm_lsn > limbo->volatile_confirmed_lsn);
+	limbo->volatile_confirmed_lsn = confirm_lsn;
+	fiber_wakeup(limbo->worker);
 }
 
 /**
@@ -518,7 +658,8 @@ txn_limbo_write_rollback(struct txn_limbo *limbo, int64_t lsn)
 	assert(lsn > limbo->confirmed_lsn);
 	assert(!limbo->is_in_rollback);
 	limbo->is_in_rollback = true;
-	txn_limbo_write_synchro(limbo, IPROTO_RAFT_ROLLBACK, lsn, 0);
+	txn_limbo_write_synchro_or_panic(limbo, IPROTO_RAFT_ROLLBACK, lsn, 0,
+					 NULL);
 	limbo->is_in_rollback = false;
 }
 
@@ -540,16 +681,12 @@ txn_limbo_read_rollback(struct txn_limbo *limbo, int64_t lsn)
 	if (last_rollback == NULL)
 		return;
 	rlist_foreach_entry_safe_reverse(e, &limbo->queue, in_queue, tmp) {
-		txn_limbo_abort(limbo, e);
-		txn_clear_flags(e->txn, TXN_WAIT_ACK);
 		/*
 		 * Should be written to WAL by now. Rollback is always written
 		 * after the affected transactions.
 		 */
 		assert(e->txn->signature >= 0);
-		e->txn->signature = TXN_SIGNATURE_SYNC_ROLLBACK;
-		e->txn->limbo_entry = NULL;
-		txn_limbo_complete(e->txn, false);
+		txn_limbo_complete_fail(limbo, e, TXN_SIGNATURE_SYNC_ROLLBACK);
 		if (e == last_rollback)
 			break;
 	}
@@ -572,10 +709,15 @@ txn_limbo_write_promote(struct txn_limbo *limbo, int64_t lsn, uint64_t term)
 		.origin_id = instance_id,
 		.lsn = lsn,
 		.term = term,
+		/*
+		 * Confirmed_vclock is only persisted in checkpoints. It doesn't
+		 * appear in WALs and replication.
+		 */
+		.confirmed_vclock = NULL,
 	};
 	if (txn_limbo_req_prepare(limbo, &req) < 0)
 		return -1;
-	synchro_request_write(&req);
+	synchro_request_write_or_panic(&req);
 	txn_limbo_req_commit(limbo, &req);
 	return 0;
 }
@@ -586,19 +728,17 @@ txn_limbo_write_promote(struct txn_limbo *limbo, int64_t lsn, uint64_t term)
  */
 static void
 txn_limbo_read_promote(struct txn_limbo *limbo, uint32_t replica_id,
-		       uint32_t prev_id, int64_t lsn)
+		       int64_t lsn)
 {
 	txn_limbo_read_confirm(limbo, lsn);
 	txn_limbo_read_rollback(limbo, lsn + 1);
 	assert(txn_limbo_is_empty(limbo));
 	limbo->owner_id = replica_id;
+	limbo->confirmed_lsn = vclock_get(&limbo->confirmed_vclock,
+					  replica_id);
+	limbo->volatile_confirmed_lsn = limbo->confirmed_lsn;
+	limbo->entry_to_confirm = NULL;
 	box_update_ro_summary();
-	/*
-	 * Only nullify confirmed_lsn when the new value is unknown. I.e. when
-	 * prev_id != replica_id.
-	 */
-	if (replica_id != prev_id)
-		limbo->confirmed_lsn = 0;
 }
 
 int
@@ -614,10 +754,11 @@ txn_limbo_write_demote(struct txn_limbo *limbo, int64_t lsn, uint64_t term)
 		.origin_id = instance_id,
 		.lsn = lsn,
 		.term = term,
+		.confirmed_vclock = NULL,
 	};
 	if (txn_limbo_req_prepare(limbo, &req) < 0)
 		return -1;
-	synchro_request_write(&req);
+	synchro_request_write_or_panic(&req);
 	txn_limbo_req_commit(limbo, &req);
 	return 0;
 }
@@ -628,9 +769,58 @@ txn_limbo_write_demote(struct txn_limbo *limbo, int64_t lsn, uint64_t term)
  * @sa txn_limbo_read_promote.
  */
 static void
-txn_limbo_read_demote(struct txn_limbo *limbo, uint32_t prev_id, int64_t lsn)
+txn_limbo_read_demote(struct txn_limbo *limbo, int64_t lsn)
 {
-	return txn_limbo_read_promote(limbo, REPLICA_ID_NIL, prev_id, lsn);
+	return txn_limbo_read_promote(limbo, REPLICA_ID_NIL, lsn);
+}
+
+/**
+ * Check that some synchronous transactions have gathered quorum and
+ * write a confirmation entry of the last confirmed transaction.
+ */
+static void
+txn_limbo_confirm(struct txn_limbo *limbo)
+{
+	assert(txn_limbo_is_owned_by_current_instance(limbo));
+	if (limbo->is_in_rollback)
+		return;
+	if (limbo->entry_to_confirm == NULL ||
+	    limbo->entry_to_confirm->lsn == -1)
+		return;
+	if (limbo->ack_count < replication_synchro_quorum)
+		return;
+	int32_t k = (int32_t)vclock_size(&limbo->vclock)
+		    - replication_synchro_quorum;
+	/**
+	 * limbo->ack_count >= replication_synchro_quorum =>
+	 * vclock_size(&limbo->vclock) >= replication_synchro_quorum
+	 */
+	assert(k >= 0);
+	int64_t confirm_lsn = vclock_nth_element(&limbo->vclock, k);
+	assert(confirm_lsn >= limbo->entry_to_confirm->lsn);
+	struct txn_limbo_entry *e = limbo->entry_to_confirm;
+	limbo->entry_to_confirm = NULL;
+	int64_t max_assigned_lsn = -1;
+	for (; !rlist_entry_is_head(e, &limbo->queue, in_queue);
+	       e = rlist_next_entry(e, in_queue)) {
+		if (!txn_has_flag(e->txn, TXN_WAIT_ACK))
+			continue;
+		if (e->lsn == -1 || e->lsn > confirm_lsn) {
+			limbo->entry_to_confirm = e;
+			/**
+			 * It may be that a quorum has been gathered, but
+			 * ack_count = 0. It's ok. CONFIRM will be written as
+			 * soon as the lsn is assigned to the transaction.
+			 */
+			limbo->ack_count = (e->lsn == -1) ? 0 :
+				vclock_count_ge(&limbo->vclock, e->lsn);
+			break;
+		} else {
+			max_assigned_lsn = e->lsn;
+		}
+	}
+	assert(max_assigned_lsn != -1);
+	txn_limbo_confirm_lsn(limbo, max_assigned_lsn);
 }
 
 void
@@ -655,9 +845,10 @@ txn_limbo_ack(struct txn_limbo *limbo, uint32_t replica_id, int64_t lsn)
 		return;
 	assert(limbo->owner_id != REPLICA_ID_NIL);
 	int64_t prev_lsn = vclock_get(&limbo->vclock, replica_id);
+	assert(lsn >= prev_lsn);
 	/*
 	 * One of the reasons why can happen - the remote instance is not
-	 * read-only and wrote something under its own insance_id. For qsync
+	 * read-only and wrote something under its own instance_id. For qsync
 	 * that most likely means that the remote instance decided to take over
 	 * the limbo ownership, and the current node is going to become a
 	 * replica very soon.
@@ -665,31 +856,15 @@ txn_limbo_ack(struct txn_limbo *limbo, uint32_t replica_id, int64_t lsn)
 	if (lsn == prev_lsn)
 		return;
 	vclock_follow(&limbo->vclock, replica_id, lsn);
-	struct txn_limbo_entry *e;
-	int64_t confirm_lsn = -1;
-	rlist_foreach_entry(e, &limbo->queue, in_queue) {
-		assert(e->ack_count <= VCLOCK_MAX);
-		if (e->lsn > lsn)
-			break;
-		/*
-		 * Sync transactions need to collect acks. Async
-		 * transactions are automatically committed right
-		 * after all the previous sync transactions are.
-		 */
-		if (!txn_has_flag(e->txn, TXN_WAIT_ACK)) {
-			continue;
-		} else if (e->lsn <= prev_lsn) {
-			continue;
-		} else if (++e->ack_count < replication_synchro_quorum) {
-			continue;
-		} else {
-			confirm_lsn = e->lsn;
-		}
+
+	if (limbo->entry_to_confirm != NULL &&
+	    limbo->entry_to_confirm->lsn != -1) {
+		if (limbo->entry_to_confirm->lsn <= prev_lsn ||
+		    lsn < limbo->entry_to_confirm->lsn)
+			return;
+		if (++limbo->ack_count >= replication_synchro_quorum)
+			txn_limbo_confirm(limbo);
 	}
-	if (confirm_lsn == -1 || confirm_lsn <= limbo->confirmed_lsn)
-		return;
-	txn_limbo_write_confirm(limbo, confirm_lsn);
-	txn_limbo_read_confirm(limbo, confirm_lsn);
 }
 
 /**
@@ -901,90 +1076,6 @@ txn_limbo_filter_generic(struct txn_limbo *limbo,
 	return 0;
 }
 
-/**
- * A common filter for all synchro requests, checking that request operates
- * over a valid lsn range.
- */
-static int
-txn_limbo_filter_queue_boundaries(struct txn_limbo *limbo,
-				  const struct synchro_request *req)
-{
-	int64_t lsn = req->lsn;
-	/*
-	 * Easy case - processed LSN matches the new one which comes inside
-	 * request, everything is consistent. This is allowed only for
-	 * PROMOTE/DEMOTE.
-	 */
-	if (limbo->confirmed_lsn == lsn) {
-		if (iproto_type_is_promote_request(req->type)) {
-			return 0;
-		} else {
-			say_error("%s. Duplicate request with confirmed lsn "
-				  "%lld = request lsn %lld", reject_str(req),
-				  (long long)limbo->confirmed_lsn,
-				  (long long)lsn);
-			diag_set(ClientError, ER_UNSUPPORTED, "Replication",
-				 "Duplicate CONFIRM/ROLLBACK request");
-			return -1;
-		}
-	}
-
-	/*
-	 * Explicit split brain situation. Request comes in with an old LSN
-	 * which we've already processed.
-	 */
-	if (limbo->confirmed_lsn > lsn) {
-		say_error("%s. confirmed lsn %lld > request lsn %lld",
-			  reject_str(req), (long long)limbo->confirmed_lsn,
-			  (long long)lsn);
-		diag_set(ClientError, ER_SPLIT_BRAIN,
-			 "got a request with lsn from an already "
-			 "processed range");
-		return -1;
-	}
-
-	/*
-	 * The last case requires a few subcases.
-	 */
-	assert(limbo->confirmed_lsn < lsn);
-
-	if (txn_limbo_is_empty(limbo)) {
-		/*
-		 * Transactions are rolled back already,
-		 * since the limbo is empty.
-		 */
-		say_error("%s. confirmed lsn %lld < request lsn %lld "
-			  "and empty limbo", reject_str(req),
-			  (long long)limbo->confirmed_lsn,
-			  (long long)lsn);
-		diag_set(ClientError, ER_SPLIT_BRAIN,
-			 "got a request mentioning future lsn");
-		return -1;
-	} else {
-		/*
-		 * Some entries are present in the limbo, we need to make sure
-		 * that request lsn lays inside limbo [first; last] range.
-		 * So that the request has some queued data to process,
-		 * otherwise it means the request comes from split brained node.
-		 */
-		int64_t first_lsn = txn_limbo_first_entry(limbo)->lsn;
-		int64_t last_lsn = txn_limbo_last_synchro_entry(limbo)->lsn;
-
-		if (lsn < first_lsn || last_lsn < lsn) {
-			say_error("%s. request lsn %lld out of range "
-				  "[%lld; %lld]", reject_str(req),
-				  (long long)lsn,
-				  (long long)first_lsn,
-				  (long long)last_lsn);
-			diag_set(ClientError, ER_SPLIT_BRAIN,
-				 "got a request lsn out of queue range");
-			return -1;
-		}
-	}
-
-	return 0;
-}
-
 /**
  * Filter CONFIRM and ROLLBACK packets.
  */
@@ -994,6 +1085,7 @@ txn_limbo_filter_confirm_rollback(struct txn_limbo *limbo,
 {
 	assert(latch_is_locked(&limbo->promote_latch));
 	assert(limbo->do_validate);
+	(void)limbo;
 	assert(req->type == IPROTO_RAFT_CONFIRM ||
 	       req->type == IPROTO_RAFT_ROLLBACK);
 	/*
@@ -1005,8 +1097,7 @@ txn_limbo_filter_confirm_rollback(struct txn_limbo *limbo,
 			 "zero LSN for CONFIRM/ROLLBACK");
 		return -1;
 	}
-
-	return txn_limbo_filter_queue_boundaries(limbo, req);
+	return 0;
 }
 
 /** A filter PROMOTE and DEMOTE packets. */
@@ -1041,8 +1132,61 @@ txn_limbo_filter_promote_demote(struct txn_limbo *limbo,
 			 "got a PROMOTE/DEMOTE with an obsolete term");
 		return -1;
 	}
-
-	return txn_limbo_filter_queue_boundaries(limbo, req);
+	/*
+	 * Explicit split brain situation. Request comes in with an old LSN
+	 * which we've already processed.
+	 */
+	if (limbo->confirmed_lsn > req->lsn) {
+		say_error("%s. confirmed lsn %lld > request lsn %lld",
+			  reject_str(req), (long long)limbo->confirmed_lsn,
+			  (long long)req->lsn);
+		diag_set(ClientError, ER_SPLIT_BRAIN,
+			 "got a request with lsn from an already "
+			 "processed range");
+		return -1;
+	}
+	/*
+	 * Easy case - processed LSN matches the new one which comes inside
+	 * request, everything is consistent. This is allowed only for
+	 * PROMOTE/DEMOTE.
+	 */
+	if (limbo->confirmed_lsn == req->lsn)
+		return 0;
+	/*
+	 * The last case requires a few subcases.
+	 */
+	if (txn_limbo_is_empty(limbo)) {
+		/*
+		 * Transactions are rolled back already,
+		 * since the limbo is empty.
+		 */
+		say_error("%s. confirmed lsn %lld < request lsn %lld "
+			  "and empty limbo", reject_str(req),
+			  (long long)limbo->confirmed_lsn,
+			  (long long)req->lsn);
+		diag_set(ClientError, ER_SPLIT_BRAIN,
+			 "got a request mentioning future lsn");
+		return -1;
+	}
+	/*
+	 * Some entries are present in the limbo, we need to make sure that
+	 * request lsn lays inside limbo [first; last] range. So that the
+	 * request has some queued data to process, otherwise it means the
+	 * request comes from split brained node.
+	 */
+	int64_t first_lsn = txn_limbo_first_entry(limbo)->lsn;
+	int64_t last_lsn = txn_limbo_last_synchro_entry(limbo)->lsn;
+	if (req->lsn < first_lsn || last_lsn < req->lsn) {
+		say_error("%s. request lsn %lld out of range "
+			  "[%lld; %lld]", reject_str(req),
+			  (long long)req->lsn,
+			  (long long)first_lsn,
+			  (long long)last_lsn);
+		diag_set(ClientError, ER_SPLIT_BRAIN,
+			 "got a request lsn out of queue range");
+		return -1;
+	}
+	return 0;
 }
 
 /** A fine-grained filter checking specific request type constraints. */
@@ -1106,8 +1250,8 @@ txn_limbo_req_prepare(struct txn_limbo *limbo,
 	case IPROTO_RAFT_PROMOTE:
 	case IPROTO_RAFT_DEMOTE: {
 		assert(limbo->svp_confirmed_lsn == -1);
-		limbo->svp_confirmed_lsn = limbo->confirmed_lsn;
-		limbo->confirmed_lsn = req->lsn;
+		limbo->svp_confirmed_lsn = limbo->volatile_confirmed_lsn;
+		limbo->volatile_confirmed_lsn = req->lsn;
 		break;
 	}
 	/*
@@ -1128,7 +1272,7 @@ txn_limbo_req_rollback(struct txn_limbo *limbo,
 	case IPROTO_RAFT_DEMOTE: {
 		assert(limbo->is_in_rollback);
 		assert(limbo->svp_confirmed_lsn >= 0);
-		limbo->confirmed_lsn = limbo->svp_confirmed_lsn;
+		limbo->volatile_confirmed_lsn = limbo->svp_confirmed_lsn;
 		limbo->svp_confirmed_lsn = -1;
 		limbo->is_in_rollback = false;
 		break;
@@ -1184,6 +1328,8 @@ txn_limbo_req_commit(struct txn_limbo *limbo, const struct synchro_request *req)
 			}
 		}
 	}
+	if (req->confirmed_vclock != NULL)
+		vclock_copy(&limbo->confirmed_vclock, req->confirmed_vclock);
 
 	int64_t lsn = req->lsn;
 	switch (req->type) {
@@ -1194,11 +1340,10 @@ txn_limbo_req_commit(struct txn_limbo *limbo, const struct synchro_request *req)
 		txn_limbo_read_rollback(limbo, lsn);
 		break;
 	case IPROTO_RAFT_PROMOTE:
-		txn_limbo_read_promote(limbo, req->origin_id, req->replica_id,
-				       lsn);
+		txn_limbo_read_promote(limbo, req->origin_id, lsn);
 		break;
 	case IPROTO_RAFT_DEMOTE:
-		txn_limbo_read_demote(limbo, req->replica_id, lsn);
+		txn_limbo_read_demote(limbo, lsn);
 		break;
 	default:
 		unreachable();
@@ -1224,23 +1369,9 @@ txn_limbo_on_parameters_change(struct txn_limbo *limbo)
 {
 	if (rlist_empty(&limbo->queue) || txn_limbo_is_frozen(limbo))
 		return;
-	struct txn_limbo_entry *e;
-	int64_t confirm_lsn = -1;
-	rlist_foreach_entry(e, &limbo->queue, in_queue) {
-		assert(e->ack_count <= VCLOCK_MAX);
-		if (!txn_has_flag(e->txn, TXN_WAIT_ACK)) {
-			continue;
-		} else if (e->ack_count < replication_synchro_quorum) {
-			continue;
-		} else {
-			confirm_lsn = e->lsn;
-			assert(confirm_lsn > 0);
-		}
-	}
-	if (confirm_lsn > limbo->confirmed_lsn && !limbo->is_in_rollback) {
-		txn_limbo_write_confirm(limbo, confirm_lsn);
-		txn_limbo_read_confirm(limbo, confirm_lsn);
-	}
+	/* The replication_synchro_quorum value may have changed. */
+	if (txn_limbo_is_owned_by_current_instance(limbo))
+		txn_limbo_confirm(limbo);
 	/*
 	 * Wakeup all the others - timed out will rollback. Also
 	 * there can be non-transactional waiters, such as CONFIRM
@@ -1273,8 +1404,22 @@ txn_limbo_filter_enable(struct txn_limbo *limbo)
 	latch_unlock(&limbo->promote_latch);
 }
 
+void
+txn_limbo_filter_disable(struct txn_limbo *limbo)
+{
+	latch_lock(&limbo->promote_latch);
+	limbo->do_validate = false;
+	latch_unlock(&limbo->promote_latch);
+}
+
 void
 txn_limbo_init(void)
 {
 	txn_limbo_create(&txn_limbo);
 }
+
+void
+txn_limbo_free(void)
+{
+	txn_limbo_destroy(&txn_limbo);
+}
