@@ -52,6 +52,20 @@ enum {
 	VY_CACHE_CLEANUP_MAX_STEPS = 10,
 };
 
+static void *
+vy_cache_tree_page_alloc(struct matras_allocator *allocator)
+{
+	(void)allocator;
+	return xmalloc(VY_CACHE_TREE_EXTENT_SIZE);
+}
+
+static void
+vy_cache_tree_page_free(struct matras_allocator *allocator, void *ptr)
+{
+	(void)allocator;
+	free(ptr);
+}
+
 void
 vy_cache_env_create(struct vy_cache_env *e, struct slab_cache *slab_cache)
 {
@@ -60,11 +74,25 @@ vy_cache_env_create(struct vy_cache_env *e, struct slab_cache *slab_cache)
 	e->mem_quota = 0;
 	mempool_create(&e->cache_node_mempool, slab_cache,
 		       sizeof(struct vy_cache_node));
+	matras_allocator_create(&e->allocator,
+				VY_CACHE_TREE_EXTENT_SIZE,
+				vy_cache_tree_page_alloc,
+				vy_cache_tree_page_free);
 }
 
+/** Delete a node from the cache. */
+static void
+vy_cache_gc_step(struct vy_cache_env *env);
+
 void
 vy_cache_env_destroy(struct vy_cache_env *e)
 {
+#if ENABLE_ASAN
+	/* Purge cache to suppress leak detector. */
+	while (e->mem_used > 0)
+		vy_cache_gc_step(e);
+#endif
+	matras_allocator_destroy(&e->allocator);
 	mempool_destroy(&e->cache_node_mempool);
 }
 
@@ -95,6 +123,8 @@ vy_cache_node_new(struct vy_cache_env *env, struct vy_cache *cache,
 	node->flags = 0;
 	node->left_boundary_level = cache->cmp_def->part_count;
 	node->right_boundary_level = cache->cmp_def->part_count;
+	node->left_lsn = 0;
+	node->right_lsn = 0;
 	rlist_add(&env->cache_lru, &node->in_lru);
 	env->mem_used += vy_cache_node_size(node);
 	vy_stmt_counter_acct_tuple(&cache->stat.count, entry.stmt);
@@ -114,26 +144,6 @@ vy_cache_node_delete(struct vy_cache_env *env, struct vy_cache_node *node)
 	mempool_free(&env->cache_node_mempool, node);
 }
 
-static void *
-vy_cache_tree_page_alloc(void *ctx)
-{
-	struct vy_env *env = (struct vy_env *)ctx;
-	(void)env;
-	void *ret = malloc(VY_CACHE_TREE_EXTENT_SIZE);
-	if (ret == NULL)
-		diag_set(OutOfMemory, VY_CACHE_TREE_EXTENT_SIZE, "malloc",
-			 "ret");
-	return ret;
-}
-
-static void
-vy_cache_tree_page_free(void *ctx, void *p)
-{
-	struct vy_env *env = (struct vy_env *)ctx;
-	(void)env;
-	free(p);
-}
-
 void
 vy_cache_create(struct vy_cache *cache, struct vy_cache_env *env,
 		struct key_def *cmp_def, bool is_primary)
@@ -142,9 +152,7 @@ vy_cache_create(struct vy_cache *cache, struct vy_cache_env *env,
 	cache->cmp_def = cmp_def;
 	cache->is_primary = is_primary;
 	cache->version = 1;
-	vy_cache_tree_create(&cache->cache_tree, cmp_def,
-			     vy_cache_tree_page_alloc,
-			     vy_cache_tree_page_free, env);
+	vy_cache_tree_create(&cache->cache_tree, cmp_def, &env->allocator);
 }
 
 void
@@ -226,7 +234,7 @@ vy_cache_env_set_quota(struct vy_cache_env *env, size_t quota)
 
 void
 vy_cache_add(struct vy_cache *cache, struct vy_entry curr,
-	     struct vy_entry prev, bool is_first,
+	     struct vy_entry prev, bool is_first, int64_t link_lsn,
 	     struct vy_entry key, enum iterator_type order)
 {
 	assert(!is_first || prev.stmt == NULL);
@@ -319,12 +327,21 @@ vy_cache_add(struct vy_cache *cache, struct vy_entry curr,
 		node->flags = replaced->flags;
 		node->left_boundary_level = replaced->left_boundary_level;
 		node->right_boundary_level = replaced->right_boundary_level;
+		node->left_lsn = replaced->left_lsn;
+		node->right_lsn = replaced->right_lsn;
 		vy_cache_node_delete(cache->env, replaced);
 	}
-	if (direction > 0 && boundary_level < node->left_boundary_level)
+	if (direction > 0 &&
+	    !(node->flags & VY_CACHE_LEFT_LINKED) &&
+	    boundary_level < node->left_boundary_level) {
 		node->left_boundary_level = boundary_level;
-	else if (direction < 0 && boundary_level < node->right_boundary_level)
+		node->left_lsn = link_lsn;
+	} else if (direction < 0 &&
+		   !(node->flags & VY_CACHE_RIGHT_LINKED) &&
+		   boundary_level < node->right_boundary_level) {
 		node->right_boundary_level = boundary_level;
+		node->right_lsn = link_lsn;
+	}
 
 	vy_stmt_counter_acct_tuple(&cache->stat.put, curr.stmt);
 
@@ -405,6 +422,8 @@ vy_cache_add(struct vy_cache *cache, struct vy_entry curr,
 		prev_node->flags = replaced->flags;
 		prev_node->left_boundary_level = replaced->left_boundary_level;
 		prev_node->right_boundary_level = replaced->right_boundary_level;
+		prev_node->left_lsn = replaced->left_lsn;
+		prev_node->right_lsn = replaced->right_lsn;
 		vy_cache_node_delete(cache->env, replaced);
 	}
 
@@ -413,6 +432,14 @@ vy_cache_add(struct vy_cache *cache, struct vy_entry curr,
 	/* Set inverted flag in the previous node */
 	prev_node->flags |= (VY_CACHE_LEFT_LINKED |
 			     VY_CACHE_RIGHT_LINKED) ^ flag;
+	/* Set chain link LSN when two nodes are linked */
+	if (direction > 0) {
+		node->left_lsn = link_lsn;
+		prev_node->right_lsn = link_lsn;
+	} else {
+		node->right_lsn = link_lsn;
+		prev_node->left_lsn = link_lsn;
+	}
 }
 
 struct vy_entry
@@ -425,9 +452,12 @@ vy_cache_get(struct vy_cache *cache, struct vy_entry key)
 	return (*node)->entry;
 }
 
-void
-vy_cache_on_write(struct vy_cache *cache, struct vy_entry entry,
-		  struct vy_entry *deleted)
+/**
+ * Invalidate the cache when a statement is committed or rolled back.
+ */
+static void
+vy_cache_invalidate(struct vy_cache *cache, struct vy_entry entry,
+		    bool rollback, struct vy_entry *deleted)
 {
 	vy_cache_gc(cache->env);
 	bool exact = false;
@@ -448,7 +478,7 @@ vy_cache_on_write(struct vy_cache *cache, struct vy_entry entry,
 	 *   ('exact' == false, 'node' == NULL)
 	 */
 
-	if (vy_stmt_type(entry.stmt) == IPROTO_DELETE && !exact) {
+	if (!rollback && vy_stmt_type(entry.stmt) == IPROTO_DELETE && !exact) {
 		/* there was nothing and there is nothing now */
 		return;
 	}
@@ -502,6 +532,19 @@ vy_cache_on_write(struct vy_cache *cache, struct vy_entry entry,
 	}
 }
 
+void
+vy_cache_on_write(struct vy_cache *cache, struct vy_entry entry,
+		  struct vy_entry *deleted)
+{
+	vy_cache_invalidate(cache, entry, /*rollback=*/false, deleted);
+}
+
+void
+vy_cache_on_rollback(struct vy_cache *cache, struct vy_entry entry)
+{
+	vy_cache_invalidate(cache, entry, /*rollback=*/true, NULL);
+}
+
 /**
  * Get a stmt by current position
  */
@@ -514,6 +557,29 @@ vy_cache_iterator_curr(struct vy_cache_iterator *itr)
 	return node ? (*node)->entry : vy_entry_none();
 }
 
+/**
+ * Return true if the given LSN is visible from the iterator read view.
+ */
+static bool
+vy_cache_iterator_lsn_is_visible(struct vy_cache_iterator *itr, int64_t lsn)
+{
+	if (lsn > (**itr->read_view).vlsn)
+		return false;
+	if (!itr->is_prepared_ok && vy_lsn_is_prepared(lsn))
+		return false;
+	return true;
+}
+
+/**
+ * Return true if the given statement is visible from the iterator read view.
+ */
+static bool
+vy_cache_iterator_stmt_is_visible(struct vy_cache_iterator *itr,
+				  struct tuple *stmt)
+{
+	return vy_cache_iterator_lsn_is_visible(itr, vy_stmt_lsn(stmt));
+}
+
 /**
  * Determine whether the merge iterator must be stopped or not.
  * That is made by examining flags of a cache record.
@@ -530,15 +596,15 @@ vy_cache_iterator_is_stop(struct vy_cache_iterator *itr,
 	/* select{} is actually an EQ iterator with part_count == 0 */
 	bool iter_is_eq = itr->iterator_type == ITER_EQ || key_level == 0;
 	if (iterator_direction(itr->iterator_type) > 0) {
-		if (node->flags & VY_CACHE_LEFT_LINKED)
-			return true;
-		if (iter_is_eq && node->left_boundary_level <= key_level)
-			return true;
+		if ((node->flags & VY_CACHE_LEFT_LINKED) ||
+		    (iter_is_eq && node->left_boundary_level <= key_level))
+			return vy_cache_iterator_lsn_is_visible(
+						itr, node->left_lsn);
 	} else {
-		if (node->flags & VY_CACHE_RIGHT_LINKED)
-			return true;
-		if (iter_is_eq && node->right_boundary_level <= key_level)
-			return true;
+		if ((node->flags & VY_CACHE_RIGHT_LINKED) ||
+		    (iter_is_eq && node->right_boundary_level <= key_level))
+			return vy_cache_iterator_lsn_is_visible(
+						itr, node->right_lsn);
 	}
 	return false;
 }
@@ -560,15 +626,17 @@ vy_cache_iterator_is_end_stop(struct vy_cache_iterator *itr,
 	/* select{} is actually an EQ iterator with part_count == 0 */
 	bool iter_is_eq = itr->iterator_type == ITER_EQ || key_level == 0;
 	if (iterator_direction(itr->iterator_type) > 0) {
-		if (last_node->flags & VY_CACHE_RIGHT_LINKED)
-			return true;
-		if (iter_is_eq && last_node->right_boundary_level <= key_level)
-			return true;
+		if ((last_node->flags & VY_CACHE_RIGHT_LINKED) ||
+		    (iter_is_eq &&
+		     last_node->right_boundary_level <= key_level))
+			return vy_cache_iterator_lsn_is_visible(
+						itr, last_node->right_lsn);
 	} else {
-		if (last_node->flags & VY_CACHE_LEFT_LINKED)
-			return true;
-		if (iter_is_eq && last_node->left_boundary_level <= key_level)
-			return true;
+		if ((last_node->flags & VY_CACHE_LEFT_LINKED) ||
+		    (iter_is_eq &&
+		     last_node->left_boundary_level <= key_level))
+			return vy_cache_iterator_lsn_is_visible(
+						itr, last_node->left_lsn);
 	}
 	return false;
 }
@@ -614,21 +682,6 @@ vy_cache_iterator_step(struct vy_cache_iterator *itr)
 	return vy_cache_iterator_is_stop(itr, node);
 }
 
-/**
- * Return true if the given statement is visible from the iterator
- * read view.
- */
-static inline bool
-vy_cache_iterator_stmt_is_visible(struct vy_cache_iterator *itr,
-				  struct tuple *stmt)
-{
-	if (vy_stmt_lsn(stmt) > (**itr->read_view).vlsn)
-		return false;
-	if (!itr->is_prepared_ok && vy_stmt_is_prepared(stmt))
-		return false;
-	return true;
-}
-
 /**
  * Skip all statements that are invisible in the read view
  * associated with the iterator.
@@ -817,10 +870,13 @@ vy_cache_iterator_restore(struct vy_cache_iterator *itr, struct vy_entry last,
 			struct vy_cache_node *node =
 				*vy_cache_tree_iterator_get_elem(tree, &pos);
 			int cmp = dir * vy_entry_compare(node->entry, key, def);
+			bool is_visible = vy_cache_iterator_stmt_is_visible(
+							itr, node->entry.stmt);
+			if (!is_visible)
+				*stop = false;
 			if (cmp < 0 || (cmp == 0 && !key_belongs))
 				break;
-			if (vy_cache_iterator_stmt_is_visible(
-					itr, node->entry.stmt)) {
+			if (is_visible) {
 				itr->curr_pos = pos;
 				if (itr->curr.stmt != NULL)
 					tuple_unref(itr->curr.stmt);
