@@ -618,6 +618,38 @@ vy_stmt_stat_decode(struct vy_stmt_stat *stat, const char **data)
 	}
 }
 
+static enum tuple_bloom_version
+iproto_to_tuple_bloom_version(uint32_t key)
+{
+	switch (key) {
+	case VY_RUN_INFO_BLOOM_FILTER_LEGACY_V1:
+		return TUPLE_BLOOM_VERSION_V1;
+	case VY_RUN_INFO_BLOOM_FILTER_LEGACY_V2:
+		return TUPLE_BLOOM_VERSION_V2;
+	case VY_RUN_INFO_BLOOM_FILTER:
+		return TUPLE_BLOOM_VERSION_V3;
+	default:
+		unreachable();
+	}
+	return 0;
+}
+
+static uint32_t
+tuple_bloom_version_to_iproto(enum tuple_bloom_version version)
+{
+	switch (version) {
+	case TUPLE_BLOOM_VERSION_V1:
+		return VY_RUN_INFO_BLOOM_FILTER_LEGACY_V1;
+	case TUPLE_BLOOM_VERSION_V2:
+		return VY_RUN_INFO_BLOOM_FILTER_LEGACY_V2;
+	case TUPLE_BLOOM_VERSION_V3:
+		return VY_RUN_INFO_BLOOM_FILTER;
+	default:
+		unreachable();
+	}
+	return 0;
+}
+
 /**
  * Decode the run metadata from xrow.
  *
@@ -669,13 +701,11 @@ vy_run_info_decode(struct vy_run_info *run_info,
 		case VY_RUN_INFO_PAGE_COUNT:
 			run_info->page_count = mp_decode_uint(&pos);
 			break;
-		case VY_RUN_INFO_BLOOM_FILTER_LEGACY:
-			run_info->bloom = tuple_bloom_decode_legacy(&pos);
-			if (run_info->bloom == NULL)
-				return -1;
-			break;
+		case VY_RUN_INFO_BLOOM_FILTER_LEGACY_V1:
+		case VY_RUN_INFO_BLOOM_FILTER_LEGACY_V2:
 		case VY_RUN_INFO_BLOOM_FILTER:
-			run_info->bloom = tuple_bloom_decode(&pos);
+			run_info->bloom = tuple_bloom_decode(
+				&pos, iproto_to_tuple_bloom_version(key));
 			if (run_info->bloom == NULL)
 				return -1;
 			break;
@@ -787,12 +817,12 @@ vy_page_stmt(struct vy_page *page, uint32_t stmt_no,
  * In terms of STL, makes lower_bound for EQ,GE,LT and upper_bound for GT,LE
  * Additionally *equal_key argument is set to true if the found value is
  * equal to given key (set to false otherwise).
- * @retval position in the page
  */
-static uint32_t
+static int
 vy_page_find_key(struct vy_page *page, struct vy_entry key,
 		 struct key_def *cmp_def, struct tuple_format *format,
-		 enum iterator_type iterator_type, bool *equal_key)
+		 enum iterator_type iterator_type, uint32_t *pos,
+		 bool *equal_key)
 {
 	uint32_t beg = 0;
 	uint32_t end = page->row_count;
@@ -805,7 +835,7 @@ vy_page_find_key(struct vy_page *page, struct vy_entry key,
 		struct vy_entry fnd_key = vy_page_stmt(page, mid, cmp_def,
 						       format);
 		if (fnd_key.stmt == NULL)
-			return end;
+			return -1;
 		int cmp = vy_entry_compare(fnd_key, key, cmp_def);
 		cmp = cmp ? cmp : zero_cmp;
 		*equal_key = *equal_key || cmp == 0;
@@ -815,7 +845,8 @@ vy_page_find_key(struct vy_page *page, struct vy_entry key,
 			end = mid;
 		tuple_unref(fnd_key.stmt);
 	}
-	return end;
+	*pos = end;
+	return 0;
 }
 
 /**
@@ -982,12 +1013,11 @@ vy_page_read_cb(struct cbus_call_msg *base)
 		return -1;
 	if (vy_page_read(task->page, task->page_info, task->run, zdctx) != 0)
 		return -1;
-	if (task->key.stmt != NULL) {
-		task->pos_in_page = vy_page_find_key(task->page, task->key,
-						     task->cmp_def, task->format,
-						     task->iterator_type,
-						     &task->equal_found);
-	}
+	if (task->key.stmt != NULL &&
+	    vy_page_find_key(task->page, task->key, task->cmp_def,
+			     task->format, task->iterator_type,
+			     &task->pos_in_page, &task->equal_found) != 0)
+		return -1;
 	return 0;
 }
 
@@ -1018,10 +1048,11 @@ vy_run_iterator_load_page(struct vy_run_iterator *itr, uint32_t page_no,
 		page = itr->curr_page;
 	}
 	if (page != NULL) {
-		if (key.stmt != NULL)
-			*pos_in_page = vy_page_find_key(page, key, itr->cmp_def,
-							itr->format, iterator_type,
-							equal_found);
+		if (key.stmt != NULL &&
+		    vy_page_find_key(page, key, itr->cmp_def,
+				     itr->format, iterator_type,
+				     pos_in_page, equal_found) != 0)
+			return -1;
 		*result = page;
 		return 0;
 	}
@@ -1551,10 +1582,8 @@ vy_run_iterator_next_lsn(struct vy_run_iterator *itr, struct vy_entry *ret)
 
 	struct vy_run_iterator_pos next_pos;
 next:
-	if (vy_run_iterator_next_pos(itr, ITER_GE, &next_pos) != 0) {
-		vy_run_iterator_stop(itr);
+	if (vy_run_iterator_next_pos(itr, ITER_GE, &next_pos) != 0)
 		return 0;
-	}
 
 	struct vy_entry next;
 	if (vy_run_iterator_read(itr, next_pos, &next) != 0)
@@ -1660,8 +1689,8 @@ vy_run_recover(struct vy_run *run, const char *dir,
 			    space_id, iid, run->id, VY_FILE_INDEX);
 
 	struct xlog_cursor cursor;
-	ERROR_INJECT_COUNTDOWN(ERRINJ_VY_RUN_OPEN, {
-		diag_set(SystemError, "failed to open '%s' file", path);
+	ERROR_INJECT_COUNTDOWN(ERRINJ_VY_RUN_RECOVER_COUNTDOWN, {
+		diag_set(ClientError, ER_INJECTION, "vinyl run recover");
 		goto fail;
 	});
 	if (xlog_cursor_open(&cursor, path))
@@ -1984,8 +2013,12 @@ vy_run_info_encode(const struct vy_run_info *run_info,
 	size_t max_key_size = tmp - run_info->max_key;
 
 	uint32_t key_count = 6;
-	if (run_info->bloom != NULL)
+	uint32_t bloom_key = 0;
+	if (run_info->bloom != NULL) {
 		key_count++;
+		bloom_key = tuple_bloom_version_to_iproto(
+			run_info->bloom->version);
+	}
 
 	size_t size = mp_sizeof_map(key_count);
 	size += mp_sizeof_uint(VY_RUN_INFO_MIN_KEY) + min_key_size;
@@ -1997,7 +2030,7 @@ vy_run_info_encode(const struct vy_run_info *run_info,
 	size += mp_sizeof_uint(VY_RUN_INFO_PAGE_COUNT) +
 		mp_sizeof_uint(run_info->page_count);
 	if (run_info->bloom != NULL)
-		size += mp_sizeof_uint(VY_RUN_INFO_BLOOM_FILTER) +
+		size += mp_sizeof_uint(bloom_key) +
 			tuple_bloom_size(run_info->bloom);
 	size += mp_sizeof_uint(VY_RUN_INFO_STMT_STAT) +
 		vy_stmt_stat_sizeof(&run_info->stmt_stat);
@@ -2024,7 +2057,7 @@ vy_run_info_encode(const struct vy_run_info *run_info,
 	pos = mp_encode_uint(pos, VY_RUN_INFO_PAGE_COUNT);
 	pos = mp_encode_uint(pos, run_info->page_count);
 	if (run_info->bloom != NULL) {
-		pos = mp_encode_uint(pos, VY_RUN_INFO_BLOOM_FILTER);
+		pos = mp_encode_uint(pos, bloom_key);
 		pos = tuple_bloom_encode(run_info->bloom, pos);
 	}
 	pos = mp_encode_uint(pos, VY_RUN_INFO_STMT_STAT);
@@ -2195,6 +2228,7 @@ vy_run_writer_start_page(struct vy_run_writer *writer,
 	if (vy_page_info_create(page, writer->data_xlog.offset,
 				key, writer->cmp_def) != 0)
 		return -1;
+	run->info.page_count++;
 	xlog_tx_begin(&writer->data_xlog);
 	return 0;
 }
@@ -2218,7 +2252,7 @@ vy_run_writer_write_to_page(struct vy_run_writer *writer, struct vy_entry entry)
 	writer->last = entry;
 	vy_stmt_ref_if_possible(entry.stmt);
 	struct vy_run *run = writer->run;
-	struct vy_page_info *page = run->page_info + run->info.page_count;
+	struct vy_page_info *page = run->page_info + run->info.page_count - 1;
 	uint32_t *offset = (uint32_t *)ibuf_alloc(&writer->row_index_buf,
 						  sizeof(uint32_t));
 	if (offset == NULL) {
@@ -2246,7 +2280,8 @@ static int
 vy_run_writer_end_page(struct vy_run_writer *writer)
 {
 	struct vy_run *run = writer->run;
-	struct vy_page_info *page = run->page_info + run->info.page_count;
+	assert(run->info.page_count > 0);
+	struct vy_page_info *page = run->page_info + run->info.page_count - 1;
 
 	assert(page->row_count > 0);
 	assert(ibuf_used(&writer->row_index_buf) ==
@@ -2268,7 +2303,6 @@ vy_run_writer_end_page(struct vy_run_writer *writer)
 	if (written < 0)
 		return -1;
 	page->size = written;
-	run->info.page_count++;
 	vy_run_acct_page(run, page);
 	ibuf_reset(&writer->row_index_buf);
 	return 0;
@@ -2641,11 +2675,13 @@ vy_slice_stream_search(struct vy_stmt_stream *virt_stream)
 		return -1;
 
 	bool unused;
-	stream->pos_in_page = vy_page_find_key(stream->page,
-					       stream->slice->begin,
-					       stream->cmp_def,
-					       stream->format,
-					       ITER_GE, &unused);
+	if (vy_page_find_key(stream->page, stream->slice->begin,
+			     stream->cmp_def, stream->format, ITER_GE,
+			     &stream->pos_in_page, &unused) != 0) {
+		vy_page_delete(stream->page);
+		stream->page = NULL;
+		return -1;
+	}
 
 	if (stream->pos_in_page == stream->page->row_count) {
 		/* The first tuple is in the beginning of the next page */
@@ -2669,11 +2705,12 @@ vy_slice_stream_next(struct vy_stmt_stream *virt_stream, struct vy_entry *ret)
 {
 	assert(virt_stream->iface->next == vy_slice_stream_next);
 	struct vy_slice_stream *stream = (struct vy_slice_stream *)virt_stream;
-	*ret = vy_entry_none();
 
 	/* If the slice is ended, return EOF */
-	if (stream->page_no > stream->slice->last_page_no)
+	if (stream->page_no > stream->slice->last_page_no) {
+		*ret = vy_entry_none();
 		return 0;
+	}
 
 	/* If current page is not already read, read it */
 	if (stream->page == NULL && vy_slice_stream_read_page(stream) != 0)
@@ -2690,6 +2727,7 @@ vy_slice_stream_next(struct vy_stmt_stream *virt_stream, struct vy_entry *ret)
 	    stream->page_no >= stream->slice->last_page_no &&
 	    vy_entry_compare(entry, stream->slice->end, stream->cmp_def) >= 0) {
 		tuple_unref(entry.stmt);
+		*ret = vy_entry_none();
 		return 0;
 	}
 
