@@ -670,7 +670,7 @@ vy_scheduler_complete_dump(struct vy_scheduler *scheduler)
 	scheduler->stat.dump_count++;
 	scheduler->dump_complete_cb(scheduler,
 			min_generation - 1, dump_duration);
-	fiber_cond_signal(&scheduler->dump_cond);
+	fiber_cond_broadcast(&scheduler->dump_cond);
 }
 
 int
@@ -746,6 +746,8 @@ vy_scheduler_end_checkpoint(struct vy_scheduler *scheduler)
 		return;
 
 	scheduler->checkpoint_in_progress = false;
+	fiber_cond_broadcast(&scheduler->dump_cond);
+
 	if (scheduler->dump_pending) {
 		/*
 		 * Dump was triggered while checkpoint was
@@ -1138,6 +1140,17 @@ vy_task_dump_complete(struct vy_task *task)
 
 	assert(lsm->is_dumping);
 
+	/*
+	 * The LSM tree could have been dropped while we were writing the new
+	 * run. In this case all the information about the LSM tree ranges
+	 * could have already been garbage collected from vylog so we can't
+	 * commit the new slice. Discard the run and exit.
+	 */
+	if (lsm->is_dropped) {
+		vy_run_discard(new_run);
+		goto delete_mems;
+	}
+
 	if (vy_run_is_empty(new_run)) {
 		/*
 		 * In case the run is empty, we can discard the run
@@ -1261,7 +1274,7 @@ delete_mems:
 	assert(scheduler->dump_task_count > 0);
 	scheduler->dump_task_count--;
 
-	say_info("%s: dump completed", vy_lsm_name(lsm));
+	say_verbose("%s: dump completed", vy_lsm_name(lsm));
 
 	vy_scheduler_complete_dump(scheduler);
 	return 0;
@@ -1415,7 +1428,7 @@ vy_task_dump_new(struct vy_scheduler *scheduler, struct vy_worker *worker,
 
 	scheduler->dump_task_count++;
 
-	say_info("%s: dump started", vy_lsm_name(lsm));
+	say_verbose("%s: dump started", vy_lsm_name(lsm));
 	*p_task = task;
 	return 0;
 
@@ -1455,12 +1468,12 @@ vy_task_compaction_complete(struct vy_task *task)
 
 	/*
 	 * The LSM tree could have been dropped while we were writing the new
-	 * run. In this case we should discard the run without committing to
-	 * vylog, because all the information about the LSM tree and its runs
-	 * could have already been garbage collected from vylog.
+	 * run. In this case all the information about the LSM tree ranges
+	 * could have already been garbage collected from vylog so we can't
+	 * commit the new slice. Discard the run and exit.
 	 */
 	if (lsm->is_dropped) {
-		vy_run_unref(new_run);
+		vy_run_discard(new_run);
 		goto out;
 	}
 
@@ -1600,8 +1613,8 @@ out:
 	vy_range_heap_insert(&lsm->range_heap, range);
 	vy_scheduler_update_lsm(scheduler, lsm);
 
-	say_info("%s: completed compacting range %s",
-		 vy_lsm_name(lsm), vy_range_str(range));
+	say_verbose("%s: completed compacting range %s",
+		    vy_lsm_name(lsm), vy_range_str(range));
 	return 0;
 }
 
@@ -1712,9 +1725,9 @@ vy_task_compaction_new(struct vy_scheduler *scheduler, struct vy_worker *worker,
 	vy_range_heap_delete(&lsm->range_heap, range);
 	vy_scheduler_update_lsm(scheduler, lsm);
 
-	say_info("%s: started compacting range %s, runs %d/%d",
-		 vy_lsm_name(lsm), vy_range_str(range),
-                 range->compaction_priority, range->slice_count);
+	say_verbose("%s: started compacting range %s, runs %d/%d",
+		    vy_lsm_name(lsm), vy_range_str(range),
+		    range->compaction_priority, range->slice_count);
 	*p_task = task;
 	return 0;
 
@@ -1839,6 +1852,11 @@ retry:
 	}
 	if (!lsm->is_dumping && lsm->pin_count == 0 &&
 	    vy_lsm_generation(lsm) == scheduler->dump_generation) {
+		/*
+		 * The code below may yield. Take a reference to the LSM tree
+		 * to make sure it isn't deleted in the meantime.
+		 */
+		vy_lsm_ref(lsm);
 		/*
 		 * Dump is in progress and there is an LSM tree that
 		 * contains data that must be dumped at the current
@@ -1846,13 +1864,17 @@ retry:
 		 */
 		if (worker == NULL) {
 			worker = vy_worker_pool_get(&scheduler->dump_pool);
-			if (worker == NULL)
+			if (worker == NULL) {
+				vy_lsm_unref(lsm);
 				return 0; /* all workers are busy */
+			}
 		}
 		if (vy_task_dump_new(scheduler, worker, lsm, ptask) != 0) {
 			vy_worker_pool_put(worker);
+			vy_lsm_unref(lsm);
 			return -1;
 		}
+		vy_lsm_unref(lsm);
 		if (*ptask != NULL)
 			return 0; /* new task */
 		/*
@@ -1909,15 +1931,24 @@ retry:
 		goto no_task; /* nothing to do */
 	if (vy_lsm_compaction_priority(lsm) <= 1)
 		goto no_task; /* nothing to do */
+	/*
+	 * The code below may yield. Take a reference to the LSM tree
+	 * to make sure it isn't deleted in the meantime.
+	 */
+	vy_lsm_ref(lsm);
 	if (worker == NULL) {
 		worker = vy_worker_pool_get(&scheduler->compaction_pool);
-		if (worker == NULL)
+		if (worker == NULL) {
+			vy_lsm_unref(lsm);
 			return 0; /* all workers are busy */
+		}
 	}
 	if (vy_task_compaction_new(scheduler, worker, lsm, ptask) != 0) {
 		vy_worker_pool_put(worker);
+		vy_lsm_unref(lsm);
 		return -1;
 	}
+	vy_lsm_unref(lsm);
 	if (*ptask == NULL)
 		goto retry; /* LSM tree dropped or range split/coalesced */
 	return 0; /* new task */
@@ -2050,7 +2081,7 @@ vy_scheduler_f(va_list va)
 		continue;
 error:
 		/* Abort pending checkpoint. */
-		fiber_cond_signal(&scheduler->dump_cond);
+		fiber_cond_broadcast(&scheduler->dump_cond);
 		/*
 		 * A task can fail either due to lack of memory or IO
 		 * error. In either case it is pointless to schedule
