@@ -51,7 +51,6 @@
 #include "trigger.h"
 #include "trivia/util.h"
 #include "tuple.h"
-#include "vy_cache.h"
 #include "vy_lsm.h"
 #include "vy_mem.h"
 #include "vy_stat.h"
@@ -477,80 +476,6 @@ vy_tx_begin(struct vy_tx_manager *xm, enum txn_isolation_level isolation)
 	return tx;
 }
 
-/**
- * Rotate the active in-memory tree if necessary and pin it to make
- * sure it is not dumped until the transaction is complete.
- */
-static int
-vy_tx_write_prepare(struct txv *v)
-{
-	struct vy_lsm *lsm = v->lsm;
-	if (vy_lsm_rotate_mem_if_required(lsm) != 0)
-		return -1;
-	vy_mem_pin(lsm->mem);
-	v->mem = lsm->mem;
-	return 0;
-}
-
-/**
- * Write a single statement into an LSM tree. If the statement has
- * an lsregion copy then use it, else create it.
- *
- * @param lsm         LSM tree to write to.
- * @param mem         In-memory tree to write to.
- * @param entry       Statement allocated with malloc().
- * @param region_stmt NULL or the same statement as stmt,
- *                    but allocated on lsregion.
- *
- * @retval  0 Success.
- * @retval -1 Memory error.
- */
-static int
-vy_tx_write(struct vy_lsm *lsm, struct vy_mem *mem,
-	    struct vy_entry entry, struct tuple **region_stmt)
-{
-	assert(vy_stmt_is_refable(entry.stmt));
-	assert(*region_stmt == NULL || !vy_stmt_is_refable(*region_stmt));
-
-	/*
-	 * The UPSERT statement can be applied to the cached
-	 * statement, because the cache always contains only
-	 * newest REPLACE statements. In such a case the UPSERT,
-	 * applied to the cached statement, can be inserted
-	 * instead of the original UPSERT.
-	 */
-	if (vy_stmt_type(entry.stmt) == IPROTO_UPSERT) {
-		struct vy_entry deleted = vy_entry_none();
-		/* Invalidate cache element. */
-		vy_cache_on_write(&lsm->cache, entry, &deleted);
-		if (deleted.stmt != NULL) {
-			struct vy_entry applied;
-			applied = vy_entry_apply_upsert(entry, deleted,
-							mem->cmp_def, false);
-			tuple_unref(deleted.stmt);
-			if (applied.stmt != NULL) {
-				enum iproto_type applied_type =
-					vy_stmt_type(applied.stmt);
-				assert(applied_type == IPROTO_REPLACE ||
-				       applied_type == IPROTO_INSERT);
-				(void) applied_type;
-				int rc = vy_lsm_set(lsm, mem, applied,
-						    region_stmt);
-				tuple_unref(applied.stmt);
-				return rc;
-			}
-			/*
-			 * Ignore a memory error, because it is
-			 * not critical to apply the optimization.
-			 */
-		}
-	} else {
-		/* Invalidate cache element. */
-		vy_cache_on_write(&lsm->cache, entry, NULL);
-	}
-	return vy_lsm_set(lsm, mem, entry, region_stmt);
-}
-
 /**
  * Try to generate a deferred DELETE statement on tx commit.
  *
@@ -639,6 +564,7 @@ vy_tx_handle_deferred_delete(struct vy_tx *tx, struct txv *v)
 		v->entry.stmt = delete_stmt;
 		tuple_ref(delete_stmt);
 		tuple_unref(stmt);
+		stmt = delete_stmt;
 	}
 
 	/*
@@ -650,22 +576,37 @@ vy_tx_handle_deferred_delete(struct vy_tx *tx, struct txv *v)
 		struct vy_lsm *lsm = vy_lsm(space->index[i]);
 		struct vy_entry entry;
 		vy_stmt_foreach_entry(entry, delete_stmt, lsm->cmp_def) {
+			/*
+			 * If there's a statement in the transaction write set
+			 * with the same key and it hasn't been overwritten by
+			 * another statement, we have the following scenarios:
+			 *
+			 *  1. It's a DELETE. Then it must have been generated
+			 *     by some UPDATE or UPSERT statement. Unless it's
+			 *     a no-op, we have nothing to do.
+			 *
+			 *  2. It's a REPLACE. Then it must be a REPLACE
+			 *     generated by the statement we're currently
+			 *     processing (must be a REPLACE, too), which
+			 *     happens not to update the secondary index key
+			 *     parts. We must not generate a DELETE for it,
+			 *     otherwise we'd lose the secondary key, but
+			 *     we may skip it because it's essentially a no-op,
+			 *     see vy_tx_set_entry().
+			 */
 			struct txv *other = write_set_search_key(&tx->write_set,
 								 lsm, entry);
 			if (other != NULL && !other->is_overwritten) {
-				/*
-				 * The write set contains a statement
-				 * for the key to be deleted. This can
-				 * only occur if it's a REPLACE that
-				 * happens not to update the secondary
-				 * index key parts. It's safe to skip it,
-				 * see vy_tx_set_entry().
-				 */
-				assert(vy_stmt_type(stmt) == IPROTO_REPLACE);
-				assert(vy_stmt_type(other->entry.stmt) ==
-								IPROTO_REPLACE);
-				other->is_nop = true;
-				continue;
+				if (vy_stmt_type(other->entry.stmt) ==
+				    IPROTO_DELETE) {
+					if (!other->is_nop)
+						continue;
+				} else {
+					assert(vy_stmt_type(stmt) ==
+					       IPROTO_REPLACE);
+					other->is_nop = true;
+					continue;
+				}
 			}
 			struct txv *delete_txv = txv_new(tx, lsm, entry);
 			if (delete_txv == NULL) {
@@ -727,18 +668,16 @@ vy_tx_prepare(struct vy_tx *tx)
 		}
 		assert(lsm->space_id == current_space_id);
 
-		if (lsm->index_id > 0 && repsert == NULL && delete == NULL) {
+		enum iproto_type type = vy_stmt_type(v->entry.stmt);
+
+		if (lsm->index_id > 0 && repsert == NULL &&
+		    type != IPROTO_DELETE) {
 			/*
-			 * This statement is for a secondary index,
-			 * and the statement corresponding to it in
-			 * the primary index was overwritten. This
-			 * can only happen if insertion of DELETE
-			 * into secondary indexes was postponed until
-			 * primary index compaction. In this case
-			 * the DELETE will not be generated, because
-			 * the corresponding statement never made it
-			 * to the primary index LSM tree. So we must
-			 * skip it for secondary indexes as well.
+			 * With deferred DELETEs enabled, a REPLACE that was
+			 * overwritten in the primary index may lack a DELETE
+			 * in a secondary index. We must skip such a REPLACE
+			 * because, since it's skipped in the primary index,
+			 * we wouldn't generate a DELETE for it on compaction.
 			 */
 			v->is_overwritten = true;
 		}
@@ -747,8 +686,6 @@ vy_tx_prepare(struct vy_tx *tx)
 		if (v->is_overwritten || v->is_nop)
 			continue;
 
-		enum iproto_type type = vy_stmt_type(v->entry.stmt);
-
 		/* Optimize out INSERT + DELETE for the same key. */
 		if (v->is_first_insert && type == IPROTO_DELETE)
 			continue;
@@ -782,9 +719,15 @@ vy_tx_prepare(struct vy_tx *tx)
 			vy_stmt_set_type(v->entry.stmt, type);
 		}
 
-		if (vy_tx_write_prepare(v) != 0)
+		/*
+		 * Rotate the active in-memory tree if necessary and pin it
+		 * to make sure it is not dumped until the transaction is
+		 * complete.
+		 */
+		if (vy_lsm_rotate_mem_if_required(lsm) != 0)
 			return -1;
-		assert(v->mem != NULL);
+		vy_mem_pin(lsm->mem);
+		v->mem = lsm->mem;
 
 		if (lsm->index_id == 0 &&
 		    vy_stmt_flags(v->entry.stmt) & VY_STMT_DEFERRED_DELETE &&
@@ -795,7 +738,7 @@ vy_tx_prepare(struct vy_tx *tx)
 		vy_stmt_set_lsn(v->entry.stmt, MAX_LSN + tx->psn);
 		struct tuple **region_stmt =
 			(type == IPROTO_DELETE) ? &delete : &repsert;
-		if (vy_tx_write(lsm, v->mem, v->entry, region_stmt) != 0)
+		if (vy_lsm_set(lsm, v->mem, v->entry, region_stmt) != 0)
 			return -1;
 		v->region_stmt = *region_stmt;
 	}
@@ -1059,6 +1002,23 @@ static int
 vy_tx_set_entry(struct vy_tx *tx, struct vy_lsm *lsm, struct vy_entry entry)
 {
 	assert(vy_stmt_type(entry.stmt) != 0);
+
+	struct txv *old = write_set_search_key(&tx->write_set, lsm, entry);
+	if (old != NULL && old->entry.stmt == entry.stmt) {
+		/*
+		 * The inserted statement is already indexed in the write set.
+		 * This may happen only if this is a multikey index and the
+		 * indexed array has duplicate entries. Inserting a duplicate
+		 * into the write set is pointless. Moreover, it may break
+		 * assumptions taken by the optimizations applied below, like
+		 * REPLACE + DELETE = NOP. Let's skip it.
+		 */
+		assert(lsm->cmp_def->is_multikey);
+		assert(!old->is_overwritten);
+		assert(old->entry.hint != entry.hint);
+		return 0;
+	}
+
 	/**
 	 * A statement in write set must have and unique lsn
 	 * in order to differ it from cachable statements in mem and run.
@@ -1066,7 +1026,6 @@ vy_tx_set_entry(struct vy_tx *tx, struct vy_lsm *lsm, struct vy_entry entry)
 	vy_stmt_set_lsn(entry.stmt, INT64_MAX);
 	struct vy_entry applied = vy_entry_none();
 
-	struct txv *old = write_set_search_key(&tx->write_set, lsm, entry);
 	/* Found a match of the previous action of this transaction */
 	if (old != NULL && vy_stmt_type(entry.stmt) == IPROTO_UPSERT) {
 		assert(lsm->index_id == 0);
@@ -1189,6 +1148,12 @@ vy_tx_manager_abort_writers_for_ddl(struct vy_tx_manager *xm,
 	struct vy_lsm *lsm = vy_lsm(space->index[0]);
 	struct vy_tx *tx;
 	rlist_foreach_entry(tx, &xm->writers, in_writers) {
+		if (tx->state == VINYL_TX_ABORT)
+			continue;
+		if (tx->last_stmt_space != space &&
+		    write_set_search_key(&tx->write_set, lsm,
+					 lsm->env->empty_key) == NULL)
+			continue;
 		/*
 		 * We can't abort prepared transactions as they have
 		 * already reached WAL. The caller needs to sync WAL
@@ -1196,11 +1161,7 @@ vy_tx_manager_abort_writers_for_ddl(struct vy_tx_manager *xm,
 		 */
 		if (tx->state == VINYL_TX_COMMIT)
 			*need_wal_sync = true;
-		if (tx->state != VINYL_TX_READY)
-			continue;
-		if (tx->last_stmt_space == space ||
-		    write_set_search_key(&tx->write_set, lsm,
-					 lsm->env->empty_key) != NULL)
+		else
 			vy_tx_abort(tx);
 	}
 }
