@@ -69,14 +69,12 @@ RLIST_HEAD(wal_on_write);
 static int
 wal_write_async(struct journal *, struct journal_entry *);
 
-static int
-wal_write(struct journal *, struct journal_entry *);
-
 static int
 wal_write_none_async(struct journal *, struct journal_entry *);
 
+/** Wait until all submitted writes are successfully flushed to disk. */
 static int
-wal_write_none(struct journal *, struct journal_entry *);
+wal_sync(struct journal *journal, struct vclock *vclock);
 
 /*
  * WAL writer - maintain a Write Ahead Log for every change
@@ -103,6 +101,8 @@ struct wal_writer
 	struct cpipe wal_pipe;
 	/** A memory pool for messages. */
 	struct mempool msg_pool;
+	/** Vclock to propagate on write. */
+	struct vclock *instance_vclock;
 	/**
 	 * A last journal entry submitted to write. This is a
 	 * 'rollback border'. When rollback starts, all
@@ -340,6 +340,7 @@ tx_complete_rollback(void)
 	if (stailq_last_entry(&writer->rollback, struct journal_entry,
 			      fifo) != writer->last_entry)
 		return;
+	journal_queue_rollback();
 	stailq_reverse(&writer->rollback);
 	tx_schedule_queue(&writer->rollback);
 	/* TX-thread can try sending transactions to WAL again. */
@@ -375,7 +376,7 @@ tx_complete_batch(struct cmsg *msg)
 		tx_complete_rollback();
 	}
 	/* Update the tx vclock to the latest written by wal. */
-	vclock_copy(&replicaset.vclock, &batch->vclock);
+	vclock_copy(writer->instance_vclock, &batch->vclock);
 	tx_schedule_queue(&batch->commit);
 	trigger_run(&wal_on_write, NULL);
 	mempool_free(&writer->msg_pool, container_of(msg, struct wal_msg, base));
@@ -410,6 +411,15 @@ tx_notify_checkpoint(struct cmsg *msg)
 	free(msg);
 }
 
+static int
+wal_sync_none(struct journal *journal, struct vclock *out)
+{
+	(void)journal;
+	if (out != NULL)
+		vclock_copy(out, &wal_writer_singleton.vclock);
+	return 0;
+}
+
 /**
  * Initialize WAL writer context. Even though it's a singleton,
  * encapsulate the details just in case we may use
@@ -418,22 +428,32 @@ tx_notify_checkpoint(struct cmsg *msg)
 static void
 wal_writer_create(struct wal_writer *writer, enum wal_mode wal_mode,
 		  const char *wal_dirname, int64_t wal_max_size,
+		  double wal_retention_period,
 		  const struct tt_uuid *instance_uuid,
+		  struct vclock *instance_vclock,
+		  struct vclock *checkpoint_vclock,
 		  wal_on_garbage_collection_f on_garbage_collection,
 		  wal_on_checkpoint_threshold_f on_checkpoint_threshold)
 {
+	writer->instance_vclock = instance_vclock;
 	writer->wal_mode = wal_mode;
 	writer->wal_max_size = wal_max_size;
 
-	journal_create(&writer->base,
-		       wal_mode == WAL_NONE ?
-		       wal_write_none_async : wal_write_async,
-		       wal_mode == WAL_NONE ?
-		       wal_write_none : wal_write);
+	if (wal_mode == WAL_NONE) {
+		journal_create(&writer->base, wal_write_none_async,
+			       wal_sync_none);
+	} else {
+		journal_create(&writer->base, wal_write_async, wal_sync);
+	}
 
 	struct xlog_opts opts = xlog_opts_default;
 	opts.sync_is_async = true;
 	xdir_create(&writer->wal_dir, wal_dirname, XLOG, instance_uuid, &opts);
+	/*
+	 * wal_retention_period must be set before gc is woken up.
+	 * Otherwise files which must be preserved can be deleted.
+	 */
+	xdir_set_retention_period(&writer->wal_dir, wal_retention_period);
 	xlog_clear(&writer->current_wal);
 	if (wal_mode == WAL_FSYNC)
 		writer->wal_dir.open_wflags |= O_SYNC;
@@ -447,6 +467,8 @@ wal_writer_create(struct wal_writer *writer, enum wal_mode wal_mode,
 
 	vclock_create(&writer->vclock);
 	vclock_create(&writer->checkpoint_vclock);
+	if (checkpoint_vclock != NULL)
+		vclock_copy(&writer->checkpoint_vclock, checkpoint_vclock);
 	rlist_create(&writer->watchers);
 
 	writer->on_garbage_collection = on_garbage_collection;
@@ -475,11 +497,30 @@ wal_open_f(struct cbus_call_msg *msg)
 	const char *path = xdir_format_filename(&writer->wal_dir,
 				vclock_sum(&writer->vclock), NONE);
 	assert(!xlog_is_open(&writer->current_wal));
+
+	uint64_t sum_already_on_disk = 0;
+	for (struct vclock *vclock = vclockset_first(&writer->wal_dir.index);
+	     vclock != NULL;
+	     vclock = vclockset_next(&writer->wal_dir.index, vclock)) {
+		if (vclock_compare(&writer->checkpoint_vclock, vclock) > 0)
+			continue;
+		const char *name = xdir_format_filename(&writer->wal_dir,
+							vclock_sum(vclock),
+							NONE);
+		struct stat attr;
+		if (stat(name, &attr) != 0)
+			say_warn("failed to get file size %s, consider it"
+				 " equal to zero", name);
+		else
+			sum_already_on_disk += (uint64_t)attr.st_size;
+	}
+	writer->checkpoint_wal_size = sum_already_on_disk;
 	return xlog_open(&writer->current_wal, path, &writer->wal_dir.opts);
 }
 
 /**
- * Try to open the current WAL file for appending if it exists.
+ * Try to open the current WAL file for appending if it exists
+ * and scan writer->wal_dir to set writer->checkpoint_wal_size.
  */
 static int
 wal_open(struct wal_writer *writer)
@@ -538,14 +579,18 @@ wal_open(struct wal_writer *writer)
 
 int
 wal_init(enum wal_mode wal_mode, const char *wal_dirname,
-	 int64_t wal_max_size, const struct tt_uuid *instance_uuid,
+	 int64_t wal_max_size, double wal_retention_period,
+	 const struct tt_uuid *instance_uuid,
+	 struct vclock *instance_vclock,
+	 struct vclock *checkpoint_vclock,
 	 wal_on_garbage_collection_f on_garbage_collection,
 	 wal_on_checkpoint_threshold_f on_checkpoint_threshold)
 {
 	/* Initialize the state. */
 	struct wal_writer *writer = &wal_writer_singleton;
 	wal_writer_create(writer, wal_mode, wal_dirname, wal_max_size,
-			  instance_uuid, on_garbage_collection,
+			  wal_retention_period, instance_uuid, instance_vclock,
+			  checkpoint_vclock, on_garbage_collection,
 			  on_checkpoint_threshold);
 
 	/* Start WAL thread. */
@@ -564,7 +609,7 @@ wal_enable(void)
 	struct wal_writer *writer = &wal_writer_singleton;
 
 	/* Initialize the writer vclock from the recovery state. */
-	vclock_copy(&writer->vclock, &replicaset.vclock);
+	vclock_copy(&writer->vclock, writer->instance_vclock);
 
 	/*
 	 * Scan the WAL directory to build an index of all
@@ -619,30 +664,29 @@ wal_sync_f(struct cbus_call_msg *data)
 	return 0;
 }
 
-int
-wal_sync(struct vclock *vclock)
+static int
+wal_sync(struct journal *journal, struct vclock *vclock)
 {
+	(void)journal;
+	struct wal_writer *writer = &wal_writer_singleton;
+	assert(writer->wal_mode != WAL_NONE);
 	ERROR_INJECT(ERRINJ_WAL_SYNC, {
 		diag_set(ClientError, ER_INJECTION, "wal sync");
 		return -1;
 	});
-
-	struct wal_writer *writer = &wal_writer_singleton;
-	if (writer->wal_mode == WAL_NONE) {
-		if (vclock != NULL)
-			vclock_copy(vclock, &writer->vclock);
-		return 0;
-	}
 	if (!stailq_empty(&writer->rollback)) {
 		/* We're rolling back a failed write. */
 		diag_set(ClientError, ER_CASCADE_ROLLBACK);
 		return -1;
 	}
+	if (journal_queue_flush() != 0)
+		return -1;
 	struct wal_vclock_msg msg;
 	int rc = cbus_call(&writer->wal_pipe, &writer->tx_prio_pipe, &msg.base,
 			   wal_sync_f);
 	if (vclock != NULL)
 		vclock_copy(vclock, &msg.vclock);
+	ERROR_INJECT_YIELD(ERRINJ_WAL_SYNC_DELAY);
 	return rc;
 }
 
@@ -666,7 +710,8 @@ wal_begin_checkpoint_f(struct cbus_call_msg *data)
 	if (xlog_is_open(&writer->current_wal) &&
 	    vclock_sum(&writer->current_wal.meta.vclock) !=
 	    vclock_sum(&writer->vclock)) {
-
+		xdir_set_retention_vclock(
+			&writer->wal_dir, &writer->current_wal.meta.vclock);
 		xlog_close(&writer->current_wal, false);
 		/*
 		 * The next WAL will be created on the first write.
@@ -698,6 +743,8 @@ wal_begin_checkpoint(struct wal_checkpoint *checkpoint)
 		diag_set(ClientError, ER_CASCADE_ROLLBACK);
 		return -1;
 	}
+	if (journal_queue_flush() != 0)
+		return -1;
 	return cbus_call(&writer->wal_pipe, &writer->tx_prio_pipe,
 			 &checkpoint->base, wal_begin_checkpoint_f);
 }
@@ -769,6 +816,59 @@ wal_set_queue_max_size(int64_t size)
 	journal_queue_set_max_size(size);
 }
 
+/** Retention delay configuration message. */
+struct wal_set_retention_period_msg {
+	/* The state of a synchronous cross-thread call. */
+	struct cbus_call_msg base;
+	/* New retention_period value. */
+	double retention_period;
+};
+
+static int
+wal_set_retention_period_f(struct cbus_call_msg *data)
+{
+	struct wal_writer *writer = &wal_writer_singleton;
+	struct wal_set_retention_period_msg *msg;
+	msg = (struct wal_set_retention_period_msg *)data;
+	xdir_set_retention_period(&writer->wal_dir, msg->retention_period);
+	return 0;
+}
+
+void
+wal_set_retention_period(double period)
+{
+	struct wal_writer *writer = &wal_writer_singleton;
+	struct wal_set_retention_period_msg msg;
+	msg.retention_period = period;
+	cbus_call(&writer->wal_pipe, &writer->tx_prio_pipe,
+		  &msg.base, wal_set_retention_period_f);
+}
+
+static int
+wal_get_retention_vclock_f(struct cbus_call_msg *data)
+{
+	struct wal_writer *writer = &wal_writer_singleton;
+	struct wal_vclock_msg *msg = (struct wal_vclock_msg *)data;
+	xdir_get_retention_vclock(&writer->wal_dir, &msg->vclock);
+	return 0;
+}
+
+void
+wal_get_retention_vclock(struct vclock *vclock)
+{
+	struct wal_writer *writer = &wal_writer_singleton;
+	if (vclock == NULL)
+		return;
+	if (writer->wal_dir.retention_period == 0) {
+		vclock_clear(vclock);
+		return;
+	}
+	struct wal_vclock_msg msg;
+	cbus_call(&writer->wal_pipe, &writer->tx_prio_pipe,
+		  &msg.base, wal_get_retention_vclock_f);
+	vclock_copy(vclock, &msg.vclock);
+}
+
 struct wal_gc_msg
 {
 	struct cbus_call_msg base;
@@ -841,6 +941,8 @@ wal_opt_rotate(struct wal_writer *writer)
 	 */
 	if (xlog_is_open(&writer->current_wal) &&
 	    writer->current_wal.offset >= writer->wal_max_size) {
+		xdir_set_retention_vclock(
+			&writer->wal_dir, &writer->current_wal.meta.vclock);
 		/*
 		 * We can not handle xlog_close()
 		 * failure in any reasonable way.
@@ -886,6 +988,14 @@ wal_fallocate(struct wal_writer *writer, size_t len)
 	 */
 	int64_t gc_lsn = vclock_sum(&writer->checkpoint_vclock);
 
+	struct vclock retention_vclock;
+	xdir_get_retention_vclock(&writer->wal_dir, &retention_vclock);
+	if (vclock_is_set(&retention_vclock)) {
+		int64_t retention_lsn = vclock_sum(&retention_vclock);
+		if (retention_lsn < gc_lsn)
+			gc_lsn = retention_lsn;
+	}
+
 	/*
 	 * The actual write size can be greater than the sum size
 	 * of encoded rows (compression, fixheaders). Double the
@@ -958,10 +1068,10 @@ wal_assign_lsn(struct vclock *vclock_diff, struct vclock *base,
 {
 	int64_t tsn = 0;
 	struct xrow_header **start = entry->rows;
-	struct xrow_header **end = entry->rows + entry->n_rows;
+	struct xrow_header **end = entry->rows + entry->n_rows - 1;
 	struct xrow_header **first_glob_row = entry->rows;
 	/** Assign LSN to all local rows. */
-	for (struct xrow_header **row = start; row < end; row++) {
+	for (struct xrow_header **row = start; row <= end; row++) {
 		if ((*row)->replica_id == 0) {
 			/*
 			 * All rows representing local space data
@@ -976,23 +1086,6 @@ wal_assign_lsn(struct vclock *vclock_diff, struct vclock *base,
 
 			(*row)->lsn = vclock_inc(vclock_diff, (*row)->replica_id) +
 				      vclock_get(base, (*row)->replica_id);
-			/*
-			 * Use lsn of the first global row as
-			 * transaction id.
-			 */
-			if ((*row)->group_id != GROUP_LOCAL && tsn == 0) {
-				tsn = (*row)->lsn;
-				/*
-				 * Remember the tail being processed.
-				 */
-				first_glob_row = row;
-			}
-			(*row)->tsn = tsn == 0 ? (*start)->lsn : tsn;
-			/* Tx meta is stored in the last tx row. */
-			if (row == end - 1) {
-				(*row)->flags = entry->flags;
-				(*row)->is_commit = true;
-			}
 		} else {
 			int64_t diff = (*row)->lsn - vclock_get(base, (*row)->replica_id);
 			if (diff <= vclock_get(vclock_diff,
@@ -1009,7 +1102,21 @@ wal_assign_lsn(struct vclock *vclock_diff, struct vclock *base,
 			} else {
 				vclock_follow(vclock_diff, (*row)->replica_id, diff);
 			}
+			/* Reset row flags taken from the remote instance. */
+			(*row)->flags = 0;
 		}
+		/*
+		 * Use lsn of the first global row as
+		 * transaction id.
+		 */
+		if (tsn == 0 && (*row)->group_id != GROUP_LOCAL) {
+			tsn = (*row)->lsn;
+			/*
+			 * Remember the tail being processed.
+			 */
+			first_glob_row = row;
+		}
+		(*row)->tsn = tsn == 0 ? (*start)->lsn : tsn;
 	}
 
 	/*
@@ -1019,6 +1126,9 @@ wal_assign_lsn(struct vclock *vclock_diff, struct vclock *base,
 	 */
 	for (struct xrow_header **row = start; row < first_glob_row; row++)
 		(*row)->tsn = tsn;
+	/* Tx meta is stored in the last tx row. */
+	(*end)->flags = entry->flags;
+	(*end)->is_commit = true;
 }
 
 static void
@@ -1151,7 +1261,7 @@ done:
 	}
 	/*
 	 * Remember the vclock of the last successfully written row so
-	 * that we can update replicaset.vclock once this message gets
+	 * that we can update instance_vclock once this message gets
 	 * back to tx.
 	 */
 	vclock_copy(&wal_msg->vclock, &writer->vclock);
@@ -1298,24 +1408,6 @@ fail:
 	return -1;
 }
 
-static int
-wal_write(struct journal *journal, struct journal_entry *entry)
-{
-	/*
-	 * We can reuse async WAL engine transparently
-	 * to the caller.
-	 */
-	if (wal_write_async(journal, entry) != 0)
-		return -1;
-
-	assert(!entry->is_complete);
-	do {
-		fiber_yield();
-	} while (!entry->is_complete);
-
-	return 0;
-}
-
 static int
 wal_write_none_async(struct journal *journal,
 		     struct journal_entry *entry)
@@ -1326,18 +1418,12 @@ wal_write_none_async(struct journal *journal,
 	vclock_create(&vclock_diff);
 	wal_assign_lsn(&vclock_diff, &writer->vclock, entry);
 	vclock_merge(&writer->vclock, &vclock_diff);
-	vclock_copy(&replicaset.vclock, &writer->vclock);
+	vclock_copy(writer->instance_vclock, &writer->vclock);
 	entry->res = vclock_sum(&writer->vclock);
 	journal_async_complete(entry);
 	return 0;
 }
 
-static int
-wal_write_none(struct journal *journal, struct journal_entry *entry)
-{
-	return wal_write_none_async(journal, entry);
-}
-
 void
 wal_init_vy_log(void)
 {
