@@ -47,6 +47,7 @@
 #include "iproto_constants.h"
 #include "errinj.h"
 #include "trivia/util.h"
+#include "retention_period.h"
 
 /*
  * FALLOC_FL_KEEP_SIZE flag has existed since fallocate() was
@@ -359,6 +360,7 @@ xdir_create(struct xdir *dir, const char *dirname, enum xdir_type type,
 		unreachable();
 	}
 	dir->type = type;
+	dir->retention_period = 0;
 }
 
 /**
@@ -419,16 +421,11 @@ xdir_index_file(struct xdir *dir, int64_t signature)
 	}
 
 	/*
-	 * Append the clock describing the file to the
-	 * directory index.
+	 * Append the clock describing the file to the directory index.
+	 * If retention feature is available (EE), then additional memory is
+	 * allocated. Time, when xlog protection should stop is saved there.
 	 */
-	struct vclock *vclock = (struct vclock *) malloc(sizeof(*vclock));
-	if (vclock == NULL) {
-		diag_set(OutOfMemory, sizeof(*vclock), "malloc", "vclock");
-		xlog_cursor_close(&cursor, false);
-		return -1;
-	}
-
+	struct vclock *vclock = retention_vclock_new();
 	vclock_copy(vclock, &meta->vclock);
 	xlog_cursor_close(&cursor, false);
 	vclockset_insert(&dir->index, vclock);
@@ -627,6 +624,8 @@ xdir_scan(struct xdir *dir, bool is_dir_required)
 			i++;
 		}
 	}
+	/* Initialize expiration time of all files, saved inside index. */
+	retention_index_update(dir, 0);
 	rc = 0;
 
 exit:
@@ -648,6 +647,31 @@ xdir_check(struct xdir *dir)
 	return 0;
 }
 
+void
+xdir_set_retention_period(struct xdir *xdir, double period)
+{
+	double old_period = xdir->retention_period;
+	xdir->retention_period = period;
+	retention_index_update(xdir, old_period);
+}
+
+void
+xdir_get_retention_vclock(struct xdir *xdir, struct vclock *vclock)
+{
+	retention_index_get(&xdir->index, vclock);
+}
+
+void
+xdir_set_retention_vclock(struct xdir *xdir, struct vclock *vclock)
+{
+	if (xdir->retention_period == 0)
+		return;
+
+	struct vclock *find = vclockset_match(&xdir->index, vclock);
+	assert(vclock_compare(find, vclock) == 0);
+	retention_vclock_set(find, xdir->retention_period);
+}
+
 const char *
 xdir_format_filename(struct xdir *dir, int64_t signature,
 		enum log_suffix suffix)
@@ -679,6 +703,23 @@ xdir_complete_gc(eio_req *req)
 void
 xdir_collect_garbage(struct xdir *dir, int64_t signature, unsigned flags)
 {
+	struct vclock retention_vclock;
+	retention_index_get(&dir->index, &retention_vclock);
+	if (vclock_is_set(&retention_vclock) &&
+	    signature > vclock_sum(&retention_vclock)) {
+		/*
+		 * This should not normally happen and can be a reason of
+		 * xdir_retention misusage: retention_vclock must be always
+		 * checked before invoking garbage collection. As we cannot
+		 * return error code from this function (gc already updated
+		 * its vclock), the best we can do is warning.
+		 */
+		assert(false);
+		say_warn("Requested xdir gc vclock is greater than "
+			 "retention vclock. Using retention vclock instead");
+		signature = vclock_sum(&retention_vclock);
+	}
+
 	struct vclock *vclock;
 	while ((vclock = vclockset_first(&dir->index)) != NULL &&
 	       vclock_sum(vclock) < signature) {
@@ -761,9 +802,7 @@ xdir_collect_inprogress(struct xdir *xdir)
 void
 xdir_add_vclock(struct xdir *xdir, const struct vclock *vclock)
 {
-	struct vclock *copy = malloc(sizeof(*vclock));
-	if (copy == NULL)
-		panic("failed to allocate vclock");
+	struct vclock *copy = retention_vclock_new();
 	vclock_copy(copy, vclock);
 	vclockset_insert(&xdir->index, copy);
 }
@@ -1096,7 +1135,7 @@ xlog_tx_write_plain(struct xlog *log)
 	 * now populate it with data.
 	 */
 	char *fixheader = (char *)log->obuf.iov[0].iov_base;
-	*(log_magic_t *)fixheader = row_marker;
+	memcpy(fixheader, &row_marker, sizeof(log_magic_t));
 	char *data = fixheader + sizeof(log_magic_t);
 
 	data = mp_encode_uint(data,
@@ -1199,7 +1238,7 @@ xlog_tx_write_zstd(struct xlog *log)
 		offset = 0;
 	}
 
-	*(log_magic_t *)fixheader = zrow_marker;
+	memcpy(fixheader, &zrow_marker, sizeof(log_magic_t));
 	char *data;
 	data = fixheader + sizeof(log_magic_t);
 	data = mp_encode_uint(data,
