@@ -64,15 +64,22 @@ mp_sizeof_vclock_ignore0(const struct vclock *vclock)
 					     mp_sizeof_uint(UINT64_MAX));
 }
 
+static inline uint32_t
+mp_sizeof_vclock(const struct vclock *vclock)
+{
+	uint32_t size = vclock_size(vclock);
+	return mp_sizeof_map(size) + size * (mp_sizeof_uint(UINT32_MAX) +
+					     mp_sizeof_uint(UINT64_MAX));
+}
+
 static inline char *
-mp_encode_vclock_ignore0(char *data, const struct vclock *vclock)
+mp_encode_vclock_impl(char *data, const struct vclock *vclock, bool ignore0)
 {
-	data = mp_encode_map(data, vclock_size_ignore0(vclock));
 	struct vclock_iterator it;
 	vclock_iterator_init(&it, vclock);
 	struct vclock_c replica;
 	replica = vclock_iterator_next(&it);
-	if (replica.id == 0)
+	if (replica.id == 0 && ignore0)
 		replica = vclock_iterator_next(&it);
 	for ( ; replica.id < VCLOCK_MAX; replica = vclock_iterator_next(&it)) {
 		data = mp_encode_uint(data, replica.id);
@@ -81,8 +88,22 @@ mp_encode_vclock_ignore0(char *data, const struct vclock *vclock)
 	return data;
 }
 
+static inline char *
+mp_encode_vclock_ignore0(char *data, const struct vclock *vclock)
+{
+	data = mp_encode_map(data, vclock_size_ignore0(vclock));
+	return mp_encode_vclock_impl(data, vclock, true);
+}
+
+static inline char *
+mp_encode_vclock(char *data, const struct vclock *vclock)
+{
+	data = mp_encode_map(data, vclock_size(vclock));
+	return mp_encode_vclock_impl(data, vclock, false);
+}
+
 static int
-mp_decode_vclock_ignore0(const char **data, struct vclock *vclock)
+mp_decode_vclock(const char **data, struct vclock *vclock)
 {
 	vclock_create(vclock);
 	if (mp_typeof(**data) != MP_MAP)
@@ -95,16 +116,21 @@ mp_decode_vclock_ignore0(const char **data, struct vclock *vclock)
 		if (mp_typeof(**data) != MP_UINT)
 			return -1;
 		int64_t lsn = mp_decode_uint(data);
-		/*
-		 * Skip vclock[0] coming from the remote
-		 * instances.
-		 */
-		if (lsn > 0 && id != 0)
+		if (lsn > 0)
 			vclock_follow(vclock, id, lsn);
 	}
 	return 0;
 }
 
+static int
+mp_decode_vclock_ignore0(const char **data, struct vclock *vclock)
+{
+	if (mp_decode_vclock(data, vclock) != 0)
+		return -1;
+	vclock_reset(vclock, 0, 0);
+	return 0;
+}
+
 /**
  * If log_level is 'verbose' or greater,
  * dump the corrupted row contents in hex to the log.
@@ -1252,21 +1278,33 @@ xrow_encode_synchro(struct xrow_header *row, char *body,
 
 	char *pos = body;
 
-	pos = mp_encode_map(pos,
-			    iproto_type_is_promote_request(req->type) ? 3 : 2);
+	/* Skip one byte for the map. */
+	pos++;
+	uint32_t map_size = 0;
 
 	pos = mp_encode_uint(pos, IPROTO_REPLICA_ID);
 	pos = mp_encode_uint(pos, req->replica_id);
+	map_size++;
 
 	pos = mp_encode_uint(pos, IPROTO_LSN);
 	pos = mp_encode_uint(pos, req->lsn);
+	map_size++;
 
-	if (iproto_type_is_promote_request(req->type)) {
+	if (req->term != 0) {
 		pos = mp_encode_uint(pos, IPROTO_TERM);
 		pos = mp_encode_uint(pos, req->term);
+		map_size++;
 	}
 
-	assert(pos - body < XROW_SYNCHRO_BODY_LEN_MAX);
+	if (req->confirmed_vclock != NULL) {
+		pos = mp_encode_uint(pos, IPROTO_VCLOCK);
+		pos = mp_encode_vclock_ignore0(pos, req->confirmed_vclock);
+		map_size++;
+	}
+
+	mp_encode_map(body, map_size);
+
+	assert(pos - body < XROW_BODY_LEN_MAX);
 
 	memset(row, 0, sizeof(*row));
 	row->type = req->type;
@@ -1276,7 +1314,8 @@ xrow_encode_synchro(struct xrow_header *row, char *body,
 }
 
 int
-xrow_decode_synchro(const struct xrow_header *row, struct synchro_request *req)
+xrow_decode_synchro(const struct xrow_header *row, struct synchro_request *req,
+		    struct vclock *vclock)
 {
 	if (row->bodycnt == 0) {
 		diag_set(ClientError, ER_INVALID_MSGPACK, "request body");
@@ -1301,7 +1340,9 @@ xrow_decode_synchro(const struct xrow_header *row, struct synchro_request *req)
 			continue;
 		}
 		uint8_t key = mp_decode_uint(&d);
-		if (key < iproto_key_MAX && iproto_key_type[key] != type) {
+		if (key < iproto_key_MAX &&
+		    iproto_key_type[key] != mp_typeof(*d)) {
+bad_msgpack:
 			xrow_on_decode_err(row, ER_INVALID_MSGPACK,
 					   "request body");
 			return -1;
@@ -1316,6 +1357,13 @@ xrow_decode_synchro(const struct xrow_header *row, struct synchro_request *req)
 		case IPROTO_TERM:
 			req->term = mp_decode_uint(&d);
 			break;
+		case IPROTO_VCLOCK:
+			if (vclock == NULL)
+				mp_next(&d);
+			else if (mp_decode_vclock_ignore0(&d, vclock) != 0)
+				goto bad_msgpack;
+			req->confirmed_vclock = vclock;
+			break;
 		default:
 			mp_next(&d);
 		}
@@ -1369,7 +1417,7 @@ xrow_encode_raft(struct xrow_header *row, struct region *region,
 	memset(row, 0, sizeof(*row));
 	row->type = IPROTO_RAFT;
 	row->body[0].iov_base = buf;
-	row->group_id = GROUP_LOCAL;
+	row->group_id = r->group_id;
 	row->bodycnt = 1;
 	const char *begin = buf;
 
@@ -1405,13 +1453,14 @@ xrow_decode_raft(const struct xrow_header *row, struct raft_request *r,
 {
 	if (row->type != IPROTO_RAFT)
 		goto bad_msgpack;
-	if (row->bodycnt != 1 || row->group_id != GROUP_LOCAL) {
+	if (row->bodycnt != 1) {
 		diag_set(ClientError, ER_INVALID_MSGPACK,
 			 "malformed raft request");
 		return -1;
 	}
 	memset(r, 0, sizeof(*r));
 
+	r->group_id = row->group_id;
 	const char *pos = row->body[0].iov_base;
 	uint32_t map_size = mp_decode_map(&pos);
 	for (uint32_t i = 0; i < map_size; ++i)
@@ -1968,6 +2017,8 @@ struct replication_request {
 	/** IPROTO_INSTANCE_UUID. */
 	struct tt_uuid *instance_uuid;
 	/** IPROTO_VCLOCK. */
+	struct vclock *vclock_ignore0;
+	/** IPROTO_VCLOCK. */
 	struct vclock *vclock;
 	/** IPROTO_ID_FILTER. */
 	uint32_t *id_filter;
@@ -1975,6 +2026,12 @@ struct replication_request {
 	uint32_t *version_id;
 	/** IPROTO_REPLICA_ANON. */
 	bool *is_anon;
+	/** IPROTO_IS_CHECKPOINT_JOIN. */
+	bool *is_checkpoint_join;
+	/** IPROTO_CHECKPOINT_VCLOCK. */
+	struct vclock *checkpoint_vclock;
+	/** IPROTO_CHECKPOINT_LSN. */
+	uint64_t *checkpoint_lsn;
 };
 
 /** Encode a replication request template. */
@@ -1985,8 +2042,14 @@ xrow_encode_replication_request(struct xrow_header *row,
 {
 	memset(row, 0, sizeof(*row));
 	size_t size = XROW_BODY_LEN_MAX;
-	if (req->vclock != NULL)
-		size += mp_sizeof_vclock_ignore0(req->vclock);
+	if (req->vclock_ignore0 != NULL) {
+		size += mp_sizeof_vclock_ignore0(req->vclock_ignore0);
+		assert(req->vclock == NULL);
+	}
+	if (req->vclock != NULL) {
+		size += mp_sizeof_vclock(req->vclock);
+		assert(req->vclock_ignore0 == NULL);
+	}
 	char *buf = xregion_alloc(&fiber()->gc, size);
 	/* Skip one byte for future map header. */
 	char *data = buf + 1;
@@ -2001,10 +2064,15 @@ xrow_encode_replication_request(struct xrow_header *row,
 		data = mp_encode_uint(data, IPROTO_INSTANCE_UUID);
 		data = xrow_encode_uuid(data, req->instance_uuid);
 	}
+	if (req->vclock_ignore0 != NULL) {
+		++map_size;
+		data = mp_encode_uint(data, IPROTO_VCLOCK);
+		data = mp_encode_vclock_ignore0(data, req->vclock_ignore0);
+	}
 	if (req->vclock != NULL) {
 		++map_size;
 		data = mp_encode_uint(data, IPROTO_VCLOCK);
-		data = mp_encode_vclock_ignore0(data, req->vclock);
+		data = mp_encode_vclock(data, req->vclock);
 	}
 	if (req->version_id != NULL) {
 		++map_size;
@@ -2082,9 +2150,10 @@ xrow_decode_replication_request(const struct xrow_header *row,
 			}
 			break;
 		case IPROTO_VCLOCK:
-			if (req->vclock == NULL)
+			if (req->vclock_ignore0 == NULL)
 				goto skip;
-			if (mp_decode_vclock_ignore0(&d, req->vclock) != 0) {
+			if (mp_decode_vclock_ignore0(
+					&d, req->vclock_ignore0) != 0) {
 				xrow_on_decode_err(row, ER_INVALID_MSGPACK,
 						   "invalid VCLOCK");
 				return -1;
@@ -2128,6 +2197,36 @@ id_filter_decode_err:		xrow_on_decode_err(row, ER_INVALID_MSGPACK,
 				*req->id_filter |= 1 << val;
 			}
 			break;
+		case IPROTO_IS_CHECKPOINT_JOIN:
+			if (req->is_checkpoint_join == NULL)
+				goto skip;
+			if (mp_typeof(*d) != MP_BOOL) {
+				xrow_on_decode_err(
+					row, ER_INVALID_MSGPACK,
+					"invalid IS_CHECKPOINT_JOIN");
+				return -1;
+			}
+			*req->is_checkpoint_join = mp_decode_bool(&d);
+			break;
+		case IPROTO_CHECKPOINT_VCLOCK:
+			if (req->checkpoint_vclock == NULL)
+				goto skip;
+			if (mp_decode_vclock(&d, req->checkpoint_vclock) != 0) {
+				xrow_on_decode_err(row, ER_INVALID_MSGPACK,
+						   "invalid CHECKPOINT_VCLOCK");
+				return -1;
+			}
+			break;
+		case IPROTO_CHECKPOINT_LSN:
+			if (req->checkpoint_lsn == NULL)
+				goto skip;
+			if (mp_typeof(*d) != MP_UINT) {
+				xrow_on_decode_err(row, ER_INVALID_MSGPACK,
+						   "invalid CHECKPOINT_LSN");
+				return -1;
+			}
+			*req->checkpoint_lsn = mp_decode_uint(&d);
+			break;
 		default: skip:
 			mp_next(&d); /* value */
 		}
@@ -2142,7 +2241,7 @@ xrow_encode_register(struct xrow_header *row,
 	struct register_request *cast = (struct register_request *)req;
 	const struct replication_request base_req = {
 		.instance_uuid = &cast->instance_uuid,
-		.vclock = &cast->vclock,
+		.vclock_ignore0 = &cast->vclock,
 	};
 	xrow_encode_replication_request(row, &base_req, IPROTO_REGISTER);
 }
@@ -2154,7 +2253,7 @@ xrow_decode_register(const struct xrow_header *row,
 	memset(req, 0, sizeof(*req));
 	struct replication_request base_req = {
 		.instance_uuid = &req->instance_uuid,
-		.vclock = &req->vclock,
+		.vclock_ignore0 = &req->vclock,
 	};
 	return xrow_decode_replication_request(row, &base_req);
 }
@@ -2167,7 +2266,7 @@ xrow_encode_subscribe(struct xrow_header *row,
 	const struct replication_request base_req = {
 		.replicaset_uuid = &cast->replicaset_uuid,
 		.instance_uuid = &cast->instance_uuid,
-		.vclock = &cast->vclock,
+		.vclock_ignore0 = &cast->vclock,
 		.is_anon = &cast->is_anon,
 		.id_filter = &cast->id_filter,
 		.version_id = &cast->version_id,
@@ -2183,7 +2282,7 @@ xrow_decode_subscribe(const struct xrow_header *row,
 	struct replication_request base_req = {
 		.replicaset_uuid = &req->replicaset_uuid,
 		.instance_uuid = &req->instance_uuid,
-		.vclock = &req->vclock,
+		.vclock_ignore0 = &req->vclock,
 		.version_id = &req->version_id,
 		.is_anon = &req->is_anon,
 		.id_filter = &req->id_filter,
@@ -2213,6 +2312,38 @@ xrow_decode_join(const struct xrow_header *row, struct join_request *req)
 	return xrow_decode_replication_request(row, &base_req);
 }
 
+void
+xrow_encode_fetch_snapshot(struct xrow_header *row,
+			   const struct fetch_snapshot_request *req)
+{
+	struct fetch_snapshot_request *cast =
+		(struct fetch_snapshot_request *)req;
+	const struct replication_request base_req = {
+		.version_id = &cast->version_id,
+	};
+	xrow_encode_replication_request(row, &base_req, IPROTO_FETCH_SNAPSHOT);
+}
+
+int
+xrow_decode_fetch_snapshot(const struct xrow_header *row,
+			   struct fetch_snapshot_request *req)
+{
+	memset(req, 0, sizeof(*req));
+	struct replication_request base_req = {
+		.version_id = &req->version_id,
+		.is_checkpoint_join = &req->is_checkpoint_join,
+		.checkpoint_vclock = &req->checkpoint_vclock,
+		.checkpoint_lsn = &req->checkpoint_lsn,
+	};
+	/*
+	 * Vclock must be cleared, as it sets -1 signature, which cannot be
+	 * done by memset above. This is done in order to distinguish not
+	 * initialized vclock from the zero one.
+	 */
+	vclock_clear(&req->checkpoint_vclock);
+	return xrow_decode_replication_request(row, &base_req);
+}
+
 void
 xrow_encode_relay_heartbeat(struct xrow_header *row,
 			    const struct relay_heartbeat *req)
@@ -2381,6 +2512,15 @@ xrow_decode_applier_heartbeat(const struct xrow_header *row,
 	return 0;
 }
 
+void
+xrow_encode_vclock_ignore0(struct xrow_header *row, const struct vclock *vclock)
+{
+	const struct replication_request base_req = {
+		.vclock_ignore0 = (struct vclock *)vclock,
+	};
+	xrow_encode_replication_request(row, &base_req, IPROTO_OK);
+}
+
 void
 xrow_encode_vclock(struct xrow_header *row, const struct vclock *vclock)
 {
@@ -2391,11 +2531,11 @@ xrow_encode_vclock(struct xrow_header *row, const struct vclock *vclock)
 }
 
 int
-xrow_decode_vclock(const struct xrow_header *row, struct vclock *vclock)
+xrow_decode_vclock_ignore0(const struct xrow_header *row, struct vclock *vclock)
 {
 	vclock_create(vclock);
 	struct replication_request base_req = {
-		.vclock = vclock,
+		.vclock_ignore0 = vclock,
 	};
 	return xrow_decode_replication_request(row, &base_req);
 }
@@ -2407,7 +2547,7 @@ xrow_encode_subscribe_response(struct xrow_header *row,
 	struct subscribe_response *cast = (struct subscribe_response *)rsp;
 	const struct replication_request base_req = {
 		.replicaset_uuid = &cast->replicaset_uuid,
-		.vclock = &cast->vclock,
+		.vclock_ignore0 = &cast->vclock,
 	};
 	xrow_encode_replication_request(row, &base_req, IPROTO_OK);
 }
@@ -2419,7 +2559,7 @@ xrow_decode_subscribe_response(const struct xrow_header *row,
 	memset(rsp, 0, sizeof(*rsp));
 	struct replication_request base_req = {
 		.replicaset_uuid = &rsp->replicaset_uuid,
-		.vclock = &rsp->vclock,
+		.vclock_ignore0 = &rsp->vclock,
 	};
 	return xrow_decode_replication_request(row, &base_req);
 }
