@@ -37,6 +37,7 @@
 
 #include "diag.h"
 #include "iproto_features.h"
+#include "iproto_constants.h"
 #include "tt_uuid.h"
 #include "vclock/vclock.h"
 
@@ -49,8 +50,7 @@ enum {
 	XROW_BODY_IOVMAX = 2,
 	XROW_IOVMAX = XROW_HEADER_IOVMAX + XROW_BODY_IOVMAX,
 	XROW_HEADER_LEN_MAX = 52,
-	XROW_BODY_LEN_MAX = 256,
-	XROW_SYNCHRO_BODY_LEN_MAX = 32,
+	XROW_BODY_LEN_MAX = 512,
 	IPROTO_HEADER_LEN = 32,
 	/** 7 = sizeof(iproto_body_bin). */
 	IPROTO_SELECT_HEADER_LEN = IPROTO_HEADER_LEN + 7,
@@ -315,9 +315,14 @@ struct synchro_request {
 	int64_t lsn;
 	/**
 	 * The new term the instance issuing this request is in. Only used for
-	 * PROMOTE request.
+	 * PROMOTE and DEMOTE requests.
 	 */
 	uint64_t term;
+	/**
+	 * Confirmed lsns of all the previous limbo owners. Only used for
+	 * PROMOTE and DEMOTE requests.
+	 */
+	struct vclock *confirmed_vclock;
 };
 
 /**
@@ -334,11 +339,13 @@ xrow_encode_synchro(struct xrow_header *row, char *body,
  * Decode synchronous replication request.
  * @param row xrow header.
  * @param[out] req Request parameters.
+ * @param[out] vclock Storage for request vclock.
  * @retval -1 on error.
  * @retval 0 success.
  */
 int
-xrow_decode_synchro(const struct xrow_header *row, struct synchro_request *req);
+xrow_decode_synchro(const struct xrow_header *row, struct synchro_request *req,
+		    struct vclock *vclock);
 
 /**
  * Raft request. It repeats Raft message to the letter, but can be extended in
@@ -351,6 +358,8 @@ struct raft_request {
 	bool is_leader_seen;
 	uint64_t state;
 	const struct vclock *vclock;
+	/** Replication group identifier. 0 - replicaset, 1 - replica-local. */
+	enum group_id group_id;
 };
 
 void
@@ -605,6 +614,27 @@ xrow_encode_join(struct xrow_header *row, const struct join_request *req);
 int
 xrow_decode_join(const struct xrow_header *row, struct join_request *req);
 
+struct fetch_snapshot_request {
+	/** Replica's version. */
+	uint32_t version_id;
+	/** Flag indicating whether checkpoint join should be done. */
+	bool is_checkpoint_join;
+	/** Checkpoint's vclock, signature of the snapshot. */
+	struct vclock checkpoint_vclock;
+	/** Checkpoint's lsn, the row number to start from. */
+	uint64_t checkpoint_lsn;
+};
+
+/** Encode FETCH_SNAPSHOT request. */
+void
+xrow_encode_fetch_snapshot(struct xrow_header *row,
+			   const struct fetch_snapshot_request *req);
+
+/** Decode FETCH_SNAPSHOT request. */
+int
+xrow_decode_fetch_snapshot(const struct xrow_header *row,
+			   struct fetch_snapshot_request *req);
+
 /**
  * Heartbeat from relay to applier. Follows the replication stream. Same
  * direction.
@@ -650,13 +680,19 @@ int
 xrow_decode_applier_heartbeat(const struct xrow_header *row,
 			      struct applier_heartbeat *req);
 
-/** Encode vclock. */
+/** Encode vclock ignoring 0th component. */
+void
+xrow_encode_vclock_ignore0(struct xrow_header *row,
+			   const struct vclock *vclock);
+
+/** Encode vclock including 0th component. */
 void
 xrow_encode_vclock(struct xrow_header *row, const struct vclock *vclock);
 
-/** Decode vclock. */
+/** Decode vclock ignoring 0th component. */
 int
-xrow_decode_vclock(const struct xrow_header *row, struct vclock *vclock);
+xrow_decode_vclock_ignore0(const struct xrow_header *row,
+			   struct vclock *vclock);
 
 /**
  * Encode any bodyless message.
@@ -1107,6 +1143,15 @@ xrow_decode_join_xc(const struct xrow_header *row, struct join_request *req)
 		diag_raise();
 }
 
+/** @copydoc xrow_decode_fetch_snapshot. */
+static inline void
+xrow_decode_fetch_snapshot_xc(const struct xrow_header *row,
+			      struct fetch_snapshot_request *req)
+{
+	if (xrow_decode_fetch_snapshot(row, req) != 0)
+		diag_raise();
+}
+
 /** @copydoc xrow_decode_register. */
 static inline void
 xrow_decode_register_xc(const struct xrow_header *row,
@@ -1134,11 +1179,12 @@ xrow_decode_relay_heartbeat_xc(const struct xrow_header *row,
 		diag_raise();
 }
 
-/** @copydoc xrow_decode_vclock. */
+/** @copydoc xrow_decode_vclock_ignore0. */
 static inline void
-xrow_decode_vclock_xc(const struct xrow_header *row, struct vclock *vclock)
+xrow_decode_vclock_ignore0_xc(const struct xrow_header *row,
+			      struct vclock *vclock)
 {
-	if (xrow_decode_vclock(row, vclock) != 0)
+	if (xrow_decode_vclock_ignore0(row, vclock) != 0)
 		diag_raise();
 }
 
