@@ -31,7 +31,6 @@
 #include "say.h"
 #include "fiber.h"
 #include "errinj.h"
-#include "tt_static.h"
 #include "tt_strerror.h"
 
 #include <errno.h>
@@ -303,7 +302,7 @@ log_set_nonblock(struct log *log)
 static int
 log_rotate(struct log *log)
 {
-	if (pm_atomic_load(&log->type) != SAY_LOGGER_FILE)
+	if (log->type != SAY_LOGGER_FILE)
 		return 0;
 
 	ERROR_INJECT(ERRINJ_LOG_ROTATE, { usleep(10); });
@@ -627,23 +626,12 @@ log_syslog_init(struct log *log, const char *init_str)
 		return -1;
 
 	log->syslog_server_type = opts.server_type;
-	if (log->syslog_server_type != SAY_SYSLOG_DEFAULT) {
-		log->path = strdup(opts.server_path);
-		if (log->path == NULL) {
-			diag_set(OutOfMemory, strlen(opts.server_path),
-				 "malloc", "server address");
-			return -1;
-		}
-	}
+	if (log->syslog_server_type != SAY_SYSLOG_DEFAULT)
+		log->path = xstrdup(opts.server_path);
 	if (opts.identity == NULL)
-		log->syslog_ident = strdup("tarantool");
+		log->syslog_ident = xstrdup("tarantool");
 	else
-		log->syslog_ident = strdup(opts.identity);
-	if (log->syslog_ident == NULL) {
-		diag_set(OutOfMemory, strlen(opts.identity), "malloc",
-		         "log->syslog_ident");
-		return -1;
-	}
+		log->syslog_ident = xstrdup(opts.identity);
 
 	if (opts.facility == syslog_facility_MAX)
 		log->syslog_facility = SYSLOG_LOCAL7;
@@ -891,8 +879,8 @@ say_format_plain(struct log *log, char *buf, int len, int level,
 
 /**
  * Format log message in json format:
- * {"time": 1507026445.23232, "level": "WARN", "message": <message>,
- * "pid": <pid>, "cord_name": <name>, "fiber_id": <id>,
+ * {"time": "2024-12-23T20:04:26.491+0300", "level": "WARN",
+ * "message": <message>, "pid": <pid>, "cord_name": <name>, "fiber_id": <id>,
  * "fiber_name": <fiber_name>, file": <filename>, "line": <fds>,
  * "module": <module_name>}
  */
@@ -901,11 +889,16 @@ say_format_json(struct log *log, char *buf, int len, int level,
 		const char *module, const char *filename, int line,
 		const char *error, const char *format, va_list ap)
 {
-	(void) log;
+	(void)log;
+
 	int total = 0;
+	char *buf_end = buf + len;
 
+	/*
+	 * Print time in format YYYY-MM-DDThh:mm:ss.ms+tz_offset.
+	 * For example 2024-12-23T20:04:26.491+0300.
+	 */
 	SNPRINT(total, snprintf, buf, len, "{\"time\": \"");
-
 	struct tm tm;
 	double tm_sec;
 	get_current_time(&tm, &tm_sec);
@@ -916,38 +909,29 @@ say_format_json(struct log *log, char *buf, int len, int level,
 	buf += written, len -= written, total += written;
 	SNPRINT(total, snprintf, buf, len, "\", ");
 
+	/* Print level. */
 	SNPRINT(total, snprintf, buf, len, "\"level\": \"%s\", ",
-			level_to_string(level));
+		level_to_string(level));
 
-	if (strncmp(format, "json", sizeof("json")) == 0) {
-		/*
-		 * Message is already JSON-formatted.
-		 * Get rid of {} brackets and append to the output buffer.
-		 */
-		const char *str = va_arg(ap, const char *);
-		assert(str != NULL);
-		int str_len = strlen(str);
-		assert(str_len > 2 && str[0] == '{' && str[str_len - 1] == '}');
-		SNPRINT(total, snprintf, buf, len, "%.*s, ",
-			str_len - 2, str + 1);
-	} else {
-		/* Format message */
-		char *tmp = tt_static_buf();
-		if (vsnprintf(tmp, TT_STATIC_BUF_LEN, format, ap) < 0)
-			return -1;
-		SNPRINT(total, snprintf, buf, len, "\"message\": \"");
-		/* Escape and print message */
-		SNPRINT(total, json_escape, buf, len, tmp);
-		SNPRINT(total, snprintf, buf, len, "\", ");
-	}
+	/*
+	 * Remember where the message starts for now, will write the message,
+	 * after writing the rest of the context and moving it to the end of
+	 * the buffer.
+	 * Currently our buffer looks like this:
+	 * | head \0| garbage |
+	 *         ^ buf       ^ buf_end
+	 */
+	char *msg_ptr = buf;
+	const int head_len = total;
 
-	/* in case of system errors */
+	/* Print error, if any. */
 	if (error) {
 		SNPRINT(total, snprintf, buf, len, "\"error\": \"");
 		SNPRINT(total, json_escape, buf, len, error);
 		SNPRINT(total, snprintf, buf, len, "\", ");
 	}
 
+	/* Print PID, cord name, fiber id and fiber name. */
 	SNPRINT(total, snprintf, buf, len, "\"pid\": %i ", getpid());
 	SNPRINT(total, snprintf, buf, len, ", \"cord_name\": \"");
 	SNPRINT(total, json_escape, buf, len, cord()->name);
@@ -960,6 +944,7 @@ say_format_json(struct log *log, char *buf, int len, int level,
 		SNPRINT(total, snprintf, buf, len, "\"");
 	}
 
+	/* Print filename, line and module name if any. */
 	if (filename) {
 		SNPRINT(total, snprintf, buf, len, ", \"file\": \"");
 		SNPRINT(total, json_escape, buf, len, filename);
@@ -970,7 +955,92 @@ say_format_json(struct log *log, char *buf, int len, int level,
 		SNPRINT(total, json_escape, buf, len, module);
 		SNPRINT(total, snprintf, buf, len, "\"");
 	}
+
 	SNPRINT(total, snprintf, buf, len, "}\n");
+
+	/*
+	 * Finished printing context for the log entry. Will move tail to the
+	 * end of the buffer.
+	 * The buffer looks like this:
+	 * | head |   tail  \0| garbage |
+	 *         ^ msg_ptr ^ buf       ^ buf_end
+	 */
+
+	const int tail_len = total - head_len;
+	char *tail_ptr = buf_end - tail_len - 1; /* Reserve space for '\0'. */
+
+	/* Move the tail of the context to the end of the buffer. */
+	memmove(tail_ptr, msg_ptr, tail_len);
+	tail_ptr[tail_len] = '\0';
+	int msg_cap = tail_ptr - msg_ptr;
+
+	/* After moving the tail, the buffer looks like this:
+	 * | head |   garbage   |   tail   \0|
+	 *         ^ msg_ptr     ^ tail_ptr   ^ buf_end
+	 */
+
+	/* Write the message. */
+	if (strncmp(format, "json", sizeof("json")) == 0) {
+		/*
+		 * Message is already JSON-formatted.
+		 * Get rid of {} brackets and print it as-is.
+		 */
+		const char *str = va_arg(ap, const char *);
+		assert(str != NULL);
+		int str_len = strlen(str);
+		assert(str_len > 2 && str[0] == '{' && str[str_len - 1] == '}');
+
+		/*
+		 * Can't use SNPRINT macro here, because it sets the pointer to
+		 * null, if the message didn't fit into the provided buffer,
+		 * but we still need to use the pointer after.
+		 */
+		int msg_len = snprintf(msg_ptr, msg_cap,
+				       "%.*s, ", str_len - 2, str + 1);
+
+		if (msg_len < 0) {
+			return -1;
+		}
+
+		msg_len = MIN(msg_len, msg_cap);
+		len -= msg_len;
+		total += msg_len;
+		msg_ptr += msg_len;
+	} else {
+		/* Print message header. */
+		SNPRINT(total, snprintf, msg_ptr, msg_cap, "\"message\": \"");
+
+		static const char msg_tail[] = "\", ";
+		msg_cap -= strlen(msg_tail);
+
+		/*
+		 * Print the message.
+		 * Need to cast msg_cap to unsigned, because otherwise
+		 * during LTO build stringop-overread warning is triggered.
+		 * Can't use SNPRINT macro here, because it sets the pointer to
+		 * null, if the message didn't fit into the provided buffer,
+		 * but we still need to use the pointer after.
+		 */
+		vsnprintf(msg_ptr, (unsigned)msg_cap, format, ap);
+
+		/* Escape the message. */
+		int msg_len = json_escape_inplace(msg_ptr, msg_cap);
+		len -= msg_len;
+		total += msg_len;
+		msg_ptr += msg_len;
+
+		/* Print message tail. */
+		SNPRINT(total, snprintf, msg_ptr, len, msg_tail);
+	}
+
+	/*
+	 * We now will move the tail with '\0' to the end of the message.
+	 * After escaping the message, the buffer looks like this:
+	 * | head | message \0| garbage |   tail   \0|
+	 *                   ^ msg_ptr   ^ tail_ptr   ^ buf_end
+	 */
+	memmove(msg_ptr, tail_ptr, tail_len + 1);
+
 	return total;
 }
 
@@ -1318,9 +1388,11 @@ log_destroy(struct log *log)
 	while(log->rotating_threads > 0)
 		tt_pthread_cond_wait(&log->rotate_cond, &log->rotate_mutex);
 	tt_pthread_mutex_unlock(&log->rotate_mutex);
-	pm_atomic_store(&log->type, SAY_LOGGER_BOOT);
-
-	if (log->fd != -1)
+	/*
+	 * Do not close stderr because it's used for reporting crashes
+	 * and memory leaks.
+	 */
+	if (log->type != SAY_LOGGER_STDERR)
 		close(log->fd);
 	free(log->syslog_ident);
 	free(log->path);
