@@ -815,6 +815,10 @@ mp_check_uint(const char *cur, const char *end);
 MP_PROTO MP_PURE ptrdiff_t
 mp_check_int(const char *cur, const char *end);
 
+/** \brief Decode unsigned int tail with the first byte \a type already known. */
+MP_PROTO uint64_t
+mp_decode_uint_data(uint8_t type, const char **data);
+
 /**
  * \brief Decode an unsigned integer from MsgPack \a data
  * \param data - the pointer to a buffer
@@ -1549,6 +1553,28 @@ mp_memcpy(char *data, const char *str, uint32_t len);
 MP_PROTO char *
 mp_memcpy_safe(char *data, ptrdiff_t *data_sz, const char *str, uint32_t len);
 
+/**
+ * \brief Decode an integer value as int8_t from MsgPack \a data.
+ * \param data - the pointer to a buffer
+ * \param[out] ret - the pointer to save a result
+ * \retval  0 on success
+ * \retval -1 if underlying mp type is not MP_INT or MP_UINT
+ * \retval -1 if the result can't be stored in int8_t
+ */
+MP_PROTO int
+mp_read_int8(const char **data, int8_t *ret);
+
+/**
+ * \brief Decode an integer value as int16_t from MsgPack \a data.
+ * \param data - the pointer to a buffer
+ * \param[out] ret - the pointer to save a result
+ * \retval  0 on success
+ * \retval -1 if underlying mp type is not MP_INT or MP_UINT
+ * \retval -1 if the result can't be stored in int16_t
+ */
+MP_PROTO int
+mp_read_int16(const char **data, int16_t *ret);
+
 /**
  * \brief Decode an integer value as int32_t from MsgPack \a data.
  * \param data - the pointer to a buffer
@@ -1657,6 +1683,52 @@ mp_read_double_lossy(const char **data, double *ret);
 MP_PROTO void
 mp_next(const char **data);
 
+/** mp_check() error type. */
+enum mp_check_error_type {
+	/** Truncated MsgPack data. */
+	MP_CHECK_ERROR_TRUNC,
+	/** Illegal MsgPack code. */
+	MP_CHECK_ERROR_ILL,
+	/** MsgPack extension check failed. */
+	MP_CHECK_ERROR_EXT,
+	/** Junk after MsgPack data (mp_check_exact() only). */
+	MP_CHECK_ERROR_JUNK,
+};
+
+/** mp_check() error info. */
+struct mp_check_error {
+	/** Error type. */
+	enum mp_check_error_type type;
+	/** Start of the checked MsgPack data. */
+	const char *data;
+	/** End of the checked MsgPack data. */
+	const char *end;
+	/** Position where mp_check() failed. */
+	const char *pos;
+	/**
+	 * Number of missing MsgPack values.
+	 * Relevant only if the type is MP_CHECK_ERROR_TRUNC.
+	 */
+	int64_t trunc_count;
+	/**
+	 * Type of the MsgPack extension that failed the check.
+	 * Relevant only if the type is MP_CHECK_ERROR_EXT.
+	 */
+	int8_t ext_type;
+	/**
+	 * Length of the MsgPack extension that failed the check.
+	 * Relevant only if the type is MP_CHECK_ERROR_EXT.
+	 */
+	uint32_t ext_len;
+};
+
+typedef void (*mp_check_on_error_f)(const struct mp_check_error *err);
+
+/**
+ * \brief Function called by mp_check() on error.
+ */
+extern mp_check_on_error_f mp_check_on_error;
+
 /**
  * \brief Equivalent to mp_next() but also validates MsgPack in \a data.
  * \param data - the pointer to a buffer
@@ -1670,6 +1742,22 @@ mp_next(const char **data);
 MP_PROTO int
 mp_check(const char **data, const char *end);
 
+/**
+ * \brief Equivalent to mp_check() but also checks that there is no more
+ * data after validated MsgPack.
+ * \param data - the pointer to a buffer
+ * \param end - the end of a buffer
+ * \retval 0 when MsgPack in \a data is valid and there is no more data
+ * after validated MsgPack.
+ * \retval != 0 when MsgPack in \a data is not valid or there is data
+ * after validated MsgPack.
+ * \post *data = end on success
+ * \post *data is not defined if MsgPack is not valid
+ * \sa mp_check()
+ */
+MP_PROTO int
+mp_check_exact(const char **data, const char *end);
+
 /**
  * The maximum msgpack nesting depth supported by mp_snprint().
  * Everything beyond that will be omitted (replaced with "...").
@@ -2191,10 +2279,9 @@ mp_encode_int_safe(char *data, ptrdiff_t *data_sz, int64_t num)
 }
 
 MP_IMPL uint64_t
-mp_decode_uint(const char **data)
+mp_decode_uint_data(uint8_t type, const char **data)
 {
-	uint8_t c = mp_load_u8(data);
-	switch (c) {
+	switch (type) {
 	case 0xcc:
 		return mp_load_u8(data);
 	case 0xcd:
@@ -2204,50 +2291,50 @@ mp_decode_uint(const char **data)
 	case 0xcf:
 		return mp_load_u64(data);
 	default:
-		if (mp_unlikely(c > 0x7f))
+		if (mp_unlikely(type > 0x7f))
 			mp_unreachable();
-		return c;
+		return type;
 	}
 }
 
+MP_IMPL uint64_t
+mp_decode_uint(const char **data)
+{
+	return mp_decode_uint_data(mp_load_u8(data), data);
+}
+
 MP_IMPL int
 mp_compare_uint(const char *data_a, const char *data_b)
 {
 	uint8_t ca = mp_load_u8(&data_a);
 	uint8_t cb = mp_load_u8(&data_b);
-
-	int r = ca - cb;
-	if (r != 0)
-		return r;
-
-	if (ca <= 0x7f)
-		return 0;
-
 	uint64_t a, b;
-	switch (ca & 0x3) {
-	case 0xcc & 0x3:
+	if (ca != cb) {
+		a = mp_decode_uint_data(ca, &data_a);
+		b = mp_decode_uint_data(cb, &data_b);
+		return a < b ? -1 : a > b;
+	}
+	switch (ca) {
+	case 0xcc:
 		a = mp_load_u8(&data_a);
 		b = mp_load_u8(&data_b);
 		break;
-	case 0xcd & 0x3:
+	case 0xcd:
 		a = mp_load_u16(&data_a);
 		b = mp_load_u16(&data_b);
 		break;
-	case 0xce & 0x3:
+	case 0xce:
 		a = mp_load_u32(&data_a);
 		b = mp_load_u32(&data_b);
 		break;
-	case 0xcf & 0x3:
+	case 0xcf:
 		a = mp_load_u64(&data_a);
 		b = mp_load_u64(&data_b);
-		return a < b ? -1 : a > b;
 		break;
 	default:
-		mp_unreachable();
+		return 0;
 	}
-
-	int64_t v = (a - b);
-	return (v > 0) - (v < 0);
+	return a < b ? -1 : a > b;
 }
 
 MP_IMPL int64_t
@@ -2720,39 +2807,158 @@ mp_decode_bool(const char **data)
 }
 
 MP_IMPL int
-mp_read_int32(const char **data, int32_t *ret)
+mp_read_int8(const char **data, int8_t *ret)
 {
-	uint32_t uval;
+	int64_t ival;
+	uint64_t uval;
 	const char *p = *data;
 	uint8_t c = mp_load_u8(&p);
 	switch (c) {
 	case 0xd0:
-		*ret = (int8_t) mp_load_u8(&p);
-		break;
+		*ret = (int8_t)mp_load_u8(&p);
+		goto commit;
 	case 0xd1:
-		*ret = (int16_t) mp_load_u16(&p);
-		break;
+		ival = (int16_t)mp_load_u16(&p);
+		goto handle_signed;
 	case 0xd2:
-		*ret = (int32_t) mp_load_u32(&p);
-		break;
+		ival = (int32_t)mp_load_u32(&p);
+		goto handle_signed;
+	case 0xd3:
+		ival = (int64_t)mp_load_u64(&p);
+		goto handle_signed;
 	case 0xcc:
-		*ret = mp_load_u8(&p);
-		break;
+		uval = mp_load_u8(&p);
+		goto handle_unsigned;
 	case 0xcd:
-		*ret = mp_load_u16(&p);
-		break;
+		uval = mp_load_u16(&p);
+		goto handle_unsigned;
 	case 0xce:
 		uval = mp_load_u32(&p);
-		if (mp_unlikely(uval > INT32_MAX))
-			return -1;
-		*ret = uval;
-		break;
+		goto handle_unsigned;
+	case 0xcf:
+		uval = mp_load_u64(&p);
+		goto handle_unsigned;
 	default:
 		if (mp_unlikely(c < 0xe0 && c > 0x7f))
 			return -1;
 		*ret = (int8_t) c;
-		break;
+		goto commit;
 	}
+handle_signed:
+	if (mp_unlikely(ival > INT8_MAX || ival < INT8_MIN))
+		return -1;
+	*ret = (int8_t)ival;
+	goto commit;
+handle_unsigned:
+	if (mp_unlikely(uval > INT8_MAX))
+		return -1;
+	*ret = (int8_t)uval;
+	goto commit;
+commit:
+	*data = p;
+	return 0;
+}
+
+MP_IMPL int
+mp_read_int16(const char **data, int16_t *ret)
+{
+	int64_t ival;
+	uint64_t uval;
+	const char *p = *data;
+	uint8_t c = mp_load_u8(&p);
+	switch (c) {
+	case 0xd0:
+		*ret = (int8_t)mp_load_u8(&p);
+		goto commit;
+	case 0xd1:
+		*ret = (int16_t)mp_load_u16(&p);
+		goto commit;
+	case 0xd2:
+		ival = (int32_t)mp_load_u32(&p);
+		goto handle_signed;
+	case 0xd3:
+		ival = (int64_t)mp_load_u64(&p);
+		goto handle_signed;
+	case 0xcc:
+		*ret = (int16_t)mp_load_u8(&p);
+		goto commit;
+	case 0xcd:
+		uval = mp_load_u16(&p);
+		goto handle_unsigned;
+	case 0xce:
+		uval = mp_load_u32(&p);
+		goto handle_unsigned;
+	case 0xcf:
+		uval = mp_load_u64(&p);
+		goto handle_unsigned;
+	default:
+		if (mp_unlikely(c < 0xe0 && c > 0x7f))
+			return -1;
+		*ret = (int8_t) c;
+		goto commit;
+	}
+handle_signed:
+	if (mp_unlikely(ival > INT16_MAX || ival < INT16_MIN))
+		return -1;
+	*ret = (int16_t)ival;
+	goto commit;
+handle_unsigned:
+	if (mp_unlikely(uval > INT16_MAX))
+		return -1;
+	*ret = (int16_t)uval;
+	goto commit;
+commit:
+	*data = p;
+	return 0;
+}
+
+MP_IMPL int
+mp_read_int32(const char **data, int32_t *ret)
+{
+	int64_t ival;
+	uint64_t uval;
+	const char *p = *data;
+	uint8_t c = mp_load_u8(&p);
+	switch (c) {
+	case 0xd0:
+		*ret = (int8_t)mp_load_u8(&p);
+		goto commit;
+	case 0xd1:
+		*ret = (int16_t)mp_load_u16(&p);
+		goto commit;
+	case 0xd2:
+		*ret = (int32_t)mp_load_u32(&p);
+		goto commit;
+	case 0xd3:
+		ival = (int64_t)mp_load_u64(&p);
+		if (mp_unlikely(ival > INT32_MAX || ival < INT32_MIN))
+			return -1;
+		*ret = (int32_t)ival;
+		goto commit;
+	case 0xcc:
+		*ret = (int32_t)mp_load_u8(&p);
+		goto commit;
+	case 0xcd:
+		*ret = (int32_t)mp_load_u16(&p);
+		goto commit;
+	case 0xce:
+		uval = mp_load_u32(&p);
+		goto handle_unsigned;
+	case 0xcf:
+		uval = mp_load_u64(&p);
+		goto handle_unsigned;
+	default:
+		if (mp_unlikely(c < 0xe0 && c > 0x7f))
+			return -1;
+		*ret = (int8_t) c;
+		goto commit;
+	}
+handle_unsigned:
+	if (mp_unlikely(uval > INT32_MAX))
+		return -1;
+	*ret = (int32_t)uval;
+	goto commit;
+commit:
 	*data = p;
 	return 0;
 }
@@ -2765,31 +2971,31 @@ mp_read_int64(const char **data, int64_t *ret)
 	uint8_t c = mp_load_u8(&p);
 	switch (c) {
 	case 0xd0:
-		*ret = (int8_t) mp_load_u8(&p);
+		*ret = (int8_t)mp_load_u8(&p);
 		break;
 	case 0xd1:
-		*ret = (int16_t) mp_load_u16(&p);
+		*ret = (int16_t)mp_load_u16(&p);
 		break;
 	case 0xd2:
-		*ret = (int32_t) mp_load_u32(&p);
+		*ret = (int32_t)mp_load_u32(&p);
 		break;
 	case 0xd3:
-		*ret = (int64_t) mp_load_u64(&p);
+		*ret = (int64_t)mp_load_u64(&p);
 		break;
 	case 0xcc:
-		*ret = mp_load_u8(&p);
+		*ret = (int64_t)mp_load_u8(&p);
 		break;
 	case 0xcd:
-		*ret = mp_load_u16(&p);
+		*ret = (int64_t)mp_load_u16(&p);
 		break;
 	case 0xce:
-		*ret = mp_load_u32(&p);
+		*ret = (int64_t)mp_load_u32(&p);
 		break;
 	case 0xcf:
 		uval = mp_load_u64(&p);
-		if (uval > INT64_MAX)
+		if (mp_unlikely(uval > INT64_MAX))
 			return -1;
-		*ret = uval;
+		*ret = (int64_t)uval;
 		break;
 	default:
 		if (mp_unlikely(c < 0xe0 && c > 0x7f))
@@ -3032,12 +3238,36 @@ mp_next(const char **data)
 MP_IMPL int
 mp_check(const char **data, const char *end)
 {
-#define MP_CHECK_LEN(_l) \
-	if (mp_unlikely((size_t)(end - *data) < (size_t)(_l))) \
-		return 1;
+	const char *begin = *data;
+
+#define MP_CHECK_LEN(_l)						\
+	if (mp_unlikely((size_t)(end - *data) < (size_t)(_l))) {	\
+		struct mp_check_error err;				\
+		err.type = MP_CHECK_ERROR_TRUNC;			\
+		err.data = begin;					\
+		err.end = end;						\
+		err.pos = pos;						\
+		err.trunc_count = k;					\
+		mp_check_on_error(&err);				\
+		return 1;						\
+	}
+
+#define MP_CHECK_EXT(_type, _data, _len)				\
+	if (mp_check_ext_data((_type), (_data), (_len)) != 0) {		\
+		struct mp_check_error err;				\
+		err.type = MP_CHECK_ERROR_EXT;				\
+		err.data = begin;					\
+		err.end = end;						\
+		err.pos = (_data);					\
+		err.ext_type = (_type);					\
+		err.ext_len = (_len);					\
+		mp_check_on_error(&err);				\
+		return 1;						\
+	}
 
 	int64_t k;
 	for (k = 1; k > 0; k--) {
+		const char *pos = *data;
 		MP_CHECK_LEN(1);
 		uint8_t c = mp_load_u8(data);
 		int l = mp_parser_hint[c];
@@ -3053,8 +3283,7 @@ mp_check(const char **data, const char *end)
 				 */
 				len = l - 1;
 				type = mp_load_u8(data);
-				if (mp_check_ext_data(type, *data, len) != 0)
-					return 1;
+				MP_CHECK_EXT(type, *data, len);
 				*data += len;
 			} else {
 				*data += l;
@@ -3113,8 +3342,7 @@ mp_check(const char **data, const char *end)
 			len = mp_load_u8(data);
 			type = mp_load_u8(data);
 			MP_CHECK_LEN(len);
-			if (mp_check_ext_data(type, *data, len) != 0)
-				return 1;
+			MP_CHECK_EXT(type, *data, len);
 			*data += len;
 			break;
 		case MP_HINT_EXT_16:
@@ -3123,8 +3351,7 @@ mp_check(const char **data, const char *end)
 			len = mp_load_u16(data);
 			type = mp_load_u8(data);
 			MP_CHECK_LEN(len);
-			if (mp_check_ext_data(type, *data, len) != 0)
-				return 1;
+			MP_CHECK_EXT(type, *data, len);
 			*data += len;
 			break;
 		case MP_HINT_EXT_32:
@@ -3133,12 +3360,19 @@ mp_check(const char **data, const char *end)
 			len = mp_load_u32(data);
 			type = mp_load_u8(data);
 			MP_CHECK_LEN(len);
-			if (mp_check_ext_data(type, *data, len) != 0)
-				return 1;
+			MP_CHECK_EXT(type, *data, len);
 			*data += len;
 			break;
 		case MP_HINT_INVALID:
+		{
+			struct mp_check_error err;
+			err.type = MP_CHECK_ERROR_ILL;
+			err.data = begin;
+			err.end = end;
+			err.pos = pos;
+			mp_check_on_error(&err);
 			return 1;
+		}
 		default:
 			mp_unreachable();
 		}
@@ -3146,6 +3380,25 @@ mp_check(const char **data, const char *end)
 
 	assert(*data <= end);
 #undef MP_CHECK_LEN
+#undef MP_CHECK_EXT
+	return 0;
+}
+
+MP_IMPL int
+mp_check_exact(const char **data, const char *end)
+{
+	const char *begin = *data;
+	if (mp_check(data, end) != 0)
+		return 1;
+	if (*data != end) {
+		struct mp_check_error err;
+		err.type = MP_CHECK_ERROR_JUNK;
+		err.data = begin;
+		err.end = end;
+		err.pos = *data;
+		mp_check_on_error(&err);
+		return 1;
+	}
 	return 0;
 }
 
