@@ -216,10 +216,8 @@ static const uint32_t LIGHT(end) = 0xFFFFFFFF;
  * @param arg - optional parameter to save for comparing function
  */
 static inline void
-LIGHT(create)(struct LIGHT(core) *ht, size_t extent_size,
-	      LIGHT(extent_alloc_t) extent_alloc_func,
-	      LIGHT(extent_free_t) extent_free_func,
-	      void *alloc_ctx, LIGHT_CMP_ARG_TYPE arg);
+LIGHT(create)(struct LIGHT(core) *ht, LIGHT_CMP_ARG_TYPE arg,
+	      struct matras_allocator *allocator);
 
 /**
  * @brief Hash table destruction. Frees all allocated memory
@@ -443,10 +441,8 @@ LIGHT(view_iterator_get_and_next)(const struct LIGHT(view) *v,
  * @param arg - optional parameter to save for comparing function
  */
 static inline void
-LIGHT(create)(struct LIGHT(core) *htab, size_t extent_size,
-	      LIGHT(extent_alloc_t) extent_alloc_func,
-	      LIGHT(extent_free_t) extent_free_func,
-	      void *alloc_ctx, LIGHT_CMP_ARG_TYPE arg)
+LIGHT(create)(struct LIGHT(core) *htab, LIGHT_CMP_ARG_TYPE arg,
+	      struct matras_allocator *allocator)
 {
 	struct LIGHT(common) *ht = &htab->common;
 	assert((LIGHT_GROW_INCREMENT & (LIGHT_GROW_INCREMENT - 1)) == 0);
@@ -455,9 +451,8 @@ LIGHT(create)(struct LIGHT(core) *htab, size_t extent_size,
 	ht->table_size = 0;
 	ht->empty_slot = LIGHT(end);
 	ht->arg = arg;
-	matras_create(&htab->mtable,
-		      extent_size, sizeof(struct LIGHT(record)),
-		      extent_alloc_func, extent_free_func, alloc_ctx, NULL);
+	matras_create(&htab->mtable, sizeof(struct LIGHT(record)),
+		      allocator, NULL);
 	matras_head_read_view(&htab->view);
 	ht->mtable = &htab->mtable;
 	ht->view = &htab->view;
@@ -529,11 +524,14 @@ LIGHT(slot)(const struct LIGHT(common) *ht, uint32_t hash)
 {
 	uint32_t cover_mask = ht->cover_mask;
 	uint32_t res = hash & cover_mask;
-	uint32_t probe = (ht->table_size - res - 1) >> 31;
+	/*
+	 * Calculate the following expression without branch instructions:
+	 * probe = res >= ht->table_size;
+	 */
+	uint32_t probe = ((size_t)ht->table_size - res - 1) >> 63;
 	uint32_t shift = __builtin_ctz(~(cover_mask >> 1));
 	res ^= (probe << shift);
 	return res;
-
 }
 
 /**
@@ -845,6 +843,15 @@ LIGHT(grow)(struct LIGHT(common) *ht)
 {
 	assert(!matras_is_read_view_created(ht->view));
 	assert(ht->empty_slot == LIGHT(end));
+	/*
+	 * The number UINT32_MAX has a special meaning (see LIGHT(end)), hence
+	 * it can not be used as a record identifier. Given that the table is
+	 * enlarged by 8 records (see LIGHT_GROW_INCREMENT), the maximum table
+	 * size is limited by (2^32)-8 records.
+	 */
+	if ((size_t)ht->table_size + LIGHT_GROW_INCREMENT >= UINT32_MAX)
+		return -1;
+
 	uint32_t new_slot;
 	struct LIGHT(record) *new_record = (struct LIGHT(record) *)
 		matras_alloc_range(ht->mtable, &new_slot, LIGHT_GROW_INCREMENT);
