@@ -52,6 +52,13 @@ const char compat_lua[] =
 "https://tarantool.io/compat/sql_seq_scan_default\n"
 "]]\n"
 "\n"
+"local C_FUNC_IPROTO_MULTIRETURN_BRIEF = [[\n"
+"Whether the results of the C stored function should be encoded with an\n"
+"additional msgpack array when returning them via iproto.\n"
+"\n"
+"https://tarantool.io/compat/c_func_iproto_multireturn\n"
+"]]\n"
+"\n"
 "-- Returns an action callback that toggles a tweak.\n"
 "local function tweak_action(tweak_name, old_tweak_value, new_tweak_value)\n"
 "    return function(is_new)\n"
@@ -97,6 +104,13 @@ const char compat_lua[] =
 "        brief = SQL_SEQ_SCAN_DEFAULT_BRIEF,\n"
 "        action = tweak_action('sql_seq_scan_default', true, false),\n"
 "    },\n"
+"    c_func_iproto_multireturn = {\n"
+"        default = 'old',\n"
+"        obsolete = nil,\n"
+"        brief = C_FUNC_IPROTO_MULTIRETURN_BRIEF,\n"
+"        run_action_now = true,\n"
+"        action = tweak_action('c_func_iproto_multireturn', false, true),\n"
+"    },\n"
 "}\n"
 "\n"
 "-- Array with option names in order of addition.\n"
