@@ -8,7 +8,7 @@ const char datetime_lua[] =
 "    values, where `dt` is a number of dates since Rata Die date (0001-01-01).\n"
 "\n"
 "    `c-dt` uses 32-bit integer number to store `dt` values, so range of\n"
-"    suported dates is limited to dates from -5879610-06-22 (INT32_MIN) to\n"
+"    supported dates is limited to dates from -5879610-06-22 (INT32_MIN) to\n"
 "    +5879611-07-11 (INT32_MAX).\n"
 "\n"
 "    For better compactness of our typical data in MessagePack stream we shift\n"
@@ -70,8 +70,7 @@ const char datetime_lua[] =
 "size_t tnt_datetime_strftime(const struct datetime *date, char *buf,\n"
 "                             uint32_t len, const char *fmt);\n"
 "ssize_t tnt_datetime_parse_full(struct datetime *date, const char *str,\n"
-"                                size_t len, const char *tzsuffix,\n"
-"                                int32_t offset);\n"
+"                                size_t len);\n"
 "ssize_t tnt_datetime_parse_tz(const char *str, size_t len, time_t base,\n"
 "                              int16_t *tzoffset, int16_t *tzindex);\n"
 "size_t tnt_datetime_strptime(struct datetime *date, const char *buf,\n"
@@ -254,7 +253,7 @@ const char datetime_lua[] =
 "end\n"
 "\n"
 "-- range may be of a form of pair {from, to} or\n"
-"-- tuple {fom, to, -1 in extra}\n"
+"-- tuple {from, to, -1 in extra}\n"
 "-- -1 is a special value (so far) used for days only\n"
 "local function check_range(v, from, to, txt, extra)\n"
 "    if type(v) ~= 'number' then\n"
@@ -306,6 +305,9 @@ const char datetime_lua[] =
 "    return b and 1 or 0\n"
 "end\n"
 "\n"
+"-- Forward declaration.\n"
+"local datetime_set\n"
+"\n"
 "local adjust_xlat = {\n"
 "    none = builtin.DT_LIMIT,\n"
 "    last = builtin.DT_SNAP,\n"
@@ -454,10 +456,7 @@ const char datetime_lua[] =
 "    return builtin.tnt_datetime_isdst(obj)\n"
 "end\n"
 "\n"
-"--[[\n"
-"    Parse timezone name similar way as datetime_parse_full parse\n"
-"    full literal.\n"
-"]]\n"
+"-- Parse timezone name.\n"
 "local function parse_tzname(base_epoch, tzname)\n"
 "    check_str(tzname, 'parse_tzname()')\n"
 "    local ptzindex = date_int16_stash_take()\n"
@@ -505,6 +504,7 @@ const char datetime_lua[] =
 "\n"
 "local function get_timezone(offset, msg)\n"
 "    if type(offset) == 'number' then\n"
+"        check_integer(offset, 'tzoffset')\n"
 "        return offset\n"
 "    elseif type(offset) == 'string' then\n"
 "        return parse_tzoffset(offset)\n"
@@ -528,31 +528,37 @@ const char datetime_lua[] =
 "    local y = obj.year\n"
 "    if y ~= nil then\n"
 "        check_range(y, MIN_DATE_YEAR, MAX_DATE_YEAR, 'year')\n"
+"        check_integer(y, 'year')\n"
 "        ymd = true\n"
 "    end\n"
 "    local M = obj.month\n"
 "    if M ~= nil then\n"
 "        check_range(M, 1, 12, 'month')\n"
+"        check_integer(M, 'month')\n"
 "        ymd = true\n"
 "    end\n"
 "    local d = obj.day\n"
 "    if d ~= nil then\n"
 "        check_range(d, 1, 31, 'day', -1)\n"
+"        check_integer(d, 'day')\n"
 "        ymd = true\n"
 "    end\n"
 "    local h = obj.hour\n"
 "    if h ~= nil then\n"
 "        check_range(h, 0, 23, 'hour')\n"
+"        check_integer(h, 'hour')\n"
 "        hms = true\n"
 "    end\n"
 "    local m = obj.min\n"
 "    if m ~= nil then\n"
 "        check_range(m, 0, 59, 'min')\n"
+"        check_integer(m, 'min')\n"
 "        hms = true\n"
 "    end\n"
 "    local s = obj.sec\n"
 "    if s ~= nil then\n"
 "        check_range(s, 0, 60, 'sec')\n"
+"        check_integer(s, 'sec')\n"
 "        hms = true\n"
 "    end\n"
 "\n"
@@ -566,12 +572,15 @@ const char datetime_lua[] =
 "        end\n"
 "        if usec ~= nil then\n"
 "            check_range(usec, 0, 1e6, 'usec')\n"
+"            check_integer(usec, 'usec')\n"
 "            nsec = usec * 1e3\n"
 "        elseif msec ~= nil then\n"
 "            check_range(msec, 0, 1e3, 'msec')\n"
+"            check_integer(msec, 'msec')\n"
 "            nsec = msec * 1e6\n"
 "        else\n"
 "            check_range(nsec, 0, 1e9, 'nsec')\n"
+"            check_integer(nsec, 'nsec')\n"
 "        end\n"
 "    else\n"
 "        nsec = 0\n"
@@ -863,10 +872,10 @@ const char datetime_lua[] =
 "       date [T] time [ ] time_zone\n"
 "    Returns constructed datetime object and length of accepted string.\n"
 "]]\n"
-"local function datetime_parse_full(str, tzname, offset)\n"
+"local function datetime_parse_full(str)\n"
 "    check_str(str, 'datetime.parse()')\n"
 "    local date = ffi.new(datetime_t)\n"
-"    local len = builtin.tnt_datetime_parse_full(date, str, #str, tzname, offset)\n"
+"    local len = builtin.tnt_datetime_parse_full(date, str, #str)\n"
 "    if len > 0 then\n"
 "        return date, tonumber(len)\n"
 "    elseif len == -builtin.TZ_NYI then\n"
@@ -905,9 +914,8 @@ const char datetime_lua[] =
 "    end\n"
 "    check_str_or_nil(fmt, 'datetime.parse()')\n"
 "\n"
-"    local offset = 0\n"
 "    if tzoffset ~= nil then\n"
-"        offset = get_timezone(tzoffset, 'tzoffset')\n"
+"        local offset = get_timezone(tzoffset, 'tzoffset')\n"
 "        check_range(offset, -720, 840, 'tzoffset')\n"
 "    end\n"
 "\n"
@@ -915,12 +923,24 @@ const char datetime_lua[] =
 "        check_str(tzname, 'datetime.parse()')\n"
 "    end\n"
 "\n"
-"    if not fmt or fmt == '' or fmt == 'iso8601' or fmt == 'rfc3339' then\n"
-"        -- Effect of .tz overrides .tzoffset\n"
-"        return datetime_parse_full(str, tzname, offset)\n"
-"    else\n"
-"        return datetime_parse_format(str, fmt)\n"
+"    if tzoffset and tzname then\n"
+"        error(\"ambiguous timezone: both tzoffset and tz are specified\")\n"
 "    end\n"
+"\n"
+"    local date, len\n"
+"    if not fmt or fmt == '' or fmt == 'iso8601' or fmt == 'rfc3339' then\n"
+"        date, len = datetime_parse_full(str)\n"
+"    else\n"
+"        date, len = datetime_parse_format(str, fmt)\n"
+"    end\n"
+"\n"
+"    -- Override timezone, if it was not specified in a parsed\n"
+"    -- string.\n"
+"    if date.tz == '' and date.tzoffset == 0 then\n"
+"        datetime_set(date, { tzoffset = tzoffset, tz = tzname })\n"
+"    end\n"
+"\n"
+"    return date, len\n"
 "end\n"
 "\n"
 "--[[\n"
@@ -974,6 +994,8 @@ const char datetime_lua[] =
 "        isdst = datetime_isdst(self),\n"
 "        nsec = self.nsec,\n"
 "        tzoffset = self.tzoffset,\n"
+"        tz = self.tz,\n"
+"        timestamp = self.timestamp,\n"
 "    }\n"
 "end\n"
 "\n"
@@ -1003,7 +1025,7 @@ const char datetime_lua[] =
 "    self.epoch = secs_day + h * 3600 + m * 60 + s\n"
 "end\n"
 "\n"
-"local function datetime_set(self, obj)\n"
+"function datetime_set(self, obj)\n"
 "    check_date(self, 'datetime.set()')\n"
 "    check_table(obj, \"datetime.set()\")\n"
 "\n"
@@ -1020,16 +1042,19 @@ const char datetime_lua[] =
 "    local y = obj.year\n"
 "    if y ~= nil then\n"
 "        check_range(y, MIN_DATE_YEAR, MAX_DATE_YEAR, 'year')\n"
+"        check_integer(y, 'year')\n"
 "        ymd = true\n"
 "    end\n"
 "    local M = obj.month\n"
 "    if M ~= nil then\n"
 "        check_range(M, 1, 12, 'month')\n"
+"        check_integer(M, 'month')\n"
 "        ymd = true\n"
 "    end\n"
 "    local d = obj.day\n"
 "    if d ~= nil then\n"
 "        check_range(d, 1, 31, 'day', -1)\n"
+"        check_integer(d, 'day')\n"
 "        ymd = true\n"
 "    end\n"
 "\n"
@@ -1041,16 +1066,19 @@ const char datetime_lua[] =
 "    local h = obj.hour\n"
 "    if h ~= nil then\n"
 "        check_range(h, 0, 23, 'hour')\n"
+"        check_integer(h, 'hour')\n"
 "        hms = true\n"
 "    end\n"
 "    local m = obj.min\n"
 "    if m ~= nil then\n"
 "        check_range(m, 0, 59, 'min')\n"
+"        check_integer(m, 'min')\n"
 "        hms = true\n"
 "    end\n"
 "    local sec = obj.sec\n"
 "    if sec ~= nil then\n"
 "        check_range(sec, 0, 60, 'sec')\n"
+"        check_integer(sec, 'sec')\n"
 "        hms = true\n"
 "    end\n"
 "\n"
@@ -1064,12 +1092,15 @@ const char datetime_lua[] =
 "        end\n"
 "        if usec ~= nil then\n"
 "            check_range(usec, 0, 1e6, 'usec')\n"
+"            check_integer(usec, 'usec')\n"
 "            self.nsec = usec * 1e3\n"
 "        elseif msec ~= nil then\n"
 "            check_range(msec, 0, 1e3, 'msec')\n"
+"            check_integer(msec, 'msec')\n"
 "            self.nsec = msec * 1e6\n"
 "        elseif nsec ~= nil then\n"
 "            check_range(nsec, 0, 1e9, 'nsec')\n"
+"            check_integer(nsec, 'nsec')\n"
 "            self.nsec = nsec\n"
 "        end\n"
 "    end\n"
