@@ -88,14 +88,12 @@ luaL_testcancel(struct lua_State *L)
 static const char *fiberlib_name = "fiber";
 
 /**
- * Trigger invoked when the fiber has stopped execution of its
- * current request. Only purpose - delete storage.lua.fid_ref and
- * storage.lua.storage_ref keeping a reference of Lua
- * fiber and fiber.storage objects. Unlike Lua stack,
- * Lua fiber storage may be created not only for fibers born from
- * Lua land. For example, an IProto request may execute a Lua
- * function, which can create the storage. Trigger guarantees,
- * that even for non-Lua fibers the Lua storage is destroyed.
+ * Trigger invoked when the fiber is stopped. Only purpose - delete
+ * storage.lua.storage_ref keeping a reference of Lua fiber.storage object.
+ * Unlike Lua stack, Lua fiber storage may be created not only for fibers born
+ * from Lua land. For example, an IProto request may execute a Lua function,
+ * which can create the storage. Trigger guarantees, that even for non-Lua
+ * fibers the Lua storage is destroyed.
  */
 static int
 lbox_fiber_on_stop(struct trigger *trigger, void *event)
@@ -103,6 +101,16 @@ lbox_fiber_on_stop(struct trigger *trigger, void *event)
 	struct fiber *f = event;
 	luaL_unref(tarantool_L, LUA_REGISTRYINDEX, f->storage.lua.storage_ref);
 	f->storage.lua.storage_ref = FIBER_LUA_NOREF;
+	trigger_clear(trigger);
+	free(trigger);
+	return 0;
+}
+
+/** Cleanup Lua fiber data when fiber is destroyed. */
+static int
+lbox_fiber_on_destroy(struct trigger *trigger, void *event)
+{
+	struct fiber *f = event;
 	luaL_unref(tarantool_L, LUA_REGISTRYINDEX, f->storage.lua.fid_ref);
 	f->storage.lua.fid_ref = FIBER_LUA_NOREF;
 	trigger_clear(trigger);
@@ -118,13 +126,10 @@ lbox_pushfiber(struct lua_State *L, struct fiber *f)
 {
 	int fid_ref = f->storage.lua.fid_ref;
 	if (fid_ref == FIBER_LUA_NOREF) {
-		struct trigger *t = malloc(sizeof(*t));
-		if (t == NULL) {
-			diag_set(OutOfMemory, sizeof(*t), "malloc", "t");
-			luaT_error(L);
-		}
-		trigger_create(t, lbox_fiber_on_stop, NULL, (trigger_f0)free);
-		trigger_add(&f->on_stop, t);
+		struct trigger *on_destroy = xmalloc(sizeof(*on_destroy));
+		trigger_create(on_destroy, lbox_fiber_on_destroy, NULL,
+			       (trigger_f0)free);
+		trigger_add(&f->on_destroy, on_destroy);
 
 		uint64_t fid = f->fid;
 		/* create a new userdata */
@@ -138,6 +143,16 @@ lbox_pushfiber(struct lua_State *L, struct fiber *f)
 	lua_rawgeti(L, LUA_REGISTRYINDEX, fid_ref);
 }
 
+/**
+ * Returns fiber id for the fiber located at the given `index` on the Lua stack.
+ */
+static uint64_t
+lbox_checkfiberid(struct lua_State *L, int index)
+{
+	uint64_t *fid = luaL_checkudata(L, index, fiberlib_name);
+	return *fid;
+}
+
 static struct fiber *
 lbox_checkfiber(struct lua_State *L, int index)
 {
@@ -145,7 +160,7 @@ lbox_checkfiber(struct lua_State *L, int index)
 	if (lua_type(L, index) == LUA_TNUMBER) {
 		fid = luaL_touint64(L, index);
 	} else {
-		fid = *(uint64_t *)luaL_checkudata(L, index, fiberlib_name);
+		fid = lbox_checkfiberid(L, index);
 	}
 	struct fiber *f = fiber_find(fid);
 	if (f == NULL)
@@ -160,7 +175,7 @@ lbox_fiber_id(struct lua_State *L)
 	if (lua_gettop(L)  == 0)
 		fid = fiber()->fid;
 	else
-		fid = *(uint64_t *)luaL_checkudata(L, 1, fiberlib_name);
+		fid = lbox_checkfiberid(L, 1);
 	luaL_pushuint64(L, fid);
 	return 1;
 }
@@ -506,8 +521,8 @@ static struct fiber *
 lbox_get_fiber(struct lua_State *L)
 {
 	if (lua_gettop(L) != 0) {
-		uint64_t *fid = luaL_checkudata(L, 1, fiberlib_name);
-		return fiber_find(*fid);
+		uint64_t fid = lbox_checkfiberid(L, 1);
+		return fiber_find(fid);
 	} else {
 		return fiber();
 	}
@@ -632,6 +647,14 @@ lbox_fiber_storage(struct lua_State *L)
 	struct fiber *f = lbox_checkfiber(L, 1);
 	int storage_ref = f->storage.lua.storage_ref;
 	if (storage_ref == FIBER_LUA_NOREF) {
+		if (f->flags & FIBER_IS_DEAD) {
+			diag_set(IllegalParams, "the fiber is dead");
+			luaT_error(L);
+		}
+		struct trigger *on_stop = xmalloc(sizeof(*on_stop));
+		trigger_create(on_stop, lbox_fiber_on_stop, NULL,
+			       (trigger_f0)free);
+		trigger_add(&f->on_stop, on_stop);
 		lua_newtable(L); /* create local storage on demand */
 		storage_ref = luaL_ref(L, LUA_REGISTRYINDEX);
 		f->storage.lua.storage_ref = storage_ref;
@@ -726,12 +749,16 @@ lbox_fiber_cancel(struct lua_State *L)
 static int
 lbox_fiber_serialize(struct lua_State *L)
 {
-	struct fiber *f = lbox_checkfiber(L, 1);
+	uint64_t fid = lbox_checkfiberid(L, 1);
+	struct fiber *f = fiber_find(fid);
+
 	lua_createtable(L, 0, 1);
-	luaL_pushuint64(L, f->fid);
+	luaL_pushuint64(L, fid);
 	lua_setfield(L, -2, "id");
-	lua_pushstring(L, fiber_name(f));
-	lua_setfield(L, -2, "name");
+	if (f != NULL) {
+		lua_pushstring(L, fiber_name(f));
+		lua_setfield(L, -2, "name");
+	}
 	lbox_fiber_status(L);
 	lua_setfield(L, -2, "status");
 	return 1;
@@ -740,10 +767,12 @@ lbox_fiber_serialize(struct lua_State *L)
 static int
 lbox_fiber_tostring(struct lua_State *L)
 {
-	char buf[32];
-	struct fiber *f = lbox_checkfiber(L, 1);
-	snprintf(buf, sizeof(buf), "fiber: %llu",
-		 (long long)f->fid);
+	char buf[35];
+	uint64_t fid = lbox_checkfiberid(L, 1);
+	bool is_dead = fiber_find(fid) == NULL;
+
+	snprintf(buf, sizeof(buf), "fiber: %llu%s", (long long)fid,
+		 is_dead ? " (dead)" : "");
 	lua_pushstring(L, buf);
 	return 1;
 }
@@ -782,6 +811,8 @@ lbox_fiber_join(struct lua_State *L)
 
 	if (!(fiber->flags & FIBER_IS_JOINABLE))
 		luaL_error(L, "the fiber is not joinable");
+	if (fid == fiber()->fid)
+		luaL_error(L, "cannot join itself");
 	double timeout = TIMEOUT_INFINITY;
 	if (!lua_isnoneornil(L, 2)) {
 		if (!lua_isnumber(L, 2) ||
