@@ -412,7 +412,7 @@ const char httpc_lua[] =
 "    end\n"
 "\n"
 "    if chunk < 0 then\n"
-"        error('io:read(): chunk can not be negative')\n"
+"        error('io:read(): chunk size can not be negative')\n"
 "    end\n"
 "\n"
 "    chunk = math.min(chunk, IO_CHUNK_DEFAULT)\n"
@@ -425,7 +425,7 @@ const char httpc_lua[] =
 "    local len = check(self, chunk, delimiter)\n"
 "    if len ~= nil then\n"
 "        local data = ffi.string(rbuf.rpos, len)\n"
-"        rbuf.rpos = rbuf.rpos + len\n"
+"        rbuf:consume(len)\n"
 "        return data\n"
 "    end\n"
 "\n"
@@ -439,15 +439,15 @@ const char httpc_lua[] =
 "            self._errno = nil\n"
 "            local len = rbuf:size()\n"
 "            local data = ffi.string(rbuf.rpos, len)\n"
-"            rbuf.rpos = rbuf.rpos + len\n"
+"            rbuf:consume(len)\n"
 "            return data\n"
 "        else\n"
-"            rbuf.wpos = rbuf.wpos + res\n"
+"            rbuf:alloc(res)\n"
 "            local len = check(self, chunk, delimiter)\n"
 "            if len ~= nil then\n"
 "                self._errno = nil\n"
 "                local data = ffi.string(rbuf.rpos, len)\n"
-"                rbuf.rpos = rbuf.rpos + len\n"
+"                rbuf:consume(len)\n"
 "                return data\n"
 "            end\n"
 "        end\n"
@@ -615,6 +615,9 @@ const char httpc_lua[] =
 "                        write = io_write,\n"
 "                        finish = io_finish,\n"
 "                    },\n"
+"                    _internal = {\n"
+"                        httpc = self,\n"
+"                    },\n"
 "                })\n"
 "            end\n"
 "        end,\n"
@@ -752,6 +755,7 @@ const char httpc_lua[] =
 "                     'connect', 'post', 'put', 'patch', 'request'}) do\n"
 "    this_module[name] = http_default_wrap(name)\n"
 "end\n"
+"this_module.curl = http_default.curl\n"
 "\n"
 "return this_module\n"
 ""
