@@ -135,6 +135,7 @@ extern char minifio_lua[],
 	tap_lua[],
 	fio_lua[],
 	error_lua[],
+	utils_lua[],
 	argparse_lua[],
 	iconv_lua[],
 	/* jit.* library */
@@ -162,13 +163,13 @@ extern char minifio_lua[],
 	/* tools.* libraries. */
 	utils_avl_lua[],
 	utils_bufread_lua[],
+	utils_evread_lua[],
 	utils_symtab_lua[],
 	memprof_parse_lua[],
 	memprof_process_lua[],
 	memprof_humanize_lua[],
 	memprof_lua[],
 	sysprof_parse_lua[],
-	sysprof_collapse_lua[],
 	sysprof_lua[],
 	datetime_lua[],
 	timezones_lua[],
@@ -186,6 +187,7 @@ extern char minifio_lua[],
 	luarocks_admin_cmd_refresh_cache_lua[],
 	luarocks_admin_cmd_remove_lua[],
 	luarocks_admin_index_lua[],
+	luarocks_argparse_lua[],
 	luarocks_build_builtin_lua[],
 	luarocks_build_cmake_lua[],
 	luarocks_build_command_lua[],
@@ -195,7 +197,6 @@ extern char minifio_lua[],
 	luarocks_cmd_config_lua[],
 	luarocks_cmd_doc_lua[],
 	luarocks_cmd_download_lua[],
-	luarocks_cmd_help_lua[],
 	luarocks_cmd_init_lua[],
 	luarocks_cmd_install_lua[],
 	luarocks_cmd_lint_lua[],
@@ -222,6 +223,7 @@ extern char minifio_lua[],
 	luarocks_core_sysdetect_lua[],
 	luarocks_core_vers_lua[],
 	luarocks_deps_lua[],
+	luarocks_deplocks_lua[],
 	luarocks_dir_lua[],
 	luarocks_download_lua[],
 	luarocks_fetch_cvs_lua[],
@@ -277,6 +279,7 @@ static const char *lua_modules[] = {
 	/* Make it first to affect load of all other modules */
 	"strict", strict_lua,
 	"compat", compat_lua,
+	"internal.utils", utils_lua,
 	"fun", fun_lua,
 	"debug", debug_lua,
 	"tarantool", init_lua,
@@ -328,12 +331,12 @@ static const char *lua_modules[] = {
 	"utils.avl", utils_avl_lua,
 	"utils.bufread", utils_bufread_lua,
 	"utils.symtab", utils_symtab_lua,
+	"utils.evread", utils_evread_lua,
 	"memprof.parse", memprof_parse_lua,
 	"memprof.process", memprof_process_lua,
 	"memprof.humanize", memprof_humanize_lua,
 	"memprof", memprof_lua,
 	"sysprof.parse", sysprof_parse_lua,
-	"sysprof.collapse", sysprof_collapse_lua,
 	"sysprof", sysprof_lua,
 	"timezones", timezones_lua,
 	"datetime", datetime_lua,
@@ -386,6 +389,7 @@ static const char *lua_modules_preload[] = {
 	"luarocks.manif", luarocks_manif_lua,
 	"luarocks.build.builtin", luarocks_build_builtin_lua,
 	"luarocks.deps", luarocks_deps_lua,
+	"luarocks.deplocks", luarocks_deplocks_lua,
 	"luarocks.cmd", luarocks_cmd_lua,
 	"luarocks.test.busted", luarocks_test_busted_lua,
 	"luarocks.test.command", luarocks_test_command_lua,
@@ -418,7 +422,6 @@ static const char *lua_modules_preload[] = {
 	"luarocks.upload.multipart", luarocks_upload_multipart_lua,
 	"luarocks.upload.api", luarocks_upload_api_lua,
 	"luarocks.cmd.upload", luarocks_cmd_upload_lua,
-	"luarocks.cmd.help", luarocks_cmd_help_lua,
 	"luarocks.cmd.doc", luarocks_cmd_doc_lua,
 	"luarocks.cmd.unpack", luarocks_cmd_unpack_lua,
 	"luarocks.cmd.config", luarocks_cmd_config_lua,
@@ -444,6 +447,7 @@ static const char *lua_modules_preload[] = {
 	"luarocks.admin.cmd.add", luarocks_admin_cmd_add_lua,
 	"luarocks.admin.cmd.remove", luarocks_admin_cmd_remove_lua,
 	"luarocks.admin.cmd.make_manifest", luarocks_admin_cmd_make_manifest_lua,
+	"luarocks.argparse", luarocks_argparse_lua,
 #endif /* defined(EMBED_LUAROCKS) */
 	NULL
 };
@@ -741,15 +745,7 @@ luaopen_tarantool(lua_State *L)
 	lua_pushstring(L, tarantool_version());
 	lua_setfield(L, LUA_GLOBALSINDEX, "_TARANTOOL");
 
-	/*
-	 * Get tarantool module.
-	 *
-	 * src/lua/init.lua is already registered as tarantool
-	 * module, so we `require` it here, not create.
-	 */
-	lua_getfield(L, LUA_GLOBALSINDEX, "require");
-	lua_pushstring(L, "tarantool");
-	lua_call(L, 1, 1);
+	luaT_newmodule(L, "tarantool", NULL);
 
 	/* package */
 	lua_pushstring(L, tarantool_package());
@@ -797,6 +793,15 @@ luaopen_tarantool(lua_State *L)
 #endif
 	lua_settable(L, -3);
 
+	/* build.asan */
+	lua_pushstring(L, "asan");
+#ifdef ENABLE_ASAN
+	lua_pushboolean(L, true);
+#else
+	lua_pushboolean(L, false);
+#endif
+	lua_settable(L, -3);
+
 	lua_settable(L, -3);    /* box.info.build */
 
 	/* debug */
@@ -937,6 +942,7 @@ tarantool_lua_init(const char *tarantool_bin, const char *script, int argc,
 #if defined(ENABLE_BACKTRACE)
 	backtrace_lua_init();
 #endif /* defined(ENABLE_BACKTRACE) */
+	luaopen_tarantool(L);
 	for (const char **s = lua_modules; *s; s += 2) {
 		const char *modname = *s;
 		const char *modsrc = *(s + 1);
@@ -957,9 +963,6 @@ tarantool_lua_init(const char *tarantool_bin, const char *script, int argc,
 		builtin_modcache_put(modname, modsrc);
 	}
 	lua_pop(L, 1); /* _PRELOAD */
-
-	luaopen_tarantool(L);
-
 #ifdef NDEBUG
 	/* Unload strict after boot in release mode */
 	if (luaL_dostring(L, "require('strict').off()") != 0)
@@ -1055,9 +1058,18 @@ run_script_f(va_list ap)
 	 * never really dead. It never returns from its function.
 	 */
 	struct diag *diag = va_arg(ap, struct diag *);
-	bool aux_loop_is_run = false;
 	bool is_option_e_ran = false;
 
+	/*
+	 * Return control to tarantool_lua_run_script.
+	 * tarantool_lua_run_script then will start an auxiliary event
+	 * loop and re-schedule this fiber.
+	 *
+	 * This also update time in ev after Tarantool startup which
+	 * reduce time slip in timers (see #9261).
+	 */
+	fiber_sleep(0.0);
+
 	/*
 	 * Execute scripts or modules pointed by TT_PRELOAD
 	 * environment variable.
@@ -1116,14 +1128,6 @@ run_script_f(va_list ap)
 		}
 	}
 
-	/*
-	 * Return control to tarantool_lua_run_script.
-	 * tarantool_lua_run_script then will start an auxiliary event
-	 * loop and re-schedule this fiber.
-	 */
-	fiber_sleep(0.0);
-	aux_loop_is_run = true;
-
 	int is_a_tty = isatty(STDIN_FILENO);
 
 	if (bytecode) {
@@ -1187,13 +1191,6 @@ run_script_f(va_list ap)
 	 * return control back to tarantool_lua_run_script.
 	 */
 end:
-	/*
-	 * Auxiliary loop in tarantool_lua_run_script
-	 * should start (ev_run()) before this fiber
-	 * invokes ev_break().
-	 */
-	if (!aux_loop_is_run)
-		fiber_sleep(0.0);
 	ev_break(loop(), EVBREAK_ALL);
 	return 0;
 
@@ -1236,8 +1233,7 @@ tarantool_lua_run_script(char *path, uint32_t opt_mask,
 	 * Run an auxiliary event loop to re-schedule run_script fiber.
 	 * When this fiber finishes, it will call ev_break to stop the loop.
 	 */
-	if (start_loop)
-		ev_run(loop(), 0);
+	ev_run(loop(), 0);
 	/* The fiber running the startup script has ended. */
 	script_fiber = NULL;
 	diag_move(&script_diag, diag_get());
