@@ -695,7 +695,7 @@ const char socket_lua[] =
 "    if len ~= nil then\n"
 "        self._errno = nil\n"
 "        local data = ffi.string(rbuf.rpos, len)\n"
-"        rbuf.rpos = rbuf.rpos + len\n"
+"        rbuf:consume(len)\n"
 "        return data\n"
 "    end\n"
 "\n"
@@ -710,15 +710,15 @@ const char socket_lua[] =
 "            self._errno = nil\n"
 "            local len = rbuf:size()\n"
 "            local data = ffi.string(rbuf.rpos, len)\n"
-"            rbuf.rpos = rbuf.rpos + len\n"
+"            rbuf:consume(len)\n"
 "            return data\n"
 "        elseif res ~= nil then\n"
-"            rbuf.wpos = rbuf.wpos + res\n"
+"            rbuf:alloc(res)\n"
 "            local len = check(self, limit, ...)\n"
 "            if len ~= nil then\n"
 "                self._errno = nil\n"
 "                local data = ffi.string(rbuf.rpos, len)\n"
-"                rbuf.rpos = rbuf.rpos + len\n"
+"                rbuf:consume(len)\n"
 "                return data\n"
 "            end\n"
 "        elseif not errno_is_transient[self._errno] then\n"
